@@ -69,7 +69,6 @@ pub const RL_CPU_ANY: i32 = bpf_intf::RL_CPU_ANY as i32;
 ///
 /// Finally the methods exited() and shutdown_and_report() can be used respectively to test
 /// whether the BPF component exited, and to shutdown and report the exit message.
-/// whether the BPF component exited, and to shutdown and report exit message.
 
 // Task queued for scheduling from the BPF component (see bpf_intf::queued_task_ctx).
 #[derive(Debug, PartialEq, Eq, PartialOrd, Clone)]
@@ -77,6 +76,7 @@ pub struct QueuedTask {
     pub pid: i32,              // pid that uniquely identifies a task
     pub cpu: i32,              // CPU where the task is running
     pub flags: u64,            // task enqueue flags
+    pub exec_runtime: u64,     // Total cpu time since last sleep
     pub sum_exec_runtime: u64, // Total cpu time
     pub nvcsw: u64,            // Total amount of voluntary context switches
     pub weight: u64,           // Task static priority
@@ -142,6 +142,7 @@ impl EnqueuedMessage {
             pid: self.inner.pid,
             cpu: self.inner.cpu,
             flags: self.inner.flags,
+            exec_runtime: self.inner.exec_runtime,
             sum_exec_runtime: self.inner.sum_exec_runtime,
             nvcsw: self.inner.nvcsw,
             weight: self.inner.weight,
@@ -193,6 +194,7 @@ impl<'cb> BpfScheduler<'cb> {
         exit_dump_len: u32,
         partial: bool,
         debug: bool,
+        builtin_idle: bool,
     ) -> Result<Self> {
         let shutdown = Arc::new(AtomicBool::new(false));
         set_ctrlc_handler(shutdown.clone()).context("Error setting Ctrl-C handler")?;
@@ -202,10 +204,6 @@ impl<'cb> BpfScheduler<'cb> {
         skel_builder.obj_builder.debug(debug);
         let mut skel = scx_ops_open!(skel_builder, open_object, rustland)?;
 
-        // Lock all the memory to prevent page faults that could trigger potential deadlocks during
-        // scheduling.
-        ALLOCATOR.lock_memory();
-
         // Copy one item from the ring buffer.
         //
         // # Safety
@@ -246,13 +244,16 @@ impl<'cb> BpfScheduler<'cb> {
         let topo = Topology::new().unwrap();
         skel.maps.rodata_data.smt_enabled = topo.smt_enabled;
 
-        // Set scheduler options (defined in the BPF part).
+        // Enable optional scheduler settings.
+        skel.struct_ops.rustland_mut().flags |=
+            *compat::SCX_OPS_ALLOW_QUEUED_WAKEUP | *compat::SCX_OPS_ENQ_MIGRATION_DISABLED;
         if partial {
             skel.struct_ops.rustland_mut().flags |= *compat::SCX_OPS_SWITCH_PARTIAL;
         }
         skel.struct_ops.rustland_mut().exit_dump_len = exit_dump_len;
 
         skel.maps.bss_data.usersched_pid = std::process::id();
+        skel.maps.rodata_data.builtin_idle = builtin_idle;
         skel.maps.rodata_data.debug = debug;
 
         // Attach BPF scheduler.
@@ -276,6 +277,11 @@ impl<'cb> BpfScheduler<'cb> {
         let dispatched = libbpf_rs::UserRingBuffer::new(&maps.dispatched)
             .expect("failed to create user ringbuf");
 
+        // Lock all the memory to prevent page faults that could trigger potential deadlocks during
+        // scheduling.
+        ALLOCATOR.lock_memory();
+        ALLOCATOR.disable_mmap().expect("Failed to disable mmap");
+
         // Make sure to use the SCHED_EXT class at least for the scheduler itself.
         match Self::use_sched_ext() {
             0 => Ok(Self {
@@ -379,7 +385,7 @@ impl<'cb> BpfScheduler<'cb> {
     }
 
     // Notify the BPF component that the user-space scheduler has completed its scheduling cycle,
-    // updating the amount tasks that are still peding.
+    // updating the amount tasks that are still pending.
     //
     // NOTE: do not set allow(dead_code) for this method, any scheduler must use this method at
     // some point, otherwise the BPF component will keep waking-up the user-space scheduler in a
@@ -504,7 +510,7 @@ impl<'cb> BpfScheduler<'cb> {
             LIBBPF_STOP => {
                 // A valid task is received, convert data to a proper task struct.
                 let task = unsafe { EnqueuedMessage::from_bytes(&BUF.0).to_queued_task() };
-                let _ = self.skel.maps.bss_data.nr_queued.saturating_sub(1);
+                self.skel.maps.bss_data.nr_queued = self.skel.maps.bss_data.nr_queued.saturating_sub(1);
 
                 Ok(Some(task))
             }
