@@ -53,6 +53,7 @@ UEI_DEFINE(uei);
  * Scheduler attributes and statistics.
  */
 u32 usersched_pid; /* User-space scheduler PID */
+u64 usersched_last_run_at; /* Timestamp of the last user-space scheduler execution */
 const volatile bool switch_partial; /* Switch all tasks or SCHED_EXT tasks */
 
 /*
@@ -87,6 +88,9 @@ volatile u64 nr_failed_dispatches, nr_sched_congested;
  /* Report additional debugging information */
 const volatile bool debug;
 
+/* Rely on the in-kernel idle CPU selection policy */
+const volatile bool builtin_idle;
+
 /* Allow to use bpf_printk() only when @debug is set */
 #define dbg_msg(_fmt, ...) do {						\
 	if (debug)							\
@@ -190,9 +194,14 @@ struct task_ctx {
 	struct bpf_cpumask __kptr *l3_cpumask;
 
 	/*
-	 * Time slice assigned to the task.
+	 * Timestamp since last time the task ran on a CPU.
 	 */
-	u64 slice_ns;
+	u64 last_run_at;
+
+	/*
+	 * Execution time (in nanoseconds) since the last sleep event.
+	 */
+	u64 exec_runtime;
 
 	/*
 	 * cpumask generation counter: used to verify the validity of the
@@ -223,46 +232,6 @@ struct task_ctx *try_lookup_task_ctx(const struct task_struct *p)
 	return tctx;
 }
 
-/*
- * Intercept when a task is executing __handle_mm_fault().
- */
-struct {
-	__uint(type, BPF_MAP_TYPE_HASH);
-	__type(key, __u32);
-	__type(value, __u64);
-	__uint(max_entries, MAX_ENQUEUED_TASKS);
-} pid_mm_fault_map SEC(".maps");
-
-SEC("fentry/handle_mm_fault")
-int BPF_PROG(kprobe_handle_mm_fault, void *vma,
-	     unsigned long address, unsigned int flags)
-{
-	pid_t pid = bpf_get_current_pid_tgid() >> 32;
-	u64 value = true;
-
-	bpf_map_update_elem(&pid_mm_fault_map, &pid, &value, BPF_ANY);
-
-	return 0;
-}
-
-SEC("fexit/handle_mm_fault")
-int BPF_PROG(kretprobe_handle_mm_fault)
-{
-	pid_t pid = bpf_get_current_pid_tgid() >> 32;
-	bpf_map_delete_elem(&pid_mm_fault_map, &pid);
-
-	return 0;
-}
-
-/*
- * Return true if a task is handling a page fault, false otherwise.
- */
-static bool in_mm_fault(pid_t pid)
-{
-	u64 *value = bpf_map_lookup_elem(&pid_mm_fault_map, &pid);
-	return value != NULL;
-}
-
 /*
  * Heartbeat timer used to periodically trigger the check to run the user-space
  * scheduler.
@@ -303,6 +272,14 @@ static inline bool is_kthread(const struct task_struct *p)
 	return p->flags & PF_KTHREAD;
 }
 
+/*
+ * Return true if @p still wants to run, false otherwise.
+ */
+static bool is_queued(const struct task_struct *p)
+{
+	return p->scx.flags & SCX_TASK_QUEUED;
+}
+
 /*
  * Flag used to wake-up the user-space scheduler.
  */
@@ -330,20 +307,20 @@ static bool test_and_clear_usersched_needed(void)
  * Return true if there's any pending activity to do for the scheduler, false
  * otherwise.
  *
- * NOTE: nr_queued is incremented by the BPF component, more exactly in
- * enqueue(), when a task is sent to the user-space scheduler, then the
- * scheduler drains the queued tasks (updating nr_queued) and adds them to its
- * internal data structures / state; at this point tasks become "scheduled" and
- * the user-space scheduler will take care of updating nr_scheduled
- * accordingly; lastly tasks will be dispatched and the user-space scheduler
- * will update nr_scheduled again.
+ * NOTE: a task is sent to the user-space scheduler using the "queued"
+ * ringbuffer, then the scheduler drains the queued tasks and adds them to
+ * its internal data structures / state; at this point tasks become
+ * "scheduled" and the user-space scheduler will take care of updating
+ * nr_scheduled accordingly; lastly tasks will be dispatched and the
+ * user-space scheduler will update nr_scheduled again.
  *
- * Checking both counters allows to determine if there is still some pending
- * work to do for the scheduler: new tasks have been queued since last check,
- * or there are still tasks "queued" or "scheduled" since the previous
- * user-space scheduler run. If the counters are both zero it is pointless to
- * wake-up the scheduler (even if a CPU becomes idle), because there is nothing
- * to do.
+ * Checking nr_scheduled and the available data in the ringbuffer allows to
+ * determine if there is still some pending work to do for the scheduler:
+ * new tasks have been queued since last check, or there are still tasks
+ * "queued" or "scheduled" since the previous user-space scheduler run.
+ *
+ * If there's no pending action, it is pointless to wake-up the scheduler
+ * (even if a CPU becomes idle), because there is nothing to do.
  *
  * Also keep in mind that we don't need any protection here since this code
  * doesn't run concurrently with the user-space scheduler (that is single
@@ -351,7 +328,10 @@ static bool test_and_clear_usersched_needed(void)
  */
 static bool usersched_has_pending_tasks(void)
 {
-	return nr_queued || nr_scheduled;
+	if (nr_scheduled)
+		return true;
+
+	return bpf_ringbuf_query(&queued, BPF_RB_AVAIL_DATA) > 0;
 }
 
 /*
@@ -367,19 +347,6 @@ static u64 cpu_to_dsq(s32 cpu)
 	return (u64)cpu;
 }
 
-/*
- * Return the time slice assigned to the task.
- */
-static inline u64 task_slice(struct task_struct *p)
-{
-	struct task_ctx *tctx;
-
-	tctx = try_lookup_task_ctx(p);
-	if (!tctx || !tctx->slice_ns)
-		return SCX_SLICE_DFL;
-	return tctx->slice_ns;
-}
-
 /*
  * Find an idle CPU in the system for the task.
  *
@@ -585,31 +552,31 @@ static void dispatch_task(const struct dispatched_task_ctx *task)
 	if (!p)
 		return;
 
-	dbg_msg("dispatch: pid=%d (%s) cpu=0x%lx vtime=%llu slice=%llu",
-		p->pid, p->comm, task->cpu, task->vtime, task->slice_ns);
-
 	/*
 	 * Update task's time slice in its context.
 	 */
 	tctx = try_lookup_task_ctx(p);
-	if (!tctx) {
-		/*
-		 * Bounce to the shared DSQ if we can't find a valid task
-		 * context.
-		 */
-		scx_bpf_dsq_insert_vtime(p, SHARED_DSQ,
-					 SCX_SLICE_DFL, task->vtime, task->flags);
-		__sync_fetch_and_add(&nr_bounce_dispatches, 1);
-		goto out_kick_idle_cpu;
-	}
-	tctx->slice_ns = task->slice_ns;
+	if (!tctx)
+		goto out_release;
+
+	dbg_msg("dispatch: pid=%d (%s) cpu=0x%lx vtime=%llu slice=%llu",
+		p->pid, p->comm, task->cpu, task->vtime, task->slice_ns);
 
 	/*
 	 * Dispatch task to the target DSQ.
 	 */
-	if (task->cpu & RL_CPU_ANY) {
-		scx_bpf_dsq_insert_vtime(p, SHARED_DSQ,
-					 SCX_SLICE_DFL, task->vtime, task->flags);
+	if (task->cpu == RL_CPU_ANY) {
+		scx_bpf_dsq_insert_vtime(p, SHARED_DSQ, task->slice_ns, task->vtime, task->flags);
+		goto out_kick_idle_cpu;
+	}
+
+	/* Read current cpumask generation counter */
+	curr_cpumask_cnt = tctx->cpumask_cnt;
+
+	/* Check if the CPU is valid, according to the cpumask */
+	if (!bpf_cpumask_test_cpu(task->cpu, p->cpus_ptr)) {
+		scx_bpf_dsq_insert_vtime(p, SHARED_DSQ, task->slice_ns, task->vtime, task->flags);
+		__sync_fetch_and_add(&nr_bounce_dispatches, 1);
 		goto out_kick_idle_cpu;
 	}
 
@@ -629,20 +596,8 @@ static void dispatch_task(const struct dispatched_task_ctx *task)
 	 */
 	dsq_id = cpu_to_dsq(task->cpu);
 
-	/* Check if the CPU is valid, according to the cpumask */
-	if (!bpf_cpumask_test_cpu(task->cpu, p->cpus_ptr)) {
-		scx_bpf_dsq_insert_vtime(p, SHARED_DSQ,
-					 SCX_SLICE_DFL, task->vtime, task->flags);
-		__sync_fetch_and_add(&nr_bounce_dispatches, 1);
-		goto out_kick_idle_cpu;
-	}
-
-	/* Read current cpumask generation counter */
-	curr_cpumask_cnt = tctx->cpumask_cnt;
-
 	/* Dispatch the task to the target per-CPU DSQ */
-	scx_bpf_dsq_insert_vtime(p, dsq_id,
-				 SCX_SLICE_DFL, task->vtime, task->flags);
+	scx_bpf_dsq_insert_vtime(p, dsq_id, task->slice_ns, task->vtime, task->flags);
 
 	/* If the cpumask is not valid anymore, ignore the dispatch event */
 	if (curr_cpumask_cnt != task->cpumask_cnt) {
@@ -668,11 +623,24 @@ out_release:
 s32 BPF_STRUCT_OPS(rustland_select_cpu, struct task_struct *p, s32 prev_cpu,
 		   u64 wake_flags)
 {
+	bool is_idle = false;
+	s32 cpu;
+
 	/*
-	 * Completely delegate the CPU selection logic to the user-space
-	 * scheduler.
+	 * If built-in idle CPU policy is not enabled completely delegate
+	 * the idle selection policy to user-space and keep re-using the
+	 * same CPU here.
 	 */
-	return prev_cpu;
+	if (!builtin_idle || is_usersched_task(p))
+		return prev_cpu;
+
+	cpu = scx_bpf_select_cpu_dfl(p, prev_cpu, wake_flags, &is_idle);
+	if (is_idle && !scx_bpf_dsq_nr_queued(SHARED_DSQ)) {
+		scx_bpf_dsq_insert_vtime(p, cpu_to_dsq(cpu), SCX_SLICE_DFL, p->scx.dsq_vtime, 0);
+		__sync_fetch_and_add(&nr_kernel_dispatches, 1);
+	}
+
+	return cpu;
 }
 
 /*
@@ -695,15 +663,6 @@ int rs_select_cpu(struct task_cpu_arg *input)
 
 	bpf_task_release(p);
 
-	/*
-	 * Wake-up the CPU if idle. Use SCX_KICK_IDLE to prevent unecessary
-	 * rescheduling events in case the CPU is already awake (since we don't
-	 * know exactly what the user-space scheduler is doing we can't
-	 * implicitly assume that the target CPU is idle here).
-	 */
-	if (cpu >= 0)
-		scx_bpf_kick_cpu(cpu, SCX_KICK_IDLE);
-
 	return cpu;
 }
 
@@ -719,6 +678,7 @@ static void get_task_info(struct queued_task_ctx *task,
 	task->pid = p->pid;
 	task->cpu = scx_bpf_task_cpu(p);
 	task->flags = enq_flags;
+	task->exec_runtime = tctx ? tctx->exec_runtime : 0;
 	task->sum_exec_runtime = p->se.sum_exec_runtime;
 	task->nvcsw = p->nvcsw;
 	task->weight = p->scx.weight;
@@ -737,23 +697,12 @@ static void sched_congested(struct task_struct *p)
 }
 
 /*
- * Try to wake up the CPU that was assigned to task @p.
+ * Return true if a task has been enqueued as a remote wakeup, false
+ * otherwise.
  */
-static void kick_task_cpu(struct task_struct *p)
+static bool is_queued_wakeup(const struct task_struct *p, u64 enq_flags)
 {
-	s32 cpu = scx_bpf_task_cpu(p);
-
-	/*
-	 * If the assigned CPU is not usable pick any other CPU usable by the
-	 * task.
-	 */
-	if (!bpf_cpumask_test_cpu(cpu, p->cpus_ptr))
-		cpu = scx_bpf_pick_any_cpu(p->cpus_ptr, 0);
-
-	/*
-	 * Wake up the selected CPU if idle.
-	 */
-	scx_bpf_kick_cpu(cpu, SCX_KICK_IDLE);
+	return !__COMPAT_is_enq_cpu_selected(enq_flags) && !scx_bpf_task_running(p);
 }
 
 /*
@@ -764,6 +713,7 @@ static void kick_task_cpu(struct task_struct *p)
 void BPF_STRUCT_OPS(rustland_enqueue, struct task_struct *p, u64 enq_flags)
 {
 	struct queued_task_ctx *task;
+	s32 cpu;
 
 	/*
 	 * Scheduler is dispatched directly in .dispatch() when needed, so
@@ -787,15 +737,19 @@ void BPF_STRUCT_OPS(rustland_enqueue, struct task_struct *p, u64 enq_flags)
 	}
 
 	/*
-	 * Bypass user-space scheduling for faulting tasks to prevent potential
-	 * deadlock conditions. They can just be dispatched to the shared DSQ
-	 * using ith the highest priority.
+	 * Give the task a chance to be directly dispatched if
+	 * ops.select_cpu() was skipped.
 	 */
-	if (in_mm_fault(p->pid)) {
-		scx_bpf_dsq_insert_vtime(p, SHARED_DSQ, SCX_SLICE_DFL, 0, enq_flags);
-		__sync_fetch_and_add(&nr_kernel_dispatches, 1);
-		kick_task_cpu(p);
-		return;
+	if (builtin_idle && is_queued_wakeup(p, enq_flags)) {
+		s32 cpu = pick_idle_cpu(p, scx_bpf_task_cpu(p));
+
+		if (cpu >= 0) {
+			scx_bpf_dsq_insert_vtime(p, cpu_to_dsq(cpu),
+					   SCX_SLICE_DFL, p->scx.dsq_vtime, enq_flags);
+			__sync_fetch_and_add(&nr_kernel_dispatches, 1);
+			scx_bpf_kick_cpu(cpu, SCX_KICK_IDLE);
+			return;
+		}
 	}
 
 	/*
@@ -809,10 +763,8 @@ void BPF_STRUCT_OPS(rustland_enqueue, struct task_struct *p, u64 enq_flags)
 	task = bpf_ringbuf_reserve(&queued, sizeof(*task), 0);
 	if (!task) {
 		sched_congested(p);
-		scx_bpf_dsq_insert_vtime(p, SHARED_DSQ,
-					 SCX_SLICE_DFL, 0, enq_flags);
+		scx_bpf_dsq_insert_vtime(p, SHARED_DSQ, SCX_SLICE_DFL, p->scx.dsq_vtime, enq_flags);
 		__sync_fetch_and_add(&nr_kernel_dispatches, 1);
-		kick_task_cpu(p);
 		return;
 	}
 	get_task_info(task, p, enq_flags);
@@ -820,6 +772,10 @@ void BPF_STRUCT_OPS(rustland_enqueue, struct task_struct *p, u64 enq_flags)
 	bpf_ringbuf_submit(task, 0);
 
 	__sync_fetch_and_add(&nr_queued, 1);
+
+	cpu = scx_bpf_task_cpu(p);
+	if (cpu != bpf_get_smp_processor_id())
+		scx_bpf_kick_cpu(cpu, SCX_KICK_IDLE);
 }
 
 /*
@@ -839,17 +795,17 @@ static bool dispatch_user_scheduler(void)
 	}
 
 	/*
-	 * Use the highest vtime possible to give the scheduler itself the
-	 * lowest priority possible.
+	 * Always append the user-space scheduler at the end of the shared
+	 * DSQ, so that it'll run after all the tasks currently dispatched
+	 * have used their assigned time slice on their target CPU.
 	 *
-	 * At the same time make sure to assign an infinite time slice, so that
-	 * it can completely drain all the pending tasks.
+	 * At the same time assign an infinite time slice, so that it can
+	 * completely drain all the pending tasks.
 	 *
 	 * The user-space scheduler will voluntarily yield the CPU upon
 	 * completion through BpfScheduler->notify_complete().
 	 */
 	scx_bpf_dsq_insert_vtime(p, SHARED_DSQ, SCX_SLICE_INF, -1ULL, 0);
-	kick_task_cpu(p);
 
 	bpf_task_release(p);
 
@@ -886,21 +842,16 @@ static long handle_dispatched_task(struct bpf_dynptr *dynptr, void *context)
  */
 void BPF_STRUCT_OPS(rustland_dispatch, s32 cpu, struct task_struct *prev)
 {
-	struct cpu_ctx *cctx;
-
-	cctx = try_lookup_cpu_ctx(cpu);
-	if (!cctx)
-		return;
-
 	/*
 	 * Consume all tasks from the @dispatched list and immediately dispatch
 	 * them on the target CPU decided by the user-space scheduler.
 	 */
 	bpf_user_ringbuf_drain(&dispatched, handle_dispatched_task, NULL, 0);
 
-	/*
-	 * Check if the user-space scheduler needs to run.
-	 */
+       /*
+	* Always dispatch the user-space scheduler every time that a CPU
+	* becomes available.
+	*/
 	dispatch_user_scheduler();
 
 	/*
@@ -916,36 +867,75 @@ void BPF_STRUCT_OPS(rustland_dispatch, s32 cpu, struct task_struct *prev)
 		return;
 
 	/*
-	 * No more tasks to process, check if we need to keep the CPU alive to
-	 * process pending tasks from the user-space scheduler.
+	 * If the previous task was the user-space scheduler and it
+	 * voluntarily released the CPU without dispatching any task, it
+	 * means that there are no pending actions to be done, so let the
+	 * CPU go idle.
 	 */
-	if (usersched_has_pending_tasks()) {
-		set_usersched_needed();
-		scx_bpf_kick_cpu(cpu, 0);
+	if (prev && is_usersched_task(prev))
+		return;
+
+	/*
+	 * If the current task expired its time slice and no other task
+	 * wants to run, simply replenish its time slice and let it run for
+	 * another round on the same CPU.
+         */
+	if (prev && is_queued(prev)) {
+		prev->scx.slice = SCX_SLICE_DFL;
+
+		/*
+		 * Notify the user-space scheduler if there are any pending
+		 * tasks to be completed, before resuming the previous
+		 * task.
+		 *
+		 * Keep in mind that if we don't refill the previous task's
+		 * time slice, this check will be performed in
+		 * ops.update_idle().
+		 */
+		if (usersched_has_pending_tasks())
+			set_usersched_needed();
 	}
 }
 
+void BPF_STRUCT_OPS(rustland_runnable, struct task_struct *p, u64 enq_flags)
+{
+	struct task_ctx *tctx;
+
+	if (is_usersched_task(p))
+		return;
+
+	tctx = try_lookup_task_ctx(p);
+	if (!tctx)
+		return;
+
+	tctx->exec_runtime = 0;
+}
+
 /*
  * Task @p starts on its selected CPU (update CPU ownership map).
  */
 void BPF_STRUCT_OPS(rustland_running, struct task_struct *p)
 {
 	s32 cpu = scx_bpf_task_cpu(p);
+	struct task_ctx *tctx;
+
+	if (is_usersched_task(p)) {
+		usersched_last_run_at = scx_bpf_now();
+		return;
+	}
 
 	dbg_msg("start: pid=%d (%s) cpu=%ld", p->pid, p->comm, cpu);
 
-	/*
-	 * Ensure time slice never exceeds slice_ns when a task is started on a
-	 * CPU.
-	 */
-	p->scx.slice = task_slice(p);
-
 	/*
 	 * Mark the CPU as busy by setting the pid as owner (ignoring the
 	 * user-space scheduler).
 	 */
-	if (!is_usersched_task(p))
-		__sync_fetch_and_add(&nr_running, 1);
+	__sync_fetch_and_add(&nr_running, 1);
+
+	tctx = try_lookup_task_ctx(p);
+	if (!tctx)
+		return;
+	tctx->last_run_at = scx_bpf_now();
 }
 
 /*
@@ -954,13 +944,23 @@ void BPF_STRUCT_OPS(rustland_running, struct task_struct *p)
 void BPF_STRUCT_OPS(rustland_stopping, struct task_struct *p, bool runnable)
 {
 	s32 cpu = scx_bpf_task_cpu(p);
+	struct task_ctx *tctx;
+
+	if (is_usersched_task(p))
+		return;
 
 	dbg_msg("stop: pid=%d (%s) cpu=%ld", p->pid, p->comm, cpu);
+
+	__sync_fetch_and_sub(&nr_running, 1);
+
+	tctx = try_lookup_task_ctx(p);
+	if (!tctx)
+		return;
+
 	/*
-	 * Mark the CPU as idle by setting the owner to 0.
+	 * Update the partial execution time since last sleep.
 	 */
-	if (!is_usersched_task(p))
-		__sync_fetch_and_sub(&nr_running, 1);
+	tctx->exec_runtime += scx_bpf_now() - tctx->last_run_at;
 }
 
 /*
@@ -988,15 +988,6 @@ void BPF_STRUCT_OPS(rustland_update_idle, s32 cpu, bool idle)
 		scx_bpf_kick_cpu(cpu, 0);
 		return;
 	}
-
-	/*
-	 * Kick the CPU if there are still tasks dispatched to the
-	 * corresponding per-CPU DSQ.
-	 */
-	if (scx_bpf_dsq_nr_queued(cpu_to_dsq(cpu)) > 0) {
-		scx_bpf_kick_cpu(cpu, 0);
-		return;
-	}
 }
 
 /*
@@ -1046,7 +1037,6 @@ s32 BPF_STRUCT_OPS(rustland_init_task, struct task_struct *p,
 				    BPF_LOCAL_STORAGE_GET_F_CREATE);
 	if (!tctx)
 		return -ENOMEM;
-	tctx->slice_ns = SCX_SLICE_DFL;
 
 	/*
 	 * Create task's L2 cache cpumask.
@@ -1082,10 +1072,27 @@ s32 BPF_STRUCT_OPS(rustland_init_task, struct task_struct *p,
  */
 static int usersched_timer_fn(void *map, int *key, struct bpf_timer *timer)
 {
+	struct task_struct *p;
 	int err = 0;
 
-	/* Kick the scheduler */
-	set_usersched_needed();
+	/*
+	 * Trigger the user-space scheduler if it has been inactive for
+	 * more than USERSCHED_TIMER_NS.
+	 */
+	if (time_delta(scx_bpf_now(), usersched_last_run_at) >= USERSCHED_TIMER_NS) {
+		bpf_rcu_read_lock();
+		p = bpf_task_from_pid(usersched_pid);
+		if (p) {
+			s32 cpu;
+
+			set_usersched_needed();
+			cpu = scx_bpf_pick_idle_cpu(p->cpus_ptr, 0);
+			if (cpu >= 0)
+				scx_bpf_kick_cpu(cpu, SCX_KICK_IDLE);
+			bpf_task_release(p);
+		}
+		bpf_rcu_read_unlock();
+	}
 
 	/* Re-arm the timer */
 	err = bpf_timer_start(timer, USERSCHED_TIMER_NS, 0);
@@ -1272,6 +1279,7 @@ SCX_OPS_DEFINE(rustland,
 	       .select_cpu		= (void *)rustland_select_cpu,
 	       .enqueue			= (void *)rustland_enqueue,
 	       .dispatch		= (void *)rustland_dispatch,
+	       .runnable		= (void *)rustland_runnable,
 	       .running			= (void *)rustland_running,
 	       .stopping		= (void *)rustland_stopping,
 	       .update_idle		= (void *)rustland_update_idle,
@@ -1280,7 +1288,7 @@ SCX_OPS_DEFINE(rustland,
 	       .init_task		= (void *)rustland_init_task,
 	       .init			= (void *)rustland_init,
 	       .exit			= (void *)rustland_exit,
-	       .flags			= SCX_OPS_ENQ_LAST | SCX_OPS_KEEP_BUILTIN_IDLE,
+	       .flags			= SCX_OPS_KEEP_BUILTIN_IDLE | SCX_OPS_ENQ_LAST,
 	       .timeout_ms		= 5000,
 	       .dispatch_max_batch	= MAX_DISPATCH_SLOT,
 	       .name			= "rustland");
