@@ -26,6 +26,7 @@
 // This software may be used and distributed according to the terms of the
 // GNU General Public License version 2.
 
+use seccomp::*;
 use std::alloc::{GlobalAlloc, Layout};
 use std::fs::File;
 use std::io::{BufRead, BufReader, Write};
@@ -210,94 +211,96 @@ impl BuddyAlloc {
     /// and must guarantee no others write to the memory range, to avoid undefined behaviors.
     /// The new function panic if memory space not enough for initialize BuddyAlloc.
     pub unsafe fn new(param: BuddyAllocParam) -> Self {
-        let BuddyAllocParam {
-            base_addr,
-            len,
-            leaf_size,
-            zero_filled,
-        } = param;
-        let mut base_addr = base_addr as usize;
-        let end_addr = base_addr + len;
-        assert!(
-            leaf_size % MIN_LEAF_SIZE_ALIGN == 0 && leaf_size != 0,
-            "{}",
-            LEAF_ALIGN_ERROR_MSG
-        );
-        let leaf2base = log2(leaf_size);
-        base_addr = roundup(base_addr, leaf2base);
-        // we use (k + 1)-th entry's split flag to test existence of k-th entry's blocks;
-        // to accoding this convention, we make a dummy (entries_size - 1)-th entry.
-        // so we plus 2 on entries_size.
-        let entries_size = log2((end_addr - base_addr) >> leaf2base) + 2;
+        unsafe {
+            let BuddyAllocParam {
+                base_addr,
+                len,
+                leaf_size,
+                zero_filled,
+            } = param;
+            let mut base_addr = base_addr as usize;
+            let end_addr = base_addr + len;
+            assert!(
+                leaf_size % MIN_LEAF_SIZE_ALIGN == 0 && leaf_size != 0,
+                "{}",
+                LEAF_ALIGN_ERROR_MSG
+            );
+            let leaf2base = log2(leaf_size);
+            base_addr = roundup(base_addr, leaf2base);
+            // we use (k + 1)-th entry's split flag to test existence of k-th entry's blocks;
+            // to accoding this convention, we make a dummy (entries_size - 1)-th entry.
+            // so we plus 2 on entries_size.
+            let entries_size = log2((end_addr - base_addr) >> leaf2base) + 2;
 
-        // alloc buddy allocator memory
-        let used_bytes = core::mem::size_of::<Entry>() * entries_size;
-        debug_assert!(end_addr >= base_addr + used_bytes, "{}", OOM_MSG);
-        let entries = base_addr as *mut Entry;
-        base_addr += used_bytes;
-
-        let buddy_list_size = core::mem::size_of::<Node>();
-        // init entries free
-        for k in 0..entries_size {
-            // use one bit for per memory block
-            debug_assert!(end_addr >= base_addr + buddy_list_size, "{}", OOM_MSG);
-            let entry = entries.add(k).as_mut().expect("entry");
-            entry.free = base_addr as *mut Node;
-            if !zero_filled {
-                core::ptr::write_bytes(entry.free, 0, buddy_list_size);
-            }
-            Node::init(entry.free);
-            base_addr += buddy_list_size;
-        }
-
-        // init alloc
-        for k in 0..entries_size {
-            // use one bit for per memory block
-            // use shift instead `/`, 8 == 1 << 3
-            let used_bytes = roundup(nblock(k, entries_size), 3) >> 3;
+            // alloc buddy allocator memory
+            let used_bytes = core::mem::size_of::<Entry>() * entries_size;
             debug_assert!(end_addr >= base_addr + used_bytes, "{}", OOM_MSG);
-            let entry = entries.add(k).as_mut().expect("entry");
-            entry.alloc = base_addr as *mut u8;
-            // mark all blocks as allocated
-            if !zero_filled {
-                core::ptr::write_bytes(entry.alloc, 0, used_bytes);
-            }
+            let entries = base_addr as *mut Entry;
             base_addr += used_bytes;
-        }
 
-        // init split
-        for k in 1..entries_size {
-            // use one bit for per memory block
-            // use shift instead `/`, 8 == 1 << 3
-            let used_bytes = roundup(nblock(k, entries_size), 3) >> 3;
-            debug_assert!(end_addr >= base_addr + used_bytes, "{}", OOM_MSG);
-            let entry = entries.add(k).as_mut().expect("entry");
-            entry.split = base_addr as *mut u8;
-            if !zero_filled {
-                core::ptr::write_bytes(entry.split, 0, used_bytes);
+            let buddy_list_size = core::mem::size_of::<Node>();
+            // init entries free
+            for k in 0..entries_size {
+                // use one bit for per memory block
+                debug_assert!(end_addr >= base_addr + buddy_list_size, "{}", OOM_MSG);
+                let entry = entries.add(k).as_mut().expect("entry");
+                entry.free = base_addr as *mut Node;
+                if !zero_filled {
+                    core::ptr::write_bytes(entry.free, 0, buddy_list_size);
+                }
+                Node::init(entry.free);
+                base_addr += buddy_list_size;
             }
-            base_addr += used_bytes;
+
+            // init alloc
+            for k in 0..entries_size {
+                // use one bit for per memory block
+                // use shift instead `/`, 8 == 1 << 3
+                let used_bytes = roundup(nblock(k, entries_size), 3) >> 3;
+                debug_assert!(end_addr >= base_addr + used_bytes, "{}", OOM_MSG);
+                let entry = entries.add(k).as_mut().expect("entry");
+                entry.alloc = base_addr as *mut u8;
+                // mark all blocks as allocated
+                if !zero_filled {
+                    core::ptr::write_bytes(entry.alloc, 0, used_bytes);
+                }
+                base_addr += used_bytes;
+            }
+
+            // init split
+            for k in 1..entries_size {
+                // use one bit for per memory block
+                // use shift instead `/`, 8 == 1 << 3
+                let used_bytes = roundup(nblock(k, entries_size), 3) >> 3;
+                debug_assert!(end_addr >= base_addr + used_bytes, "{}", OOM_MSG);
+                let entry = entries.add(k).as_mut().expect("entry");
+                entry.split = base_addr as *mut u8;
+                if !zero_filled {
+                    core::ptr::write_bytes(entry.split, 0, used_bytes);
+                }
+                base_addr += used_bytes;
+            }
+
+            // align base_addr to leaf size
+            base_addr = roundup(base_addr, leaf2base);
+            assert!(end_addr >= base_addr, "{}", OOM_MSG);
+            debug_assert_eq!(
+                (base_addr >> leaf2base) << leaf2base,
+                base_addr,
+                "misalignment"
+            );
+
+            let mut allocator = BuddyAlloc {
+                base_addr,
+                end_addr,
+                entries,
+                entries_size,
+                leaf2base,
+                unavailable: 0,
+            };
+            allocator.init_free_list();
+            allocator
         }
-
-        // align base_addr to leaf size
-        base_addr = roundup(base_addr, leaf2base);
-        assert!(end_addr >= base_addr, "{}", OOM_MSG);
-        debug_assert_eq!(
-            (base_addr >> leaf2base) << leaf2base,
-            base_addr,
-            "misalignment"
-        );
-
-        let mut allocator = BuddyAlloc {
-            base_addr,
-            end_addr,
-            entries,
-            entries_size,
-            leaf2base,
-            unavailable: 0,
-        };
-        allocator.init_free_list();
-        allocator
     }
 
     fn init_free_list(&mut self) {
@@ -447,9 +450,29 @@ impl UserAllocator {
     }
 
     unsafe fn fetch_buddy_alloc<R, F: FnOnce(&mut BuddyAlloc) -> R>(&self, f: F) -> R {
-        let mut inner = self.inner_buddy_alloc.lock().unwrap();
-        let alloc = inner.get_or_insert_with(|| BuddyAlloc::new(self.buddy_alloc_param));
-        f(alloc)
+        unsafe {
+            let mut inner = self.inner_buddy_alloc.lock().unwrap();
+            let alloc = inner.get_or_insert_with(|| BuddyAlloc::new(self.buddy_alloc_param));
+            f(alloc)
+        }
+    }
+
+    // Enable a seccomp filter that sends a SIGSYS when mmap() is called.
+    #[allow(static_mut_refs)]
+    pub fn disable_mmap(&self) -> Result<(), Box<dyn std::error::Error>> {
+        let mut ctx = seccomp::Context::default(Action::Allow)?;
+        let syscall_nr = libc::SYS_mmap as usize;
+        let cmp = Compare::arg(0)
+            .using(Op::MaskedEq)
+            .with(0)
+            .build()
+            .ok_or("Failed to build seccomp filter")?;
+
+        let rule = Rule::new(syscall_nr, cmp, Action::Errno(libc::EPERM));
+        ctx.add_rule(rule)?;
+        ctx.load()?;
+
+        Ok(())
     }
 
     #[allow(static_mut_refs)]
@@ -492,12 +515,16 @@ impl UserAllocator {
 // Override global allocator methods.
 unsafe impl GlobalAlloc for UserAllocator {
     unsafe fn alloc(&self, layout: Layout) -> *mut u8 {
-        let bytes = layout.size();
-        self.fetch_buddy_alloc(|alloc| alloc.malloc(bytes))
+        unsafe {
+            let bytes = layout.size();
+            self.fetch_buddy_alloc(|alloc| alloc.malloc(bytes))
+        }
     }
 
     unsafe fn dealloc(&self, ptr: *mut u8, _layout: Layout) {
-        self.fetch_buddy_alloc(|alloc| alloc.free(ptr));
+        unsafe {
+            self.fetch_buddy_alloc(|alloc| alloc.free(ptr));
+        }
     }
 }
 
