@@ -79,7 +79,17 @@ pub fn set_rlimit_infinity() {
     };
 }
 
-pub fn read_file_usize(path: &Path) -> Result<usize> {
+/// Read a file and parse its content into the specified type.
+///
+/// Trims whitespace before parsing.
+///
+/// # Errors
+/// Returns an error if reading or parsing fails.
+pub fn read_from_file<T>(path: &Path) -> Result<T>
+where
+    T: std::str::FromStr,
+    T::Err: std::error::Error + Send + Sync + 'static,
+{
     let val = match std::fs::read_to_string(path) {
         Ok(val) => val,
         Err(_) => {
@@ -87,10 +97,10 @@ pub fn read_file_usize(path: &Path) -> Result<usize> {
         }
     };
 
-    match val.trim().parse::<usize>() {
+    match val.trim().parse::<T>() {
         Ok(parsed) => Ok(parsed),
         Err(_) => {
-            bail!("Failed to parse {}", val);
+            bail!("Failed to parse content '{}' from {:?}", val.trim(), path);
         }
     }
 }
@@ -107,6 +117,28 @@ pub fn read_file_usize_vec(path: &Path, separator: char) -> Result<Vec<usize>> {
         .collect::<Result<Vec<usize>>>()
 }
 
+pub fn read_file_byte(path: &Path) -> Result<usize> {
+    let val = std::fs::read_to_string(path)?;
+    let val = val.trim();
+
+    // E.g., 10K, 10M, 10G, 10
+    if let Some(sval) = val.strip_suffix("K") {
+        let byte = sval.parse::<usize>()?;
+        return Ok(byte * 1024);
+    }
+    if let Some(sval) = val.strip_suffix("M") {
+        let byte = sval.parse::<usize>()?;
+        return Ok(byte * 1024 * 1024);
+    }
+    if let Some(sval) = val.strip_suffix("G") {
+        let byte = sval.parse::<usize>()?;
+        return Ok(byte * 1024 * 1024 * 1024);
+    }
+
+    let byte = val.parse::<usize>()?;
+    Ok(byte)
+}
+
 /* Load is reported as weight * duty cycle
  *
  * In the Linux kernel, EEDVF uses default weight = 1 s.t.
