@@ -69,8 +69,10 @@
 //! hierarchy are entirely read-only. If the host topology were to change (due
 //! to e.g. hotplug), a new Topology object should be created.
 
-use crate::misc::read_file_usize;
+use crate::cpumask::read_cpulist;
+use crate::misc::read_file_byte;
 use crate::misc::read_file_usize_vec;
+use crate::misc::read_from_file;
 use crate::Cpumask;
 use anyhow::bail;
 use anyhow::Result;
@@ -116,13 +118,16 @@ pub struct Cpu {
     /// Base operational frqeuency. Only available on Intel Turbo Boost
     /// CPUs. If not available, this will simply return maximum frequency.
     pub base_freq: usize,
-    /// The best-effort guessting of cpu_capacity scaled to 1024
+    /// The best-effort guessing of cpu_capacity scaled to 1024.
     pub cpu_capacity: usize,
+    pub smt_level: usize,
     /// CPU idle resume latency
     pub pm_qos_resume_latency_us: usize,
     pub trans_lat_ns: usize,
     pub l2_id: usize,
     pub l3_id: usize,
+    /// Per-CPU cache size of all levels.
+    pub cache_size: usize,
     pub core_type: CoreType,
 
     /// Ancestor IDs.
@@ -130,7 +135,7 @@ pub struct Cpu {
     pub llc_id: usize,
     pub node_id: usize,
     pub package_id: usize,
-    pub cluster_id: usize,
+    pub cluster_id: isize,
 }
 
 #[derive(Clone, Debug, Eq, Ord, PartialEq, PartialOrd)]
@@ -139,7 +144,7 @@ pub struct Core {
     pub id: usize,
     /// The sysfs value of core_id
     pub kernel_id: usize,
-    pub cluster_id: usize,
+    pub cluster_id: isize,
     pub cpus: BTreeMap<usize, Arc<Cpu>>,
     /// Cpumask of all CPUs in this core.
     pub span: Cpumask,
@@ -215,7 +220,13 @@ impl Topology {
                 let mut llc_cpus = BTreeMap::new();
 
                 for (&core_id, core) in llc_mut.cores.iter_mut() {
-                    for (&cpu_id, cpu) in core.cpus.iter() {
+                    let core_mut = Arc::get_mut(core).unwrap();
+                    let smt_level = core_mut.cpus.len();
+
+                    for (&cpu_id, cpu) in core_mut.cpus.iter_mut() {
+                        let cpu_mut = Arc::get_mut(cpu).unwrap();
+                        cpu_mut.smt_level = smt_level;
+
                         if topo_cpus
                             .insert(cpu_id, cpu.clone())
                             .or(node_cpus.insert(cpu_id, cpu.clone()))
@@ -353,24 +364,7 @@ impl TopoCtx {
 fn cpus_online() -> Result<Cpumask> {
     let path = "/sys/devices/system/cpu/online";
     let online = std::fs::read_to_string(path)?;
-    let online_groups: Vec<&str> = online.split(',').collect();
-    let mut mask = Cpumask::new();
-    for group in online_groups.iter() {
-        let (min, max) = match sscanf!(group.trim(), "{usize}-{usize}") {
-            Ok((x, y)) => (x, y),
-            Err(_) => match sscanf!(group.trim(), "{usize}") {
-                Ok(x) => (x, x),
-                Err(_) => {
-                    bail!("Failed to parse online cpus {}", group.trim());
-                }
-            },
-        };
-        for i in min..(max + 1) {
-            mask.set_cpu(i)?;
-        }
-    }
-
-    Ok(mask)
+    Cpumask::from_cpulist(&online)
 }
 
 fn get_cache_id(topo_ctx: &mut TopoCtx, cache_level_path: &PathBuf, cache_level: usize) -> usize {
@@ -393,7 +387,7 @@ fn get_cache_id(topo_ctx: &mut TopoCtx, cache_level_path: &PathBuf, cache_level:
     }
 
     // In case of a cache miss, try to get the id from the sysfs first.
-    let id = read_file_usize(&cache_level_path.join("id")).unwrap_or(usize::MAX);
+    let id = read_from_file(&cache_level_path.join("id")).unwrap_or(usize::MAX);
     if id != usize::MAX {
         // Keep the id in the map
         id_map.insert(key, id);
@@ -407,6 +401,26 @@ fn get_cache_id(topo_ctx: &mut TopoCtx, cache_level_path: &PathBuf, cache_level:
     id
 }
 
+fn get_per_cpu_cache_size(cache_path: &PathBuf) -> Result<usize> {
+    let path_str = cache_path.to_str().unwrap();
+    let paths = glob(&(path_str.to_owned() + "/index[0-9]*"))?;
+    let mut tot_size = 0;
+
+    for index in paths.filter_map(Result::ok) {
+        // If there is no size information under sysfs (e.g., many ARM SoCs),
+        // give 1024 as a default value. 1024 is small enough compared to the
+        // real cache size of the CPU, but it is large enough to give a penalty
+        // when multiple CPUs share the cache.
+        let size = read_file_byte(&index.join("size")).unwrap_or(1024_usize);
+        let cpulist: String = read_from_file(&index.join("shared_cpu_list"))?;
+        let num_cpus = read_cpulist(&cpulist)?.len();
+        tot_size += size / num_cpus;
+    }
+
+    Ok(tot_size)
+}
+
+#[allow(clippy::too_many_arguments)]
 fn create_insert_cpu(
     id: usize,
     node: &mut Node,
@@ -429,9 +443,9 @@ fn create_insert_cpu(
 
     // Physical core ID
     let top_path = cpu_path.join("topology");
-    let core_kernel_id = read_file_usize(&top_path.join("core_id"))?;
-    let package_id = read_file_usize(&top_path.join("physical_package_id"))?;
-    let cluster_id = read_file_usize(&top_path.join("cluster_id"))?;
+    let core_kernel_id = read_from_file(&top_path.join("core_id"))?;
+    let package_id = read_from_file(&top_path.join("physical_package_id"))?;
+    let cluster_id = read_from_file(&top_path.join("cluster_id"))?;
 
     // Evaluate L2, L3 and LLC cache IDs.
     //
@@ -449,24 +463,28 @@ fn create_insert_cpu(
         l3_id
     };
 
+    // Per-CPU cache size
+    let cache_size = get_per_cpu_cache_size(&cache_path).unwrap_or(0_usize);
+
     // Min and max frequencies. If the kernel is not compiled with
     // CONFIG_CPU_FREQ, just assume 0 for both frequencies.
     let freq_path = cpu_path.join("cpufreq");
-    let min_freq = read_file_usize(&freq_path.join("scaling_min_freq")).unwrap_or(0);
-    let max_freq = read_file_usize(&freq_path.join("scaling_max_freq")).unwrap_or(0);
-    let base_freq = read_file_usize(&freq_path.join("base_frequency")).unwrap_or(max_freq);
-    let trans_lat_ns = read_file_usize(&freq_path.join("cpuinfo_transition_latency")).unwrap_or(0);
+    let min_freq = read_from_file(&freq_path.join("scaling_min_freq")).unwrap_or(0_usize);
+    let max_freq = read_from_file(&freq_path.join("scaling_max_freq")).unwrap_or(0_usize);
+    let base_freq = read_from_file(&freq_path.join("base_frequency")).unwrap_or(max_freq);
+    let trans_lat_ns =
+        read_from_file(&freq_path.join("cpuinfo_transition_latency")).unwrap_or(0_usize);
 
     // Cpu capacity
     let (cap_suffix, _avg_rcap, max_rcap) = capacity_src.unwrap_or(("".to_string(), 1024, 1024));
     let cap_path = cpu_path.join(cap_suffix);
-    let rcap = read_file_usize(&cap_path).unwrap_or(max_rcap);
+    let rcap = read_from_file(&cap_path).unwrap_or(max_rcap);
     let cpu_capacity = (rcap * 1024) / max_rcap;
 
     // Power management
     let power_path = cpu_path.join("power");
     let pm_qos_resume_latency_us =
-        read_file_usize(&power_path.join("pm_qos_resume_latency_us")).unwrap_or(0);
+        read_from_file(&power_path.join("pm_qos_resume_latency_us")).unwrap_or(0_usize);
 
     let num_llcs = topo_ctx.node_llc_kernel_ids.len();
     let llc_id = topo_ctx
@@ -529,10 +547,12 @@ fn create_insert_cpu(
             max_freq,
             base_freq,
             cpu_capacity,
+            smt_level: 0, // Will be initialized at instantiate().
             pm_qos_resume_latency_us,
             trans_lat_ns,
             l2_id,
             l3_id,
+            cache_size,
             core_type: core_type.clone(),
 
             core_id: *core_id,
@@ -588,7 +608,7 @@ fn cpu_capacity_source() -> Option<(String, usize, usize)> {
     for src in sources {
         let path_str = [prefix, src].join("/");
         let path = Path::new(&path_str);
-        raw_capacity = read_file_usize(&path).unwrap_or(0);
+        raw_capacity = read_from_file(&path).unwrap_or(0_usize);
         if raw_capacity > 0 {
             suffix = src;
             break;
@@ -604,7 +624,7 @@ fn cpu_capacity_source() -> Option<(String, usize, usize)> {
     let mut nr_cpus = 0;
     let cpu_paths = glob("/sys/devices/system/cpu/cpu[0-9]*").ok()?;
     for cpu_path in cpu_paths.filter_map(Result::ok) {
-        let raw_capacity = read_file_usize(&cpu_path.join(suffix)).unwrap_or(0);
+        let raw_capacity = read_from_file(&cpu_path.join(suffix)).unwrap_or(0_usize);
         if max_raw_capacity < raw_capacity {
             max_raw_capacity = raw_capacity;
         }
@@ -630,8 +650,8 @@ fn avg_cpu_freq() -> Option<(usize, usize)> {
     let cpu_paths = glob("/sys/devices/system/cpu/cpu[0-9]*").ok()?;
     for cpu_path in cpu_paths.filter_map(Result::ok) {
         let freq_path = cpu_path.join("cpufreq");
-        let max_freq = read_file_usize(&freq_path.join("scaling_max_freq")).unwrap_or(0);
-        let base_freq = read_file_usize(&freq_path.join("base_frequency")).unwrap_or(max_freq);
+        let max_freq = read_from_file(&freq_path.join("scaling_max_freq")).unwrap_or(0_usize);
+        let base_freq = read_from_file(&freq_path.join("base_frequency")).unwrap_or(max_freq);
         if base_freq > 0 {
             if max_freq > top_max_freq {
                 top_max_freq = max_freq;
@@ -652,7 +672,7 @@ fn has_big_little() -> Option<bool> {
     let cpu_paths = glob("/sys/devices/system/cpu/cpu[0-9]*").ok()?;
     for cpu_path in cpu_paths.filter_map(Result::ok) {
         let top_path = cpu_path.join("topology");
-        let cluster_id = read_file_usize(&top_path.join("cluster_id")).unwrap_or(0);
+        let cluster_id = read_from_file(&top_path.join("cluster_id")).unwrap_or(-1);
         clusters.insert(cluster_id);
     }
 
@@ -660,7 +680,7 @@ fn has_big_little() -> Option<bool> {
 }
 
 fn is_smt_active() -> Option<bool> {
-    let smt_on = read_file_usize(Path::new("/sys/devices/system/cpu/smt/active")).ok()?;
+    let smt_on: u8 = read_from_file(Path::new("/sys/devices/system/cpu/smt/active")).ok()?;
     Some(smt_on == 1)
 }
 
@@ -770,6 +790,7 @@ fn create_numa_nodes(
         let big_little = has_big_little().unwrap_or(false);
         let capacity_src = cpu_capacity_source();
         let avg_cpu_freq = avg_cpu_freq();
+        let mut cpu_ids = vec![];
         for cpu_path in cpu_paths.filter_map(Result::ok) {
             let cpu_str = cpu_path.to_str().unwrap().trim();
             let cpu_id = match sscanf!(cpu_str, "/sys/devices/system/node/node{usize}/cpu{usize}") {
@@ -778,7 +799,11 @@ fn create_numa_nodes(
                     bail!("Failed to parse cpu ID {}", cpu_str);
                 }
             };
+            cpu_ids.push(cpu_id);
+        }
+        cpu_ids.sort();
 
+        for cpu_id in cpu_ids {
             create_insert_cpu(
                 cpu_id,
                 &mut node,
