@@ -38,11 +38,15 @@ const volatile u64 slice_min = 1ULL * NSEC_PER_MSEC;
 const volatile s64 slice_lag = 20ULL * NSEC_PER_MSEC;
 
 /*
- * If enabled, never allow tasks to preempt others before their assigned
- * time slice expires.
+ * If enabled, never allow tasks to be directly dispatched.
  */
 const volatile bool no_preempt;
 
+/*
+ * Ignore synchronous wakeup events.
+ */
+const volatile bool no_wake_sync;
+
 /*
  * When enabled always dispatch per-CPU kthreads directly.
  *
@@ -245,6 +249,12 @@ struct task_ctx {
 	 * for latency-sensitive tasks.
 	 */
 	u64 deadline;
+
+	/*
+	 * Task's recently used CPU: used to determine whether we need to
+	 * refresh the task's cpumasks.
+	 */
+	s32 recent_used_cpu;
 };
 
 /* Map that contains task-local storage. */
@@ -323,6 +333,28 @@ static u64 nr_tasks_waiting(int node)
 	return scx_bpf_dsq_nr_queued(node) + 1;
 }
 
+/*
+ * Scale a value inversely proportional to the task's normalized weight.
+ */
+static inline u64 scale_by_task_normalized_weight_inverse(const struct task_struct *p, u64 value)
+{
+	/*
+	 * Original weight range:   [1, 10000], default = 100
+	 * Normalized weight range: [1, 128], default = 64
+	 *
+	 * This normalization reduces the impact of extreme weight differences,
+	 * preventing highly prioritized tasks from starving lower-priority ones.
+	 *
+	 * The goal is to ensure a more balanced scheduling that is
+	 * influenced more by the task's behavior rather than its priority
+	 * difference and prevent potential stalls due to large priority
+	 * gaps.
+	*/
+	u64 weight = 1 + (127 * log2_u64(p->scx.weight) / log2_u64(10000));
+
+	return value * 64 / weight;
+}
+
 /*
  * Update and return the task's deadline.
  */
@@ -331,21 +363,30 @@ static u64 task_deadline(const struct task_struct *p, struct task_ctx *tctx)
 	u64 vtime_min;
 
 	/*
-	 * Limit the amount of vtime budget that an idling task can
-	 * accumulate to prevent excessive prioritization of sleeping
-	 * tasks.
+	 * Cap the vruntime budget that an idle task can accumulate to
+	 * slice_lag, preventing sleeping tasks from gaining excessive
+	 * priority.
 	 *
-	 * Tasks with a higher weight get a bigger "bucket" for their
-	 * allowed accumulated time budget.
+	 * A larger slice_lag favors tasks that sleep longer by allowing
+	 * them to accumulate more credit, leading to shorter deadlines and
+	 * earlier execution. A smaller slice_lag reduces the advantage of
+	 * long sleeps, treating short and long sleeps equally once they
+	 * exceed the threshold.
+	 *
+	 * If slice_lag is negative, it can be used to de-emphasize the
+	 * deadline-based scheduling altogether by charging all tasks a
+	 * fixed vruntime penalty (equal to the absolute value of
+	 * slice_lag), effectively approximating FIFO behavior as the
+	 * penalty increases.
 	 */
-	vtime_min = vtime_now - slice_max;
+	vtime_min = vtime_now - slice_lag;
 	if (time_before(tctx->deadline, vtime_min))
 		tctx->deadline = vtime_min;
 
 	/*
 	 * Add the execution vruntime to the deadline.
 	 */
-	tctx->deadline += scale_by_task_weight_inverse(p, tctx->exec_runtime);
+	tctx->deadline += scale_by_task_normalized_weight_inverse(p, tctx->exec_runtime);
 
 	return tctx->deadline;
 }
@@ -357,6 +398,12 @@ static void task_update_domain(struct task_struct *p, struct task_ctx *tctx,
 	struct bpf_cpumask *p_mask, *l2_mask, *l3_mask;
 	struct cpu_ctx *cctx;
 
+	/*
+	 * Refresh task's recently used CPU every time the task's domain
+	 * is updated..
+	 */
+	tctx->recent_used_cpu = cpu;
+
 	cctx = try_lookup_cpu_ctx(cpu);
 	if (!cctx)
 		return;
@@ -365,13 +412,8 @@ static void task_update_domain(struct task_struct *p, struct task_ctx *tctx,
 	if (!primary)
 		return;
 
-	l3_domain = cctx->l3_cpumask;
-	if (!l3_domain)
-		l3_domain = primary;
-
 	l2_domain = cctx->l2_cpumask;
-	if (!l2_domain)
-		l2_domain = l3_domain;
+	l3_domain = cctx->l3_cpumask;
 
 	p_mask = tctx->cpumask;
 	if (!p_mask) {
@@ -402,21 +444,54 @@ static void task_update_domain(struct task_struct *p, struct task_ctx *tctx,
 	 * primary cpumask and the L2 cache domain mask of the previously used
 	 * CPU.
 	 */
-	bpf_cpumask_and(l2_mask, cast_mask(p_mask), cast_mask(l2_domain));
+	if (l2_domain)
+		bpf_cpumask_and(l2_mask, cast_mask(p_mask), cast_mask(l2_domain));
 
 	/*
 	 * Determine the L3 cache domain as the intersection of the task's
 	 * primary cpumask and the L3 cache domain mask of the previously used
 	 * CPU.
 	 */
-	bpf_cpumask_and(l3_mask, cast_mask(p_mask), cast_mask(l3_domain));
+	if (l3_domain)
+		bpf_cpumask_and(l3_mask, cast_mask(p_mask), cast_mask(l3_domain));
 }
 
-static bool is_wake_sync(const struct task_struct *p,
-			 const struct task_struct *current,
-			 s32 prev_cpu, s32 cpu, u64 wake_flags)
+/*
+ * Return true if all the CPUs in the LLC of @cpu are busy, false
+ * otherwise.
+ */
+static bool is_llc_busy(const struct cpumask *idle_cpumask, s32 cpu)
 {
-	if (wake_flags & SCX_WAKE_SYNC)
+	const struct cpumask *primary, *l3_mask;
+	struct cpu_ctx *cctx;
+
+	primary = cast_mask(primary_cpumask);
+	if (!primary)
+		return false;
+
+	cctx = try_lookup_cpu_ctx(cpu);
+	if (!cctx)
+		return false;
+
+	l3_mask = cast_mask(cctx->l3_cpumask);
+	if (!l3_mask)
+		l3_mask = primary;
+
+	return !bpf_cpumask_intersects(l3_mask, idle_cpumask);
+}
+
+/*
+ * Return true if the waker commits to release the CPU after waking up @p,
+ * false otherwise.
+ */
+static bool is_wake_sync(s32 prev_cpu, s32 this_cpu, u64 wake_flags)
+{
+	const struct task_struct *current = (void *)bpf_get_current_task_btf();
+
+	if (no_wake_sync)
+		return false;
+
+	if ((wake_flags & SCX_WAKE_SYNC) && !(current->flags & PF_EXITING))
 		return true;
 
 	/*
@@ -425,15 +500,43 @@ static bool is_wake_sync(const struct task_struct *p,
 	 *
 	 * The assumption is that the wakee had queued work for the per-CPU
 	 * kthread, which has now finished, making the wakeup effectively
-	 * synchronous. An example of this behavior is seen in IO completions.
+	 * synchronous. An example of this behavior is seen in IO
+	 * completions.
 	 */
 	if (is_kthread(current) && (current->nr_cpus_allowed == 1) &&
-	    (prev_cpu == cpu))
+	    (prev_cpu == this_cpu))
 		return true;
 
 	return false;
 }
 
+/*
+ * Return the target CPU for @p in case of a sync wakeup.
+ *
+ * During a sync wakeup, the waker commits to releasing the CPU immediately
+ * after the wakeup event, so we should consider a sync wakeup almost like
+ * a direct function call between a waker and a wakee.
+ */
+static s32 try_sync_wakeup(const struct task_struct *p, s32 prev_cpu, s32 this_cpu)
+{
+	/*
+	 * If @prev_cpu is idle, keep using it, since there is no guarantee
+	 * that the cache hot data from the waker's CPU is more important
+	 * than cache hot data in the wakee's CPU.
+	 */
+	if ((this_cpu != prev_cpu) && scx_bpf_test_and_clear_cpu_idle(prev_cpu))
+		return prev_cpu;
+
+	/*
+	 * If waker and wakee are on the same CPU and no other tasks are
+	 * queued, consider the waker's CPU as idle.
+	 */
+	if (!scx_bpf_dsq_nr_queued(SCX_DSQ_LOCAL_ON | this_cpu))
+		return this_cpu;
+
+	return -EBUSY;
+}
+
 /*
  * Find an idle CPU in the system.
  *
@@ -447,10 +550,10 @@ static s32 pick_idle_cpu(struct task_struct *p, s32 prev_cpu, u64 wake_flags, bo
 {
 	const struct cpumask *idle_smtmask, *idle_cpumask;
 	const struct cpumask *primary, *p_mask, *l2_mask, *l3_mask;
-	struct task_struct *current = (void *)bpf_get_current_task_btf();
 	struct task_ctx *tctx;
 	int node;
-	s32 cpu;
+	s32 this_cpu = bpf_get_smp_processor_id(), cpu;
+	bool share_llc;
 
 	primary = cast_mask(primary_cpumask);
 	if (!primary)
@@ -472,9 +575,12 @@ static s32 pick_idle_cpu(struct task_struct *p, s32 prev_cpu, u64 wake_flags, bo
 	}
 
 	/*
-	 * Refresh task domain based on the previously used cpu.
+	 * Refresh task domain based on the previously used cpu. If we keep
+	 * selecting the same CPU, the task's domain doesn't need to be
+	 * updated and we can save some cpumask ops.
 	 */
-	task_update_domain(p, tctx, prev_cpu, p->cpus_ptr);
+	if (tctx->recent_used_cpu != prev_cpu)
+		task_update_domain(p, tctx, prev_cpu, p->cpus_ptr);
 
 	p_mask = cast_mask(tctx->cpumask);
 	if (p_mask && bpf_cpumask_empty(p_mask))
@@ -483,8 +589,8 @@ static s32 pick_idle_cpu(struct task_struct *p, s32 prev_cpu, u64 wake_flags, bo
 	if (l2_mask && bpf_cpumask_empty(l2_mask))
 		l2_mask = NULL;
 	l3_mask = cast_mask(tctx->l3_cpumask);
-	if (l3_mask)
-		l2_mask = NULL;
+	if (l3_mask && bpf_cpumask_empty(l3_mask))
+		l3_mask = NULL;
 
 	/*
 	 * Acquire the CPU masks to determine the idle CPUs in the system.
@@ -494,50 +600,20 @@ static s32 pick_idle_cpu(struct task_struct *p, s32 prev_cpu, u64 wake_flags, bo
 	idle_cpumask = __COMPAT_scx_bpf_get_idle_cpumask_node(node);
 
 	/*
-	 * If the current task is waking up another task and releasing the CPU
-	 * (WAKE_SYNC), attempt to migrate the wakee on the same CPU as the
-	 * waker.
+	 * In case of a sync wakeup, attempt to run the wakee on the
+	 * waker's CPU if possible, as it's going to release the CPU right
+	 * after the wakeup, so it can be considered as idle and, possibly,
+	 * cache hot.
+	 *
+	 * However, ignore this optimization if the LLC is completely
+	 * saturated, since it's just more efficient to dispatch the task
+	 * on the first CPU available.
 	 */
-	cpu = bpf_get_smp_processor_id();
-	if (is_wake_sync(p, current, cpu, prev_cpu, wake_flags)) {
-		const struct cpumask *curr_l3_domain;
-		struct cpu_ctx *cctx;
-		bool share_llc;
-
-		/*
-		 * Determine waker CPU scheduling domain.
-		 */
-		cctx = try_lookup_cpu_ctx(cpu);
-		if (!cctx) {
-			cpu = -EINVAL;
-			goto out_put_cpumask;
-		}
-
-		curr_l3_domain = cast_mask(cctx->l3_cpumask);
-		if (!curr_l3_domain)
-			curr_l3_domain = primary;
-
-		/*
-		 * If both the waker and wakee share the same L3 cache keep
-		 * using the same CPU if possible.
-		 */
-		share_llc = bpf_cpumask_test_cpu(prev_cpu, curr_l3_domain);
-		if (share_llc &&
-		    scx_bpf_test_and_clear_cpu_idle(prev_cpu)) {
-			cpu = prev_cpu;
-			*is_idle = true;
-			goto out_put_cpumask;
-		}
-
-		/*
-		 * If the waker's L3 domain is not saturated attempt to migrate
-		 * the wakee on the same CPU as the waker (since it's going to
-		 * block and release the current CPU).
-		 */
-		if (!(current->flags & PF_EXITING) &&
-		    bpf_cpumask_intersects(curr_l3_domain, idle_cpumask) &&
-		    p_mask && bpf_cpumask_test_cpu(cpu, p_mask) &&
-		    scx_bpf_dsq_nr_queued(SCX_DSQ_LOCAL_ON | cpu) == 0) {
+	share_llc = l3_mask && bpf_cpumask_test_cpu(this_cpu, l3_mask);
+	if (is_wake_sync(prev_cpu, this_cpu, wake_flags) &&
+	    share_llc && !is_llc_busy(idle_cpumask, this_cpu)) {
+		cpu = try_sync_wakeup(p, prev_cpu, this_cpu);
+		if (cpu >= 0) {
 			*is_idle = true;
 			goto out_put_cpumask;
 		}
@@ -674,6 +750,25 @@ out_put_cpumask:
 	return cpu;
 }
 
+/*
+ * Return true if we can perform a direct dispatch on @node, false
+ * otherwise.
+ */
+static bool can_direct_dispatch(int node)
+{
+	/*
+	 * Never allow direct dispatch if preemption is disabled.
+	 */
+	if (no_preempt)
+		return false;
+
+	/*
+	 * Allow direct dispatch when @local_pcpu is enabled, or when there
+	 * are no tasks queued in the node DSQ.
+	 */
+	return local_pcpu || !scx_bpf_dsq_nr_queued(node);
+}
+
 /*
  * Pick a target CPU for a task which is being woken up.
  *
@@ -690,7 +785,7 @@ s32 BPF_STRUCT_OPS(bpfland_select_cpu, struct task_struct *p,
 	if (is_idle) {
 		int node = __COMPAT_scx_bpf_cpu_node(cpu);
 
-		if (local_pcpu || !scx_bpf_dsq_nr_queued(node)) {
+		if (can_direct_dispatch(node)) {
 			scx_bpf_dsq_insert(p, SCX_DSQ_LOCAL, slice_max, 0);
 			__sync_fetch_and_add(&nr_direct_dispatches, 1);
 		}
@@ -767,24 +862,9 @@ static bool try_direct_dispatch(struct task_struct *p, struct task_ctx *tctx,
 	 * taken by a higher priority scheduling class, force it to follow
 	 * the regular scheduling path and give it a chance to run on a
 	 * different CPU.
-	 *
-	 * However, if the task can only run on a single CPU, re-scheduling
-	 * is unnecessary, as it can only be dispatched on that specific
-	 * CPU. In this case, dispatch it immediately to maximize its
-	 * chances of reclaiming the CPU quickly and avoiding stalls.
-	 *
-	 * This approach will be effective once dl_server support is added
-	 * to the sched_ext core.
 	 */
-	if (enq_flags & SCX_ENQ_REENQ) {
-		if (p->nr_cpus_allowed == 1) {
-			scx_bpf_dsq_insert(p, SCX_DSQ_LOCAL, slice_max, enq_flags);
-			__sync_fetch_and_add(&nr_kthread_dispatches, 1);
-
-			return true;
-		}
+	if (enq_flags & SCX_ENQ_REENQ)
 		return false;
-	}
 
 	/*
 	 * If local_kthread is specified dispatch per-CPU kthreads
@@ -802,19 +882,13 @@ static bool try_direct_dispatch(struct task_struct *p, struct task_ctx *tctx,
 	 */
 	if (!__COMPAT_is_enq_cpu_selected(enq_flags)) {
 		int node = __COMPAT_scx_bpf_cpu_node(prev_cpu);
-		struct rq *rq = scx_bpf_cpu_rq(prev_cpu);
 
 		/*
-		 * Allow to preempt the task currently running on the
-		 * assigned CPU if our deadline is earlier.
+		 * Stop here if direct dispatch is not allowed in the
+		 * target node.
 		 */
-		if (!no_preempt && tctx->deadline < rq->curr->scx.dsq_vtime) {
-			scx_bpf_dsq_insert(p, SCX_DSQ_LOCAL_ON | prev_cpu,
-					   slice, enq_flags | SCX_ENQ_PREEMPT);
-			__sync_fetch_and_add(&nr_direct_dispatches, 1);
-
-			return true;
-		}
+		if (!can_direct_dispatch(node))
+			return false;
 
 		/*
 		 * If local_pcpu is enabled always dispatch tasks that can
@@ -830,18 +904,11 @@ static bool try_direct_dispatch(struct task_struct *p, struct task_ctx *tctx,
 			return true;
 		}
 
-		/*
-		 * Skip direct dispatch if the shared DSQ has tasks waiting
-		 * to prevent starvation.
-		 */
-		if (!local_pcpu && scx_bpf_dsq_nr_queued(node))
-			return false;
-
 		/*
 		 * If the task can only run on a single CPU and that CPU is
 		 * idle, perform a direct dispatch.
 		 */
-		if (p->nr_cpus_allowed == 1) {
+		if (p->nr_cpus_allowed == 1 || is_migration_disabled(p)) {
 			if (scx_bpf_test_and_clear_cpu_idle(prev_cpu)) {
 				scx_bpf_dsq_insert(p, SCX_DSQ_LOCAL,
 						   slice_max, enq_flags);
@@ -1105,7 +1172,7 @@ void BPF_STRUCT_OPS(bpfland_stopping, struct task_struct *p, bool runnable)
 	/*
 	 * Evaluate the time slice used by the task.
 	 */
-	slice = scx_bpf_now() - tctx->last_run_at;
+	slice = now - tctx->last_run_at;
 
 	/*
 	 * Update task's execution time (exec_runtime), but never account
@@ -1119,7 +1186,7 @@ void BPF_STRUCT_OPS(bpfland_stopping, struct task_struct *p, bool runnable)
 	/*
 	 * Update task's vruntime.
 	 */
-	tctx->deadline += scale_by_task_weight_inverse(p, slice);
+	tctx->deadline += scale_by_task_normalized_weight_inverse(p, slice);
 
 	/*
 	 * Update CPU runtime.
