@@ -145,8 +145,8 @@ struct Opts {
 
     /// Disable preemption.
     ///
-    /// Never allow tasks to preempt others before their assigned time slice expires. This can help
-    /// to increase system throughput over responsiveness.
+    /// Never allow tasks to be directly dispatched. This can help to increase fairness
+    /// over responsiveness.
     #[clap(short = 'n', long, action = clap::ArgAction::SetTrue)]
     no_preempt: bool,
 
@@ -170,6 +170,15 @@ struct Opts {
     #[clap(short = 'k', long, action = clap::ArgAction::SetTrue)]
     local_kthreads: bool,
 
+    /// Disable direct dispatch during synchronous wakeups.
+    ///
+    /// Enabling this option can lead to a more uniform load distribution across available cores,
+    /// potentially improving performance in certain scenarios. However, it may come at the cost of
+    /// reduced efficiency for pipe-intensive workloads that benefit from tighter producer-consumer
+    /// coupling.
+    #[clap(short = 'w', long, action = clap::ArgAction::SetTrue)]
+    no_wake_sync: bool,
+
     /// Specifies the initial set of CPUs, represented as a bitmask in hex (e.g., 0xff), that the
     /// scheduler will use to dispatch tasks, until the system becomes saturated, at which point
     /// tasks may overflow to other available CPUs.
@@ -296,6 +305,7 @@ impl<'a> Scheduler<'a> {
         skel.maps.rodata_data.local_pcpu = opts.local_pcpu;
         skel.maps.rodata_data.local_kthreads = opts.local_kthreads;
         skel.maps.rodata_data.no_preempt = opts.no_preempt;
+        skel.maps.rodata_data.no_wake_sync = opts.no_wake_sync;
         skel.maps.rodata_data.slice_max = opts.slice_us * 1000;
         skel.maps.rodata_data.slice_min = opts.slice_us_min * 1000;
         skel.maps.rodata_data.slice_lag = opts.slice_us_lag * 1000;
