@@ -12,6 +12,17 @@ typedef unsigned long long u64;
 enum chaos_consts {
 	CHAOS_DSQ_BASE_SHIFT	= 16,
 	CHAOS_DSQ_BASE		= 1 << CHAOS_DSQ_BASE_SHIFT,
+
+	CHAOS_NUM_PPIDS_CHECK	= 1 << 20,
+};
+
+enum chaos_match {
+	CHAOS_MATCH_UNKNOWN		= 0,
+	CHAOS_MATCH_COMPLETE		= 1 << 0,
+	CHAOS_MATCH_EXCLUDED		= 1 << 1,
+	CHAOS_MATCH_HAS_PARENT		= 1 << 2,
+
+	CHAOS_MATCH_MAX			= 1 << 3,
 };
 
 enum chaos_trait_kind {
@@ -22,9 +33,11 @@ enum chaos_trait_kind {
 
 struct chaos_task_ctx {
 	// chaos_task_ctx is initialised zero'd
+	enum chaos_match	match;
 
 	enum chaos_trait_kind	next_trait;
 	u64			enq_flags;
+	u64			p2dq_vtime;
 };
 
 #endif /* __CHAOS_INTF_H */
