@@ -50,7 +50,7 @@ extern void bpf_iter_task_destroy(struct bpf_iter_task *it) __weak __ksym;
 enum {
 	LAVD_CPU_ID_MAX			= 512,
 
-	LAVD_CPDOM_MAX_NR		= 32, /* maximum number of compute domain */
+	LAVD_CPDOM_MAX_NR		= 16, /* maximum number of compute domain */
 	LAVD_CPDOM_MAX_DIST		= 4,  /* maximum distance from one compute domain to another */
 
 	LAVD_STATUS_STR_LEN		= 4, /* {LR: Latency-critical, Regular}
@@ -63,32 +63,35 @@ enum {
  * System-wide stats
  */
 struct sys_stat {
-	volatile u64	last_update_clk;
-	volatile u64	util;		/* average of the CPU utilization */
+	u64	last_update_clk;
+	u64	avg_util;	/* average of the CPU utilization */
+	u64	avg_sc_util;	/* average of the scaled CPU utilization,
+				   which is capacity and frequency invariant */
 
-	volatile u64	avg_svc_time;	/* average service time per task */
-	volatile u64	nr_queued_task;
+	u64	avg_svc_time;	/* average service time per task */
+	u64	nr_queued_task;
+	u64	slice;		/* base time slice */
 
-	volatile u32	avg_lat_cri;	/* average latency criticality (LC) */
-	volatile u32	max_lat_cri;	/* maximum latency criticality (LC) */
-	volatile u32	thr_lat_cri;	/* latency criticality threshold for kicking */
+	u32	avg_lat_cri;	/* average latency criticality (LC) */
+	u32	max_lat_cri;	/* maximum latency criticality (LC) */
+	u32	thr_lat_cri;	/* latency criticality threshold for kicking */
 
-	volatile u32	min_perf_cri;	/* minimum performance criticality */
-	volatile u32	avg_perf_cri;	/* average performance criticality */
-	volatile u32	max_perf_cri;	/* maximum performance criticality */
-	volatile u32	thr_perf_cri;	/* performance criticality threshold */
+	u32	min_perf_cri;	/* minimum performance criticality */
+	u32	avg_perf_cri;	/* average performance criticality */
+	u32	max_perf_cri;	/* maximum performance criticality */
+	u32	thr_perf_cri;	/* performance criticality threshold */
 
-	volatile u32	nr_stealee;	/* number of compute domains to be migrated */
-	volatile u32	nr_violation;	/* number of utilization violation */
-	volatile u32	nr_active;	/* number of active cores */
+	u32	nr_stealee;	/* number of compute domains to be migrated */
+	u32	nr_active;	/* number of active cores */
 
-	volatile u64	nr_sched;	/* total scheduling so far */
-	volatile u64	nr_perf_cri;	/* number of performance-critical tasks scheduled */
-	volatile u64	nr_lat_cri;	/* number of latency-critical tasks scheduled */
-	volatile u64	nr_x_migration; /* number of cross domain migration */
-	volatile u64	nr_big;		/* scheduled on big core */
-	volatile u64	nr_pc_on_big;	/* performance-critical tasks scheduled on big core */
-	volatile u64	nr_lc_on_big;	/* latency-critical tasks scheduled on big core */
+	u64	nr_sched;	/* total scheduling so far */
+	u64	nr_preempt;	/* total number of preemption operations triggered */
+	u64	nr_perf_cri;	/* number of performance-critical tasks scheduled */
+	u64	nr_lat_cri;	/* number of latency-critical tasks scheduled */
+	u64	nr_x_migration; /* number of cross domain migration */
+	u64	nr_big;		/* scheduled on big core */
+	u64	nr_pc_on_big;	/* performance-critical tasks scheduled on big core */
+	u64	nr_lc_on_big;	/* latency-critical tasks scheduled on big core */
 };
 
 /*
@@ -106,12 +109,12 @@ struct task_ctx {
 	/*
 	 * Task running statistics for latency criticality calculation
 	 */
-	u64	acc_run_time_ns;	/* accmulated runtime from runnable to quiescent state */
-	u64	run_time_ns;		/* average runtime per schedule */
+	u64	acc_runtime;		/* accmulated runtime from runnable to quiescent state */
+	u64	avg_runtime;		/* average runtime per schedule */
 	u64	run_freq;		/* scheduling frequency in a second */
 	u64	wait_freq;		/* waiting frequency in a second */
 	u64	wake_freq;		/* waking-up frequency in a second */
-	u64	svc_time;		/* total CPU time consumed for this task */
+	u64	svc_time;		/* total CPU time consumed for this task scaled by task's weight */
 	u64	dsq_id;			/* DSQ id where a task run for statistics */
 
 	/*
@@ -126,7 +129,6 @@ struct task_ctx {
 	u8	need_lock_boost;	/* need to boost lock for deadline calculation */
 	u8	lock_holder_xted;	/* slice is already extended for a lock holder task */
 	u8	wakeup_ft;		/* regular wakeup = 1, sync wakeup = 2 */
-	u8	slice_boost_prio;	/* how many times a task fully consumed the slice */
 	u8	on_big;			/* executable on a big core */
 	u8	on_little;		/* executable on a little core */
 	u8	is_affinitized;		/* is this task pinned to a subset of all CPUs? */
@@ -142,6 +144,7 @@ struct task_ctx_x {
 	u16	static_prio;	/* nice priority */
 	u32	cpu_id;		/* where a task ran */
 	u64	cpu_util;	/* cpu utilization in [0..100] */
+	u64	cpu_sutil;	/* scaled cpu utilization in [0..100] */
 	u32	thr_perf_cri;	/* performance criticality threshold */
 	u32	avg_lat_cri;	/* average latency criticality */
 	u32	nr_active;	/* number of active cores */
