@@ -191,18 +191,6 @@
 
 char _license[] SEC("license") = "GPL";
 
-/*
- * Include sub-modules
- */
-#include "util.bpf.c"
-#include "power.bpf.c"
-#include "introspec.bpf.c"
-#include "sys_stat.bpf.c"
-#include "preempt.bpf.c"
-#include "lock.bpf.c"
-#include "idle.bpf.c"
-#include "balance.bpf.c"
-
 /*
  * Logical current clock
  */
@@ -220,26 +208,30 @@ static u64		cur_svc_time;
 const volatile u64	slice_min_ns = LAVD_SLICE_MIN_NS_DFL;
 const volatile u64	slice_max_ns = LAVD_SLICE_MAX_NS_DFL;
 
+/*
+ * Include sub-modules
+ */
+#include "util.bpf.c"
+#include "power.bpf.c"
+#include "introspec.bpf.c"
+#include "sys_stat.bpf.c"
+#include "preempt.bpf.c"
+#include "lock.bpf.c"
+#include "idle.bpf.c"
+#include "balance.bpf.c"
+
 static u32 calc_greedy_ratio(struct task_ctx *taskc)
 {
-	struct sys_stat *stat_cur = get_sys_stat_cur();
 	u32 ratio;
 
-	if (!have_scheduled(taskc)) {
-		ratio = LAVD_GREEDY_RATIO_NEW;
-		goto out;
-	}
-
 	/*
 	 * The greedy ratio of a task represents how much time the task
 	 * overspent CPU time compared to the ideal, fair CPU allocation. It is
 	 * the ratio of task's actual service time to average service time in a
 	 * system.
 	 */
-	ratio = (1000 * taskc->svc_time) / stat_cur->avg_svc_time;
-
-out:
-	taskc->is_greedy = ratio > 1000;
+	ratio = (taskc->svc_time << LAVD_SHIFT) / sys_stat.avg_svc_time;
+	taskc->is_greedy = ratio > LAVD_SCALE;
 	return ratio;
 }
 
@@ -248,26 +240,24 @@ static u32 calc_greedy_factor(u32 greedy_ratio)
 	/*
 	 * For all under-utilized tasks, we treat them equally.
 	 */
-	if (greedy_ratio <= 1000)
-		return 1000;
+	if (greedy_ratio <= LAVD_SCALE)
+		return LAVD_SCALE;
 
 	/*
 	 * For over-utilized tasks, we give some mild penalty.
 	 */
-	return 1000 + ((greedy_ratio - 1000) / LAVD_LC_GREEDY_PENALTY);
+	return LAVD_SCALE + ((greedy_ratio - LAVD_SCALE) / LAVD_LC_GREEDY_PENALTY);
 
 }
 
-static u64 calc_runtime_factor(u64 runtime, u64 weight_ft)
+static inline u64 calc_runtime_factor(u64 runtime)
 {
-	u64 ft = rsigmoid_u64(runtime, LAVD_LC_RUNTIME_MAX);
-	return (ft / weight_ft) + 1;
+	return rsigmoid_u64(runtime, LAVD_LC_RUNTIME_MAX);
 }
 
-static u64 calc_freq_factor(u64 freq, u64 weight_ft)
+static inline u64 calc_freq_factor(u64 freq)
 {
-	u64 ft = sigmoid_u64(freq, LAVD_LC_FREQ_MAX);
-	return (ft * weight_ft * LAVD_LC_FREQ_OVER_RUNTIME) + 1;
+	return sigmoid_u64(freq, LAVD_LC_FREQ_MAX);
 }
 
 static u64 calc_weight_factor(struct task_struct *p, struct task_ctx *taskc)
@@ -288,6 +278,12 @@ static u64 calc_weight_factor(struct task_struct *p, struct task_ctx *taskc)
 	if (is_kernel_task(p))
 		weight_boost += LAVD_LC_WEIGHT_BOOST;
 
+	/*
+	 * Further prioritize kworkers.
+	 */
+	if (is_kernel_worker(p))
+		weight_boost += LAVD_LC_WEIGHT_BOOST;
+
 	/*
 	 * Prioritize an affinitized task since it has restrictions
 	 * in placement so it tends to be delayed.
@@ -296,10 +292,10 @@ static u64 calc_weight_factor(struct task_struct *p, struct task_ctx *taskc)
 		weight_boost += LAVD_LC_WEIGHT_BOOST;
 
 	/*
-	 * Prioritize a migration-disabled task since it has restrictions
-	 * in placement so it tends to be delayed.
+	 * Prioritize a pinned task since it has restrictions in placement
+	 * so it tends to be delayed.
 	 */
-	if (is_per_cpu_task(p))
+	if (is_pinned(p) || is_migration_disabled(p))
 		weight_boost += LAVD_LC_WEIGHT_BOOST;
 
 	/*
@@ -319,7 +315,7 @@ static u64 calc_weight_factor(struct task_struct *p, struct task_ctx *taskc)
 
 static void calc_perf_cri(struct task_struct *p, struct task_ctx *taskc)
 {
-	u64 wait_freq_ft, wake_freq_ft, perf_cri = 1000;
+	u64 wait_freq_ft, wake_freq_ft, perf_cri = LAVD_SCALE;
 
 	/*
 	 * A task is more CPU-performance sensitive when it meets the following
@@ -336,11 +332,11 @@ static void calc_perf_cri(struct task_struct *p, struct task_ctx *taskc)
 	 * Note that we use unadjusted weight to reflect the pure task priority.
 	 */
 	if (have_little_core) {
-		wait_freq_ft = calc_freq_factor(taskc->wait_freq, p->scx.weight);
-		wake_freq_ft = calc_freq_factor(taskc->wake_freq, p->scx.weight);
+		wait_freq_ft = calc_freq_factor(taskc->wait_freq);
+		wake_freq_ft = calc_freq_factor(taskc->wake_freq);
 		perf_cri = log2_u64(wait_freq_ft * wake_freq_ft);
 		perf_cri += log2_u64(max(taskc->run_freq, 1) *
-				     max(taskc->run_time_ns, 1) * p->scx.weight);
+				     max(taskc->avg_runtime, 1) * p->scx.weight);
 	}
 
 	taskc->perf_cri = perf_cri;
@@ -367,9 +363,9 @@ static void calc_lat_cri(struct task_struct *p, struct task_ctx *taskc)
 	 * is monotonically increasing since higher frequencies mean more
 	 * latency-critical.
 	 */
-	wait_freq_ft = calc_freq_factor(taskc->wait_freq, weight_ft);
-	wake_freq_ft = calc_freq_factor(taskc->wake_freq, weight_ft);
-	runtime_ft = calc_runtime_factor(taskc->run_time_ns, weight_ft);
+	wait_freq_ft = calc_freq_factor(taskc->wait_freq) + 1;
+	wake_freq_ft = calc_freq_factor(taskc->wake_freq) + 1;
+	runtime_ft = calc_runtime_factor(taskc->avg_runtime) + 1;
 
 	/*
 	 * Wake frequency and wait frequency represent how much a task is used
@@ -379,9 +375,8 @@ static void calc_lat_cri(struct task_struct *p, struct task_ctx *taskc)
 	 * add +1 to guarantee the latency criticality (log2-ed) is always
 	 * positive.
 	 */
-	lat_cri = log2_u64(wait_freq_ft);
-	lat_cri += log2_u64(wake_freq_ft);
-	lat_cri += log2_u64(runtime_ft);
+	lat_cri = log2_u64(wait_freq_ft * wake_freq_ft) +
+		  log2_u64(runtime_ft * weight_ft);
 
 	/*
 	 * Determine latency criticality of a task in a context-aware manner by
@@ -391,8 +386,18 @@ static void calc_lat_cri(struct task_struct *p, struct task_ctx *taskc)
 	taskc->lat_cri = max(lat_cri, taskc->lat_cri_waker);
 }
 
-static u64 calc_adj_runtime(u64 runtime)
+static u64 calc_adjusted_runtime(struct task_ctx *taskc)
 {
+	u64 runtime;
+
+	/*
+	 * Prefer a short-running (avg_runtime) and recently woken-up
+	 * (acc_runtime) task. To avoid the starvation of CPU-bound tasks,
+	 * which rarely sleep, limit the impact of acc_runtime.
+	 */
+	runtime = taskc->avg_runtime +
+		  min(taskc->acc_runtime, LAVD_ACC_RUNTIME_MAX);
+
 	/*
 	 * Convert highly skewed runtime distribution to
 	 * mildly skewed distribution.
@@ -402,9 +407,9 @@ static u64 calc_adj_runtime(u64 runtime)
 }
 
 static u64 calc_virtual_deadline_delta(struct task_struct *p,
-					struct task_ctx *taskc)
+				       struct task_ctx *taskc)
 {
-	u64 deadline, adj_runtime;
+	u64 deadline, adjusted_runtime;
 	u32 greedy_ratio, greedy_ft;
 
 	/*
@@ -415,61 +420,20 @@ static u64 calc_virtual_deadline_delta(struct task_struct *p,
 	calc_lat_cri(p, taskc);
 	greedy_ratio = calc_greedy_ratio(taskc);
 	greedy_ft = calc_greedy_factor(greedy_ratio);
-	adj_runtime = calc_adj_runtime(taskc->run_time_ns);
+	adjusted_runtime = calc_adjusted_runtime(taskc);
 
-	deadline = (adj_runtime / taskc->lat_cri) * greedy_ft;
+	deadline = (adjusted_runtime * greedy_ft) / taskc->lat_cri;
 
 	return deadline;
 }
 
-static u32 clamp_time_slice_ns(u32 slice)
+static u64 calc_time_slice(struct task_ctx *taskc)
 {
-	if (slice < slice_min_ns)
-		slice = slice_min_ns;
-	else if (slice > slice_max_ns)
-		slice = slice_max_ns;
-	return slice;
-}
-
-static u64 calc_time_slice(struct task_struct *p, struct task_ctx *taskc)
-{
-	struct sys_stat *stat_cur = get_sys_stat_cur();
-	u64 nr_queued;
-	u32 slice;
-
 	if (!taskc)
 		return LAVD_SLICE_MAX_NS_DFL;
 
-	/*
-	 * The time slice should be short enough to schedule all runnable tasks
-	 * at least once within a targeted latency.
-	 */
-	nr_queued = stat_cur->nr_queued_task + 1;
-	slice = (LAVD_TARGETED_LATENCY_NS * stat_cur->nr_active) / nr_queued;
-
-	/*
-	 * Keep the slice in [slice_min_ns, slice_max_ns].
-	 */
-	slice = clamp_time_slice_ns(slice);
-
-	/*
-	 * Boost time slice for CPU-bound tasks.
-	 */
-	slice += (LAVD_SLICE_BOOST_MAX_FT * slice *
-		  taskc->slice_boost_prio) / LAVD_SLICE_BOOST_MAX_STEP;
-
-	/*
-	 * If a task has yet to be scheduled (i.e., a freshly forked task or a
-	 * task just under sched_ext), don't give a fair amount of time slice
-	 * until knowing its properties. This helps to mitigate potential
-	 * system starvation caused by massively forking tasks (i.e., fork-bomb
-	 * attacks).
-	 */
-	if (!have_scheduled(taskc))
-		slice >>= 2;
-
-	taskc->slice_ns = slice;
-	return slice;
+	taskc->slice_ns = sys_stat.slice;
+	return taskc->slice_ns;
 }
 
 static void reset_suspended_duration(struct cpu_ctx *cpuc)
@@ -499,18 +463,8 @@ static u64 get_suspended_duration_and_reset(struct cpu_ctx *cpuc)
 	return duration;
 }
 
-static void update_stat_for_runnable(struct task_struct *p,
-				     struct task_ctx *taskc)
-{
-	/*
-	 * Reflect task's load immediately.
-	 */
-	taskc->acc_run_time_ns = 0;
-}
-
 static void advance_cur_logical_clk(struct task_struct *p)
 {
-	struct sys_stat *stat_cur = get_sys_stat_cur();
 	u64 vlc, clc, ret_clc;
 	u64 nr_queued, delta, new_clk;
 	int i;
@@ -529,7 +483,7 @@ static void advance_cur_logical_clk(struct task_struct *p)
 		 * Advance the clock up to the task's deadline. When overloaded,
 		 * advance the clock slower so other can jump in the run queue.
 		 */
-		nr_queued = max(stat_cur->nr_queued_task, 1);
+		nr_queued = max(sys_stat.nr_queued_task, 1);
 		delta = (vlc - clc) / nr_queued;
 		new_clk = clc + delta;
 
@@ -548,7 +502,6 @@ static void update_stat_for_running(struct task_struct *p,
 				    struct task_ctx *taskc,
 				    struct cpu_ctx *cpuc, u64 now)
 {
-	struct sys_stat *stat_cur = get_sys_stat_cur();
 	u64 wait_period, interval;
 
 	/*
@@ -562,7 +515,7 @@ static void update_stat_for_running(struct task_struct *p,
 	 */
 	if (p->scx.slice == SCX_SLICE_DFL) {
 		p->scx.dsq_vtime = READ_ONCE(cur_logical_clk);
-		p->scx.slice = calc_time_slice(p, taskc);
+		p->scx.slice = calc_time_slice(taskc);
 	}
 
 	/*
@@ -576,7 +529,7 @@ static void update_stat_for_running(struct task_struct *p,
 	 */
 	if (have_scheduled(taskc)) {
 		wait_period = time_delta(now, taskc->last_quiescent_clk);
-		interval = taskc->run_time_ns + wait_period;
+		interval = taskc->avg_runtime + wait_period;
 		taskc->run_freq = calc_avg_freq(taskc->run_freq, interval);
 	}
 
@@ -616,10 +569,10 @@ static void update_stat_for_running(struct task_struct *p,
 	/*
 	 * Update statistics information.
 	 */
-	if (is_lat_cri(taskc, stat_cur))
+	if (is_lat_cri(taskc))
 		cpuc->nr_lat_cri++;
 
-	if (is_perf_cri(taskc, stat_cur))
+	if (is_perf_cri(taskc))
 		cpuc->nr_perf_cri++;
 
 	if (taskc->dsq_id != cpuc->cpdom_id) {
@@ -639,10 +592,10 @@ static void update_stat_for_stopping(struct task_struct *p,
 				     struct cpu_ctx *cpuc)
 {
 	u64 now = scx_bpf_now();
-	u64 suspended_duration, task_run_time;
+	u64 suspended_duration, task_runtime;
 
 	/*
-	 * Update task's run_time. When a task is scheduled consecutively
+	 * Update task's runtime. When a task is scheduled consecutively
 	 * without ops.quiescent(), the task's runtime is accumulated for
 	 * statistics. Suppose a task is scheduled 2ms, 2ms, and 2ms with the
 	 * time slice exhausted. If 6ms of time slice was given in the first
@@ -651,26 +604,12 @@ static void update_stat_for_stopping(struct task_struct *p,
 	 * calculation of runtime statistics.
 	 */
 	suspended_duration = get_suspended_duration_and_reset(cpuc);
-	task_run_time = time_delta(now, taskc->last_running_clk + suspended_duration);
-	taskc->acc_run_time_ns += task_run_time;
-	taskc->run_time_ns = calc_avg(taskc->run_time_ns, taskc->acc_run_time_ns);
+	task_runtime = time_delta(now, taskc->last_running_clk + suspended_duration);
+	taskc->acc_runtime += task_runtime;
+	taskc->avg_runtime = calc_avg(taskc->avg_runtime, taskc->acc_runtime);
 	taskc->last_stopping_clk = now;
 
-	taskc->svc_time += task_run_time / p->scx.weight;
-
-	/*
-	 * Count how many times a task completely consumed the assigned time
-	 * slice to boost the task's slice. If not fully consumed, decrease
-	 * the slice boost priority by half.
-	 */
-	if (task_run_time > 0 && task_run_time >= taskc->slice_ns) {
-		if (taskc->slice_boost_prio < LAVD_SLICE_BOOST_MAX_STEP)
-			taskc->slice_boost_prio++;
-	}
-	else {
-		if (taskc->slice_boost_prio)
-			taskc->slice_boost_prio >>= 1;
-	}
+	taskc->svc_time += task_runtime / p->scx.weight;
 
 	/*
 	 * Reset waker's latency criticality here to limit the latency boost of
@@ -689,6 +628,12 @@ static void update_stat_for_stopping(struct task_struct *p,
 	if (READ_ONCE(cur_svc_time) < taskc->svc_time)
 		WRITE_ONCE(cur_svc_time, taskc->svc_time);
 
+	/*
+	 * Increase total scaled CPU time of this CPU,
+	 * whcih is capacity and frequency invariant.
+	 */
+	cpuc->tot_sc_time += scale_cap_freq(task_runtime, cpuc->cpu_id);
+
 	/*
 	 * Reset task's lock and futex boost count
 	 * for a lock holder to be boosted only once.
@@ -747,7 +692,7 @@ s32 BPF_STRUCT_OPS(lavd_select_cpu, struct task_struct *p, s32 prev_cpu,
 
 		if (!scx_bpf_dsq_nr_queued(dsq_id)) {
 			p->scx.dsq_vtime = calc_when_to_run(p, taskc);
-			p->scx.slice = calc_time_slice(p, taskc);
+			p->scx.slice = calc_time_slice(taskc);
 			scx_bpf_dsq_insert(p, SCX_DSQ_LOCAL, p->scx.slice, 0);
 			return cpu_id;
 		}
@@ -772,10 +717,9 @@ static bool can_direct_dispatch(u64 dsq_id, s32 cpu, bool is_idle)
 
 void BPF_STRUCT_OPS(lavd_enqueue, struct task_struct *p, u64 enq_flags)
 {
-	struct cpu_ctx *cpuc_cur;
 	struct task_ctx *taskc;
 	s32 task_cpu, cpu = -ENOENT;
-	u64 dsq_id, now;
+	u64 dsq_id;
 	bool is_idle = false;
 
 	taskc = get_task_ctx(p);
@@ -787,7 +731,7 @@ void BPF_STRUCT_OPS(lavd_enqueue, struct task_struct *p, u64 enq_flags)
 	 */
 	taskc->wakeup_ft += !!(enq_flags & SCX_ENQ_WAKEUP);
 	p->scx.dsq_vtime = calc_when_to_run(p, taskc);
-	p->scx.slice = calc_time_slice(p, taskc);
+	p->scx.slice = calc_time_slice(taskc);
 
 	/*
 	 * Find a proper DSQ for the task, which is either the task's
@@ -824,44 +768,19 @@ void BPF_STRUCT_OPS(lavd_enqueue, struct task_struct *p, u64 enq_flags)
 	 * Try to find and kick a victim CPU, which runs a less urgent task,
 	 * from dsq_id. The kick will be done asynchronously.
 	 */
-	cpuc_cur = get_cpu_ctx();
-	if (cpuc_cur) {
-		now = scx_bpf_now();
-		try_find_and_kick_victim_cpu(p, taskc, cpuc_cur, dsq_id, now);
-	}
-}
-
-static bool try_continue_lock_holder(struct task_struct *prev,
-				     struct task_ctx *taskc,
-				     struct cpu_ctx *cpuc)
-{
-	if (!(prev->scx.flags & SCX_TASK_QUEUED) || !is_lock_holder(taskc))
-		return false;
-
-	/*
-	 * Refill the time slice.
-	 */
-	prev->scx.slice = calc_time_slice(prev, taskc);
-
-	/*
-	 * Reset prev task's lock and futex boost count
-	 * for a lock holder to be boosted only once.
-	 */
-	reset_lock_futex_boost(taskc, cpuc);
-	taskc->lock_holder_xted = true;
-
-	return true;
+	if (!no_preemption)
+		try_find_and_kick_victim_cpu(p, taskc, dsq_id);
 }
 
 void BPF_STRUCT_OPS(lavd_dispatch, s32 cpu, struct task_struct *prev)
 {
 	struct cpu_ctx *cpuc;
-	struct task_ctx *taskc;
-	struct bpf_cpumask *active, *ovrflw, *cpdom_mask, *new_mask;
+	struct task_ctx *taskc_prev = NULL;
+	struct bpf_cpumask *active, *ovrflw;
 	struct task_struct *p;
 	u64 dsq_id;
 	s32 new_cpu;
-	bool try_consume = false;
+	bool try_consume;
 
 	cpuc = get_cpu_ctx_id(cpu);
 	if (!cpuc) {
@@ -874,23 +793,32 @@ void BPF_STRUCT_OPS(lavd_dispatch, s32 cpu, struct task_struct *prev)
 	 * If a task is holding a new lock, continue to execute it
 	 * to make system-wide forward progress.
 	 */
-	if (prev) {
-		taskc = get_task_ctx(prev);
-		if (!taskc) {
+	if (prev && (prev->scx.flags & SCX_TASK_QUEUED) && cpuc->lock_holder) {
+		taskc_prev = get_task_ctx(prev);
+		if (!taskc_prev) {
 			scx_bpf_error("Failed to look up task context");
 			return;
 		}
-		if (try_continue_lock_holder(prev, taskc, cpuc))
-			return;
+
+		/*
+		 * Refill the time slice.
+		 */
+		prev->scx.slice = calc_time_slice(taskc_prev);
+
+		/*
+		 * Reset prev task's lock and futex boost count
+		 * for a lock holder to be boosted only once.
+		 */
+		reset_lock_futex_boost(taskc_prev, cpuc);
+		taskc_prev->lock_holder_xted = true;
+		return;
 	}
 
 	/*
 	 * If all CPUs are using, directly consume without checking CPU masks.
 	 */
-	if (use_full_cpus()) {
-		try_consume = true;
+	if (use_full_cpus())
 		goto consume_out;
-	}
 
 	/*
 	 * Prepare cpumasks.
@@ -899,10 +827,9 @@ void BPF_STRUCT_OPS(lavd_dispatch, s32 cpu, struct task_struct *prev)
 
 	active = active_cpumask;
 	ovrflw = ovrflw_cpumask;
-	cpdom_mask = MEMBER_VPTR(cpdom_cpumask, [cpuc->cpdom_id]);
-	new_mask = cpuc->tmp_t_mask;
-	if (!active || !ovrflw || !cpdom_mask || !new_mask) {
+	if (!active || !ovrflw) {
 		scx_bpf_error("Failed to prepare cpumasks.");
+		try_consume = false;
 		goto unlock_out;
 	}
 
@@ -912,39 +839,48 @@ void BPF_STRUCT_OPS(lavd_dispatch, s32 cpu, struct task_struct *prev)
 	 */
 	if (bpf_cpumask_test_cpu(cpu, cast_mask(active)) ||
 	    bpf_cpumask_test_cpu(cpu, cast_mask(ovrflw))) {
-		try_consume = true;
-		goto unlock_out;
+		bpf_rcu_read_unlock();
+		goto consume_out;
 	}
 	/* NOTE: This CPU belongs to neither active nor overflow set. */
 
-	/*
-	 * If the previous task is a per-CPU task on this CPU,
-	 * extend the overflow set and go.
-	 */
-	if (prev && is_per_cpu_task(prev)) {
-		bpf_cpumask_set_cpu(cpu, ovrflw);
-		try_consume = true;
-		goto unlock_out;
-	}
+	if (prev) {
+		/*
+		 * If the previous task is pinned to this CPU,
+		 * extend the overflow set and go.
+		 */
+		if (is_pinned(prev)) {
+			bpf_cpumask_set_cpu(cpu, ovrflw);
+			bpf_rcu_read_unlock();
+			goto consume_out;
+		} else if (is_migration_disabled(prev)) {
+			bpf_rcu_read_unlock();
+			goto consume_out;
+		}
 
-	/*
-	 * If the previous task can run on this CPU but not on either active
-	 * or overflow set, extend the overflow set and go.
-	 */
-	if (prev && taskc->is_affinitized &&
-	    bpf_cpumask_test_cpu(cpu, prev->cpus_ptr) &&
-	    !bpf_cpumask_intersects(cast_mask(active), prev->cpus_ptr) &&
-	    !bpf_cpumask_intersects(cast_mask(ovrflw), prev->cpus_ptr)) {
-		bpf_cpumask_set_cpu(cpu, ovrflw);
-		try_consume = true;
-		goto unlock_out;
+		/*
+		 * If the previous task can run on this CPU but not on either
+		 * active or overflow set, extend the overflow set and go.
+		 */
+		taskc_prev = get_task_ctx(prev);
+		if (taskc_prev && taskc_prev->is_affinitized &&
+		    bpf_cpumask_test_cpu(cpu, prev->cpus_ptr) &&
+		    !bpf_cpumask_intersects(cast_mask(active), prev->cpus_ptr) &&
+		    !bpf_cpumask_intersects(cast_mask(ovrflw), prev->cpus_ptr)) {
+			bpf_cpumask_set_cpu(cpu, ovrflw);
+			bpf_rcu_read_unlock();
+			goto consume_out;
+		}
 	}
 
 	/*
 	 * If this CPU is neither in active nor overflow CPUs,
 	 * try to find and run the task affinitized on this CPU.
 	 */
+	try_consume = false;
 	bpf_for_each(scx_dsq, p, dsq_id, 0) {
+		struct task_ctx *taskc;
+
 		/*
 		 * Note that this is a hack to bypass the restriction of the
 		 * current BPF not trusting the pointer p. Once the BPF
@@ -955,11 +891,11 @@ void BPF_STRUCT_OPS(lavd_dispatch, s32 cpu, struct task_struct *prev)
 			break;
 
 		/*
-		 * If the task is a per-CPU task on this CPU,
+		 * If the task is pinned to this CPU,
 		 * extend the overflow set and go.
 		 * But not on this CPU, try another task.
 		 */
-		if (is_per_cpu_task(p)) {
+		if (is_pinned(p)) {
 			new_cpu = scx_bpf_task_cpu(p);
 			if (new_cpu == cpu) {
 				bpf_cpumask_set_cpu(new_cpu, ovrflw);
@@ -967,9 +903,17 @@ void BPF_STRUCT_OPS(lavd_dispatch, s32 cpu, struct task_struct *prev)
 				try_consume = true;
 				break;
 			}
-			else if (!bpf_cpumask_test_and_set_cpu(new_cpu, ovrflw))
+			if (!bpf_cpumask_test_and_set_cpu(new_cpu, ovrflw))
 				scx_bpf_kick_cpu(new_cpu, SCX_KICK_IDLE);
-
+			bpf_task_release(p);
+			continue;
+		} else if (is_migration_disabled(p)) {
+			new_cpu = scx_bpf_task_cpu(p);
+			if (new_cpu == cpu) {
+				bpf_task_release(p);
+				try_consume = true;
+				break;
+			}
 			bpf_task_release(p);
 			continue;
 		}
@@ -1008,52 +952,26 @@ void BPF_STRUCT_OPS(lavd_dispatch, s32 cpu, struct task_struct *prev)
 unlock_out:
 	bpf_rcu_read_unlock();
 
-consume_out:
 	/*
 	 * If this CPU should go idle, do nothing.
 	 */
 	if (!try_consume)
 		return;
 
+consume_out:
 	/*
 	 * Otherwise, consume a task.
 	 */
-	if (consume_task(cpuc))
+	if (consume_task(dsq_id))
 		return;
 
 	/*
 	 * If nothing to run, continue running the previous task.
 	 */
-	if (prev && prev->scx.flags & SCX_TASK_QUEUED)
-		prev->scx.slice = calc_time_slice(prev, taskc);
-}
-
-void BPF_STRUCT_OPS(lavd_tick, struct task_struct *p_run)
-{
-	struct cpu_ctx *cpuc_run;
-	struct task_ctx *taskc_run;
-	u64 now;
-	bool preempted;
-
-	cpuc_run = get_cpu_ctx();
-	taskc_run = get_task_ctx(p_run);
-	if (!cpuc_run || !taskc_run)
-		goto update_cpuperf;
-
-	/*
-	 * Try to yield the current CPU if there is a higher priority task in
-	 * the run queue.
-	 */
-	now = scx_bpf_now();
-	preempted = try_yield_current_cpu(p_run, cpuc_run, taskc_run, now);
-
-	/*
-	 * Update performance target of the current CPU if the current running
-	 * task continues to run.
-	 */
-	if (!preempted) {
-update_cpuperf:
-		try_decrease_cpuperf_target(cpuc_run);
+	if (prev && prev->scx.flags & SCX_TASK_QUEUED) {
+		taskc_prev = taskc_prev ?: get_task_ctx(prev);
+		if (taskc_prev)
+			prev->scx.slice = calc_time_slice(taskc_prev);
 	}
 }
 
@@ -1064,15 +982,12 @@ void BPF_STRUCT_OPS(lavd_runnable, struct task_struct *p, u64 enq_flags)
 	u64 now, interval;
 
 	/*
-	 * Add task load based on the current statistics regardless of a target
-	 * rq. Statistics will be adjusted when more accurate statistics become
-	 * available (ops.running).
+	 * Clear the accumulated runtime.
 	 */
 	p_taskc = get_task_ctx(p);
 	if (!p_taskc)
 		return;
-
-	update_stat_for_runnable(p, p_taskc);
+	p_taskc->acc_runtime = 0;
 
 	/*
 	 * When a task @p is wakened up, the wake frequency of its waker task
@@ -1112,7 +1027,6 @@ void BPF_STRUCT_OPS(lavd_runnable, struct task_struct *p, u64 enq_flags)
 
 void BPF_STRUCT_OPS(lavd_running, struct task_struct *p)
 {
-	struct sys_stat *stat_cur = get_sys_stat_cur();
 	struct cpu_ctx *cpuc;
 	struct task_ctx *taskc;
 	u64 now = scx_bpf_now();
@@ -1120,7 +1034,7 @@ void BPF_STRUCT_OPS(lavd_running, struct task_struct *p)
 	/*
 	 * Update task statistics
 	 */
-	cpuc = get_cpu_ctx();
+	cpuc = get_cpu_ctx_task(p);
 	taskc = get_task_ctx(p);
 	if (!cpuc || !taskc)
 		return;
@@ -1130,7 +1044,7 @@ void BPF_STRUCT_OPS(lavd_running, struct task_struct *p)
 	/*
 	 * Calculate the task's time slice.
 	 */
-	p->scx.slice = calc_time_slice(p, taskc);
+	p->scx.slice = calc_time_slice(taskc);
 
 	/*
 	 * Calculate the task's CPU performance target and update if the new
@@ -1138,8 +1052,7 @@ void BPF_STRUCT_OPS(lavd_running, struct task_struct *p)
 	 * urgently increases according to task's target but it decreases
 	 * gradually according to EWMA of past performance targets.
 	 */
-	calc_cpuperf_target(stat_cur, taskc, cpuc);
-	try_increase_cpuperf_target(cpuc);
+	update_cpuperf_target(cpuc);
 
 	/*
 	 * Update running task's information for preemption
@@ -1161,7 +1074,7 @@ void BPF_STRUCT_OPS(lavd_stopping, struct task_struct *p, bool runnable)
 	/*
 	 * Update task statistics
 	 */
-	cpuc = get_cpu_ctx();
+	cpuc = get_cpu_ctx_task(p);
 	taskc = get_task_ctx(p);
 	if (!cpuc || !taskc)
 		return;
@@ -1178,7 +1091,7 @@ void BPF_STRUCT_OPS(lavd_quiescent, struct task_struct *p, u64 deq_flags)
 	/*
 	 * Substract task load from the current CPU's load.
 	 */
-	cpuc = get_cpu_ctx();
+	cpuc = get_cpu_ctx_task(p);
 	taskc = get_task_ctx(p);
 	if (!cpuc || !taskc)
 		return;
@@ -1350,27 +1263,71 @@ void BPF_STRUCT_OPS(lavd_set_cpumask, struct task_struct *p,
 	set_on_core_type(taskc, cpumask);
 }
 
+void BPF_STRUCT_OPS(lavd_cpu_acquire, s32 cpu,
+		    struct scx_cpu_acquire_args *args)
+{
+	struct cpu_ctx *cpuc;
+	u64 dur, scaled_dur;
+
+	cpuc = get_cpu_ctx_id(cpu);
+	if (!cpuc) {
+		scx_bpf_error("Failed to lookup cpu_ctx %d", cpu);
+		return;
+	}
+
+	/*
+	 * When regaining control of a CPU under the higher priority scheduler
+	 * class, measure how much time the higher priority scheduler class
+	 * used -- i.e., [lavd_cpu_release, lavd_cpu_acquire]. This will be
+	 * used to calculate capacity-invariant and frequency-invariant CPU
+	 * utilization.
+	 */
+	dur = time_delta(scx_bpf_now(), cpuc->cpu_release_clk);
+	scaled_dur = scale_cap_freq(dur, cpu);
+	cpuc->tot_sc_time += scaled_dur;
+
+	/*
+	 * The higher-priority scheduler class could change the CPU frequency,
+	 * so let's keep track of the frequency when we gain the CPU control.
+	 * This helps to make the frequency update decision.
+	 */
+	cpuc->cpuperf_cur = scx_bpf_cpuperf_cur(cpu);
+}
+
 void BPF_STRUCT_OPS(lavd_cpu_release, s32 cpu,
 		    struct scx_cpu_release_args *args)
 {
 	struct cpu_ctx *cpuc;
 
-	/*
-	 * When the scx scheduler loses control of a CPU,
-	 * reset the CPU's preemption information so it cannot be a victim.
-	 */
 	cpuc = get_cpu_ctx_id(cpu);
 	if (!cpuc) {
 		scx_bpf_error("Failed to lookup cpu_ctx %d", cpu);
 		return;
 	}
-	reset_cpu_preemption_info(cpuc, true);
 
 	/*
 	 * When a CPU is released to serve higher priority scheduler class,
-	 * requeue the tasks in a local DSQ to the global enqueue.
+	 * reset the CPU's preemption information so it cannot be a victim.
+	 */
+	reset_cpu_preemption_info(cpuc, true);
+
+	/*
+	 * Requeue the tasks in a local DSQ to the global enqueue.
 	 */
 	scx_bpf_reenqueue_local();
+
+	/*
+	 * Reset the current CPU's performance target, so we can set
+	 * the target properly after regaining the control.
+	 */
+	reset_cpuperf_target(cpuc);
+
+	/*
+	 * Keep track of when the higher-priority scheduler class takes
+	 * the CPUto calculate capacity-invariant and frequency-invariant
+	 * CPU utilization.
+	 */
+	cpuc->cpu_release_clk = scx_bpf_now();
 }
 
 void BPF_STRUCT_OPS(lavd_enable, struct task_struct *p)
@@ -1391,15 +1348,14 @@ void BPF_STRUCT_OPS(lavd_enable, struct task_struct *p)
 
 static void init_task_ctx(struct task_struct *p, struct task_ctx *taskc)
 {
-	struct sys_stat *stat_cur = get_sys_stat_cur();
 	u64 now = scx_bpf_now();
 
 	__builtin_memset(taskc, 0, sizeof(*taskc));
 	taskc->is_affinitized = bpf_cpumask_weight(p->cpus_ptr) != nr_cpu_ids;
-	taskc->last_running_clk = now; /* for run_time_ns */
-	taskc->last_stopping_clk = now; /* for run_time_ns */
-	taskc->run_time_ns = slice_max_ns;
-	taskc->svc_time = stat_cur->avg_svc_time * LAVD_NEW_PROC_PENALITY;
+	taskc->last_running_clk = now; /* for avg_runtime */
+	taskc->last_stopping_clk = now; /* for avg_runtime */
+	taskc->avg_runtime = slice_max_ns;
+	taskc->svc_time = sys_stat.avg_svc_time;
 
 	set_on_core_type(taskc, p->cpus_ptr);
 }
@@ -1595,10 +1551,11 @@ static s32 init_per_cpu_ctx(u64 now)
 		cpuc->stopping_tm_est_ns = SCX_SLICE_INF;
 		cpuc->online_clk = now;
 		cpuc->offline_clk = now;
+		cpuc->cpu_release_clk = now;
 		cpuc->is_online = bpf_cpumask_test_cpu(cpu, online_cpumask);
 		cpuc->capacity = get_cpuperf_cap(cpu);
 		cpuc->cpdom_poll_pos = cpu % LAVD_CPDOM_MAX_NR;
-		cpuc->min_perf_cri = 1000;
+		cpuc->min_perf_cri = LAVD_SCALE;
 		cpuc->futex_op = LAVD_FUTEX_OP_INVALID;
 
 		if (cpuc->capacity > 0) {
@@ -1642,7 +1599,7 @@ static s32 init_per_cpu_ctx(u64 now)
 			debugln("CPU %d is a turbo core.", cpu);
 		}
 	}
-	default_big_core_ratio = (1000 * big_capacity) / sum_capacity;
+	default_big_core_scale = (big_capacity << LAVD_SHIFT) / sum_capacity;
 
 	/*
 	 * Initialize compute domain id.
@@ -1765,7 +1722,6 @@ SCX_OPS_DEFINE(lavd_ops,
 	       .select_cpu		= (void *)lavd_select_cpu,
 	       .enqueue			= (void *)lavd_enqueue,
 	       .dispatch		= (void *)lavd_dispatch,
-	       .tick			= (void *)lavd_tick,
 	       .runnable		= (void *)lavd_runnable,
 	       .running			= (void *)lavd_running,
 	       .stopping		= (void *)lavd_stopping,
@@ -1774,6 +1730,7 @@ SCX_OPS_DEFINE(lavd_ops,
 	       .cpu_offline		= (void *)lavd_cpu_offline,
 	       .update_idle		= (void *)lavd_update_idle,
 	       .set_cpumask		= (void *)lavd_set_cpumask,
+	       .cpu_acquire		= (void *)lavd_cpu_acquire,
 	       .cpu_release		= (void *)lavd_cpu_release,
 	       .enable			= (void *)lavd_enable,
 	       .init_task		= (void *)lavd_init_task,
