@@ -17,7 +17,7 @@ static bool		have_little_core;
 
 const volatile u16	cpu_order_performance[LAVD_CPU_ID_MAX]; /* CPU preference order for performance and balanced mode */
 const volatile u16	cpu_order_powersave[LAVD_CPU_ID_MAX]; /* CPU preference order for powersave mode */
-const volatile u16	__cpu_capacity_hint[LAVD_CPU_ID_MAX]; /* CPU capacity based on 1000 */
+const volatile u16	cpu_capacity[LAVD_CPU_ID_MAX]; /* CPU capacity based on 1024 */
 
 static int		nr_cpdoms; /* number of compute domains */
 struct cpdom_ctx	cpdom_ctxs[LAVD_CPDOM_MAX_NR]; /* contexts for compute domains */
@@ -25,14 +25,14 @@ private(LAVD) struct bpf_cpumask cpdom_cpumask[LAVD_CPDOM_MAX_NR]; /* online CPU
 
 
 /*
- * Big core's compute ratio among currently active cores
+ * Big core's compute ratio among currently active cores scaled by 1024.
  */
-static u32		cur_big_core_ratio;
+static u32		cur_big_core_scale;
 
 /*
- * Big core's compute ratio when all cores are active
+ * Big core's compute ratio when all cores are active scaled by 1024.
  */
-static u32		default_big_core_ratio;
+static u32		default_big_core_scale;
 
 /*
  * Statistics
@@ -43,36 +43,16 @@ volatile u64		performance_mode_ns;
 volatile u64		balanced_mode_ns;
 volatile u64		powersave_mode_ns;
 
-static bool is_perf_cri(struct task_ctx *taskc, struct sys_stat *stat_cur)
+static bool is_perf_cri(struct task_ctx *taskc)
 {
 	if (!have_little_core)
 		return true;
 
 	if (READ_ONCE(taskc->on_big) && READ_ONCE(taskc->on_little))
-		return taskc->perf_cri >= stat_cur->thr_perf_cri;
+		return taskc->perf_cri >= sys_stat.thr_perf_cri;
 	return READ_ONCE(taskc->on_big);
 }
 
-static u64 calc_nr_active_cpus(struct sys_stat *stat_cur)
-{
-	u64 nr_active;
-
-	/*
-	 * nr_active = ceil(nr_cpus_onln * cpu_util * per_core_max_util)
-	 */
-	nr_active  = (nr_cpus_onln * stat_cur->util * 1000) + 500;
-	nr_active /= (LAVD_CC_PER_CORE_MAX_CTUIL * 1000);
-
-	/*
-	 * If a few CPUs are particularly busy, boost the active CPUs more.
-	 */
-	nr_active += min(LAVD_CC_NR_OVRFLW, (stat_cur->nr_violation) / 1000);
-	nr_active = max(min(nr_active, nr_cpus_onln),
-			LAVD_CC_NR_ACTIVE_MIN);
-
-	return nr_active;
-}
-
 static bool clear_cpu_periodically(u32 cpu, struct bpf_cpumask *cpumask)
 {
 	u32 clear;
@@ -100,14 +80,56 @@ static const volatile u16 *get_cpu_order(void)
 		return cpu_order_performance;
 }
 
+static int calc_nr_active_cpus(void)
+{
+	const volatile u16 *cpu_order;
+	u64 req_cap, cap_cpu, cap_sum = 0;
+	u16 cpu_id, i;
+
+	/*
+	 * Calculate the required compute capacity:
+	 *
+	 * Scaled utilization assumes all the CPUs are the fastest ones
+	 * running at the highest frequency. So the required compute capacity
+	 * given the scaled utilization is defined as follows:
+	 *
+	 * req_cpacity = total_compute_capaticy * scaled_utilization
+	 *             = (nr_cpus_onln * 1024) * (avg_sc_util / 1024)
+	 *             = nr_cpus_only * scaled_utilization
+	 */
+	req_cap = nr_cpus_onln * sys_stat.avg_sc_util;
+
+	/*
+	 * Fill the required compute capacity in the CPU preference order,
+	 * utilizing each CPU in a certain % (LAVD_CC_PER_CORE_UTIL or
+	 * LAVD_CC_PER_CORE_SHIFT).
+	 */
+	cpu_order = get_cpu_order();
+	bpf_for(i, 0, nr_cpu_ids) {
+		if (i >= LAVD_CPU_ID_MAX)
+			return nr_cpu_ids;
+
+		cpu_id = cpu_order[i];
+		if (cpu_id >= LAVD_CPU_ID_MAX)
+			return nr_cpu_ids;
+
+		cap_cpu = cpu_capacity[cpu_id];
+		cap_sum += cap_cpu >> LAVD_CC_PER_CORE_SHIFT;
+		if (cap_sum >= req_cap)
+			return i+1;
+	}
+
+	/* Should not be here. */
+	return nr_cpu_ids;
+}
+
 static void do_core_compaction(void)
 {
-	struct sys_stat *stat_cur = get_sys_stat_cur();
 	const volatile u16 *cpu_order = get_cpu_order();
 	struct cpu_ctx *cpuc;
 	struct bpf_cpumask *active, *ovrflw, *cd_cpumask;
 	struct cpdom_ctx *cpdomc;
-	int nr_cpus, nr_active, nr_active_old, cpu, i;
+	int nr_active, nr_active_old, cpu, i;
 	u32 sum_capacity = 0, big_capacity = 0;
 	bool clear;
 	u64 cpdom_id;
@@ -127,9 +149,8 @@ static void do_core_compaction(void)
 	/*
 	 * Assign active and overflow cores
 	 */
-	nr_active_old = stat_cur->nr_active;
-	nr_active = calc_nr_active_cpus(stat_cur);
-	nr_cpus = nr_active + LAVD_CC_NR_OVRFLW;
+	nr_active_old = sys_stat.nr_active;
+	nr_active = calc_nr_active_cpus();
 	bpf_for(i, 0, nr_cpu_ids) {
 		if (i >= LAVD_CPU_ID_MAX)
 			break;
@@ -148,19 +169,13 @@ static void do_core_compaction(void)
 		/*
 		 * Assign an online cpu to active and overflow cpumasks
 		 */
-		if (i < nr_cpus) {
-			if (i < nr_active) {
-				bpf_cpumask_set_cpu(cpu, active);
-				bpf_cpumask_clear_cpu(cpu, ovrflw);
+		if (i < nr_active) {
+			bpf_cpumask_set_cpu(cpu, active);
+			bpf_cpumask_clear_cpu(cpu, ovrflw);
 
-				cpdomc = MEMBER_VPTR(cpdom_ctxs, [cpu]);
-				if (cpdomc)
-					WRITE_ONCE(cpdomc->is_active, true);
-			}
-			else {
-				bpf_cpumask_set_cpu(cpu, ovrflw);
-				bpf_cpumask_clear_cpu(cpu, active);
-			}
+			cpdomc = MEMBER_VPTR(cpdom_ctxs, [cpu]);
+			if (cpdomc)
+				WRITE_ONCE(cpdomc->is_active, true);
 			scx_bpf_kick_cpu(cpu, SCX_KICK_IDLE);
 
 			/*
@@ -169,37 +184,33 @@ static void do_core_compaction(void)
 			sum_capacity += cpuc->capacity;
 			if (cpuc->big_core)
 				big_capacity += cpuc->capacity;
-		}
-		else {
-			if (i < nr_active_old) {
-				bpf_cpumask_clear_cpu(cpu, active);
-				bpf_cpumask_clear_cpu(cpu, ovrflw);
-			}
-			else {
-				/*
-				 * This is the case when a CPU belongs to the
-				 * overflow set even though that CPU was not an
-				 * overflow set initially. This can happen only
-				 * when a pinned userspace task ran on this
-				 * CPU. In this case, we keep the CPU in an
-				 * overflow set since the CPU will be used
-				 * anyway for the task. This will promote equal
-				 * use of all used CPUs, lowering the energy
-				 * consumption by avoiding a few CPUs being
-				 * turbo-boosted. Hence, we do not clear the
-				 * overflow cpumask here for a while,
-				 * approximately for LAVD_CC_CPU_PIN_INTERVAL.
-				 */
-				bpf_cpumask_clear_cpu(cpu, active);
-				clear = clear_cpu_periodically(cpu, ovrflw);
-				if (!clear)
-					scx_bpf_kick_cpu(cpu, SCX_KICK_IDLE);
-			}
+		} else if (i < nr_active_old) {
+			bpf_cpumask_clear_cpu(cpu, active);
+			bpf_cpumask_clear_cpu(cpu, ovrflw);
+		} else {
+			/*
+			 * This is the case when a CPU belongs to the
+			 * overflow set even though that CPU was not an
+			 * overflow set initially. This can happen only
+			 * when a pinned userspace task ran on this
+			 * CPU. In this case, we keep the CPU in an
+			 * overflow set since the CPU will be used
+			 * anyway for the task. This will promote equal
+			 * use of all used CPUs, lowering the energy
+			 * consumption by avoiding a few CPUs being
+			 * turbo-boosted. Hence, we do not clear the
+			 * overflow cpumask here for a while,
+			 * approximately for LAVD_CC_CPU_PIN_INTERVAL.
+			 */
+			bpf_cpumask_clear_cpu(cpu, active);
+			clear = clear_cpu_periodically(cpu, ovrflw);
+			if (!clear)
+				scx_bpf_kick_cpu(cpu, SCX_KICK_IDLE);
 		}
 	}
 
-	cur_big_core_ratio = (1000 * big_capacity) / sum_capacity;
-	stat_cur->nr_active = nr_active;
+	cur_big_core_scale = (big_capacity << LAVD_SHIFT) / sum_capacity;
+	sys_stat.nr_active = nr_active;
 
 	/*
 	 * Maintain cpdomc->is_active reflecting the active set.
@@ -266,8 +277,6 @@ static int do_set_power_profile(s32 pm, int util)
 	switch (pm) {
 	case LAVD_PM_PERFORMANCE:
 		no_core_compaction = true;
-		no_freq_scaling = false;
-		no_prefer_turbo_core = false;
 		is_powersave_mode = false;
 
 		/*
@@ -284,16 +293,12 @@ static int do_set_power_profile(s32 pm, int util)
 		break;
 	case LAVD_PM_BALANCED:
 		no_core_compaction = false;
-		no_freq_scaling = false;
-		no_prefer_turbo_core = false;
 		is_powersave_mode = false;
 		reinit_cpumask_for_performance = false;
 		debugln("Set the scheduler's power profile to balanced mode: %d", util);
 		break;
 	case LAVD_PM_POWERSAVE:
 		no_core_compaction = false;
-		no_freq_scaling = false;
-		no_prefer_turbo_core = true;
 		is_powersave_mode = true;
 		reinit_cpumask_for_performance = false;
 		debugln("Set the scheduler's power profile to power-save mode: %d", util);
@@ -307,45 +312,42 @@ static int do_set_power_profile(s32 pm, int util)
 
 static int do_autopilot(void)
 {
-	struct sys_stat *stat_cur = get_sys_stat_cur();
-
 	/*
 	 * If the CPU utiulization is very low (say <= 5%), it means high
 	 * performance is not required. We run the scheduler in powersave mode
 	 * to save energy consumption.
 	 */
-	if (stat_cur->util <= LAVD_AP_LOW_UTIL)
-		return do_set_power_profile(LAVD_PM_POWERSAVE, stat_cur->util);
+	if (sys_stat.avg_util <= LAVD_AP_LOW_UTIL)
+		return do_set_power_profile(LAVD_PM_POWERSAVE, sys_stat.avg_util);
 
 	/*
 	 * If the CPU utiulization is moderate (say > 5%, <= 30%), we run the
 	 * scheduler in balanced mode. Actually, balanced mode can save energy
 	 * consumption only under moderate CPU load.
 	 */
-	if (stat_cur->util <= LAVD_AP_HIGH_UTIL)
-		return do_set_power_profile(LAVD_PM_BALANCED, stat_cur->util);
+	if (sys_stat.avg_util <= LAVD_AP_HIGH_UTIL)
+		return do_set_power_profile(LAVD_PM_BALANCED, sys_stat.avg_util);
 
 	/*
 	 * If the CPU utilization is high enough (say > 30%), we run the
 	 * scheduler in performance mode. The system indeed needs perrformance
 	 * also there is little energy benefit even under balanced mode anyway.
 	 */
-	return do_set_power_profile(LAVD_PM_PERFORMANCE, stat_cur->util);
+	return do_set_power_profile(LAVD_PM_PERFORMANCE, sys_stat.avg_util);
 }
 
 static void update_thr_perf_cri(void)
 {
-	struct sys_stat *stat_cur = get_sys_stat_cur();
-	u32 little_core_ratio, delta, diff, thr;
+	u32 little_core_scale, delta, diff, thr;
 
 	if (no_core_compaction || !have_little_core)
-		cur_big_core_ratio = default_big_core_ratio;
+		cur_big_core_scale = default_big_core_scale;
 
 	/*
 	 * If all active cores are big, all tasks should run on the big cores.
 	 */
-	if (cur_big_core_ratio == 1000) {
-		stat_cur->thr_perf_cri = 0;
+	if (cur_big_core_scale == LAVD_SCALE) {
+		sys_stat.thr_perf_cri = 0;
 		return;
 	}
 
@@ -365,12 +367,12 @@ static void update_thr_perf_cri(void)
 	 *
 	 *   <------------><------------------->
 	 *   |            |                    |
-	 *   |            |                    1000
-	 *   |            1000 - big_core_ratio (i.e., little_core_ratio)
+	 *   |            |                    1024
+	 *   |            1024 - big_core_scale (i.e., little_core_scale)
 	 *   0
 	 */
-	little_core_ratio = 1000 - cur_big_core_ratio;
-	if (little_core_ratio < 500) {
+	little_core_scale = LAVD_SCALE - cur_big_core_scale;
+	if (little_core_scale < p2s(50)) {
 		/*
 		 *   min_perf_cri
 		 *   |         avg_perf_cri
@@ -380,13 +382,13 @@ static void update_thr_perf_cri(void)
 		 *
 		 *   <-///-><-------------------------->
 		 *   |     |                           |
-		 *   |     |                           1000
-		 *   |     little_core_ratio
+		 *   |     |                           1024
+		 *   |     little_core_scale
 		 *   0
 		 */
-		delta = stat_cur->avg_perf_cri - stat_cur->min_perf_cri;
-		diff = (delta * little_core_ratio) / 1000;
-		thr = diff + stat_cur->min_perf_cri;
+		delta = sys_stat.avg_perf_cri - sys_stat.min_perf_cri;
+		diff = (delta * little_core_scale) >> LAVD_SHIFT;
+		thr = diff + sys_stat.min_perf_cri;
 	}
 	else {
 		/*
@@ -398,22 +400,23 @@ static void update_thr_perf_cri(void)
 		 *
 		 *   <---------------------><-////////->
 		 *   |                     |           |
-		 *   |                     |           1000
-		 *   |                     little_core_ratio
+		 *   |                     |           1024
+		 *   |                     little_core_scale
 		 *   0
 		 */
-		delta = stat_cur->max_perf_cri - stat_cur->avg_perf_cri;
-		diff = (delta * cur_big_core_ratio) / 1000;
-		thr = stat_cur->max_perf_cri - diff;
+		delta = sys_stat.max_perf_cri - sys_stat.avg_perf_cri;
+		diff = (delta * cur_big_core_scale) >> LAVD_SHIFT;
+		thr = sys_stat.max_perf_cri - diff;
 	}
 
-	stat_cur->thr_perf_cri = thr;
+	sys_stat.thr_perf_cri = thr;
 }
 
 static int reinit_active_cpumask_for_performance(void)
 {
 	struct cpu_ctx *cpuc;
 	struct bpf_cpumask *active, *ovrflw;
+	const struct cpumask *online_cpumask;
 	int cpu, err = 0;
 
 	barrier();
@@ -432,125 +435,81 @@ static int reinit_active_cpumask_for_performance(void)
 
 
 	/*
-	 * Once core compaction becomes off in performance mode,
-	 * reinitialize active/overflow cpumasks to reflect the mode change.
+	 * Once core compaction becomes off in performance mode, reinitialize
+	 * active/overflow cpumasks to reflect the mode change.
+	 * In an asymmetric system, big cores belong to the active set but
+	 * little cores the overflow set to prefer big cores for performance.
+	 * In a symmetric system, all online CPUs belong to the active set.
 	 */
-	bpf_for(cpu, 0, nr_cpu_ids) {
-		cpuc = get_cpu_ctx_id(cpu);
-		if (!cpuc) {
-			scx_bpf_error("Failed to lookup cpu_ctx: %d", cpu);
-			err = -ESRCH;
-			goto unlock_out;
-		}
+	if (have_little_core) {
+		bpf_for(cpu, 0, nr_cpu_ids) {
+			cpuc = get_cpu_ctx_id(cpu);
+			if (!cpuc) {
+				err = -ESRCH;
+				goto unlock_out;
+			}
 
-		if (cpuc->big_core) {
-			bpf_cpumask_set_cpu(cpu, active);
-			bpf_cpumask_clear_cpu(cpu, ovrflw);
-		}
-		else {
-			bpf_cpumask_set_cpu(cpu, ovrflw);
-			bpf_cpumask_clear_cpu(cpu, active);
+			if (cpuc->big_core) {
+				bpf_cpumask_set_cpu(cpu, active);
+				bpf_cpumask_clear_cpu(cpu, ovrflw);
+			}
+			else {
+				bpf_cpumask_set_cpu(cpu, ovrflw);
+				bpf_cpumask_clear_cpu(cpu, active);
+			}
 		}
+	} else {
+		online_cpumask = scx_bpf_get_online_cpumask();
+		nr_cpus_onln = bpf_cpumask_weight(online_cpumask);
+		bpf_cpumask_copy(active, online_cpumask);
+		scx_bpf_put_cpumask(online_cpumask);
+
+		bpf_cpumask_clear(ovrflw);
 	}
+	sys_stat.nr_active = nr_cpus_onln;
 
 unlock_out:
 	bpf_rcu_read_unlock();
 	return err;
 }
 
-static int calc_cpuperf_target(struct sys_stat *stat_cur,
-			       struct task_ctx *taskc, struct cpu_ctx *cpuc)
+static void update_cpuperf_target(struct cpu_ctx *cpuc)
 {
-	u64 max_load, cpu_load;
-	u32 cpuperf_target;
-
-	if (!stat_cur || !taskc || !cpuc)
-		return -EINVAL;
-
-	if (no_freq_scaling) {
-		cpuc->cpuperf_task = SCX_CPUPERF_ONE;
-		cpuc->cpuperf_avg = SCX_CPUPERF_ONE;
-		return 0;
-	}
+	u32 util, cpuperf_target;
 
 	/*
-	 * We determine the clock frequency of a CPU using two factors: 1) the
-	 * current CPU utilization (cpuc->util) and 2) the current task's
-	 * performance criticality (taskc->perf_cri) compared to the
-	 * system-wide average performance criticality
-	 * (stat_cur->thr_perf_cri).
-	 *
-	 * When a current CPU utilization is 85% and the current task's
-	 * performance criticality is the same as the system-wide average
-	 * criticality, we set the target CPU frequency to the maximum.
-	 *
-	 * In other words, even if CPU utilization is not so high, the target
-	 * CPU frequency could be high when the task's performance criticality
-	 * is high enough (i.e., boosting CPU frequency). On the other hand,
-	 * the target CPU frequency could be low even if CPU utilization is
-	 * high when a non-performance-critical task is running (i.e.,
-	 * deboosting CPU frequency).
+	 * The CPU utilization decides the frequency. The bigger one between
+	 * the running average and the recent utilization is used to respond
+	 * quickly upon load spikes. When the utilization is greater than
+	 * LAVD_CPU_UTIL_MAX_FOR_CPUPERF (85%), ceil to 100%.
 	 */
-	max_load = stat_cur->thr_perf_cri * LAVD_CPU_UTIL_MAX_FOR_CPUPERF;
-	cpu_load = taskc->perf_cri * cpuc->util;
-	cpuperf_target = (cpu_load * SCX_CPUPERF_ONE) / max_load;
-	cpuperf_target = min(cpuperf_target, SCX_CPUPERF_ONE);
+	if (!no_freq_scaling) {
+		util = max(cpuc->avg_util, cpuc->cur_util) <
+			LAVD_CPU_UTIL_MAX_FOR_CPUPERF? : LAVD_SCALE;
+		cpuperf_target = (util * SCX_CPUPERF_ONE) >> LAVD_SHIFT;
+	} else
+		cpuperf_target = SCX_CPUPERF_ONE;
 
-	cpuc->cpuperf_task = cpuperf_target;
-	cpuc->cpuperf_avg = calc_avg32(cpuc->cpuperf_avg, cpuperf_target);
-	return 0;
+	/*
+	 * Update the performance target once it changes.
+	 */
+	if (cpuc->cpuperf_cur != cpuperf_target) {
+		scx_bpf_cpuperf_set(cpuc->cpu_id, cpuperf_target);
+		cpuc->cpuperf_cur = cpuperf_target;
+	}
 }
 
-static bool try_increase_cpuperf_target(struct cpu_ctx *cpuc)
+static void reset_cpuperf_target(struct cpu_ctx *cpuc)
 {
-	/*
-	 * When a task becomes running, update CPU's performance target only
-	 * when the current task's target performance is higher. This helps
-	 * rapidly adopt workload changes by rapidly increasing CPU's
-	 * performance target.
-	 */
-	u32 target;
-
-	if (!cpuc)
-		return false;
-
-	target = max(cpuc->cpuperf_task, cpuc->cpuperf_avg);
-	if (cpuc->cpuperf_cur < target) {
-		cpuc->cpuperf_cur = target;
-		scx_bpf_cpuperf_set(cpuc->cpu_id, target);
-		return true;
+	if (!no_freq_scaling) {
+		cpuc->cpuperf_cur = 0;
 	}
-
-	return false;
-}
-
-static bool try_decrease_cpuperf_target(struct cpu_ctx *cpuc)
-{
-	/*
-	 * Upon every tick interval, we try to decrease the CPU's performance
-	 * target if the current one is higher than both the current task's
-	 * target and EWMA of past targets. This helps gradually adopt workload
-	 * changes upon sudden down falls.
-	 */
-	u32 target;
-
-	if (!cpuc)
-		return false;
-
-	target = max(cpuc->cpuperf_task, cpuc->cpuperf_avg);
-	if (cpuc->cpuperf_cur != target) {
-		cpuc->cpuperf_cur = target;
-		scx_bpf_cpuperf_set(cpuc->cpu_id, target);
-		return true;
-	}
-
-	return false;
 }
 
 static u16 get_cpuperf_cap(s32 cpu)
 {
 	if (cpu >= 0 && cpu < nr_cpu_ids && cpu < LAVD_CPU_ID_MAX)
-		return __cpu_capacity_hint[cpu];
+		return cpu_capacity[cpu];
 
 	debugln("Infeasible CPU id: %d", cpu);
 	return 0;
@@ -565,8 +524,8 @@ static u16 get_cputurbo_cap(void)
 	 * Find the maximum CPU capacity
 	 */
 	for (cpu = 0; cpu < nr_cpu_ids && cpu < LAVD_CPU_ID_MAX; cpu++) {
-		if (__cpu_capacity_hint[cpu] > turbo_cap) {
-			turbo_cap = __cpu_capacity_hint[cpu];
+		if (cpu_capacity[cpu] > turbo_cap) {
+			turbo_cap = cpu_capacity[cpu];
 			nr_turbo++;
 		}
 	}
@@ -580,6 +539,21 @@ static u16 get_cputurbo_cap(void)
 	return turbo_cap;
 }
 
+static u64 scale_cap_freq(u64 dur, s32 cpu)
+{
+	u64 cap, freq, scaled_dur;
+
+	/*
+	 * Scale the duration by CPU capacity and frequency, so calculate
+	 * capacity-invariant and frequency-invariant time duration.
+	 */
+	cap = get_cpuperf_cap(cpu);
+	freq = scx_bpf_cpuperf_cur(cpu);
+	scaled_dur = (dur * cap * freq) >> (LAVD_SHIFT * 2);
+
+	return scaled_dur;
+}
+
 static void init_autopilot_low_util(void)
 {
 	if (nr_cpus_big < nr_cpus_onln) {
@@ -587,14 +561,14 @@ static void init_autopilot_low_util(void)
 		 * When there are little cores, we move up to the balanced mode
 		 * if one little core is fully utilized.
 		 */
-		LAVD_AP_LOW_UTIL = 1000 / nr_cpus_onln;
+		LAVD_AP_LOW_UTIL = LAVD_SCALE / nr_cpus_onln;
 	}
 	else {
 		/*
 		 * When there are only big cores, we move up to the balanced
 		 * mode if two big cores are fully utilized.
 		 */
-		LAVD_AP_LOW_UTIL = (2 * 1000) / nr_cpus_onln;
+		LAVD_AP_LOW_UTIL = (2 * LAVD_SCALE) / nr_cpus_onln;
 	}
 }
 
