@@ -14,19 +14,19 @@ struct preemption_info {
 
 static u64 get_est_stopping_time(struct task_ctx *taskc, u64 now)
 {
-	return now + taskc->run_time_ns;
+	return now + taskc->avg_runtime;
 }
 
 static int comp_preemption_info(struct preemption_info *prm_a,
 				struct preemption_info *prm_b)
 {
 	/*
-	 * Check if one's latency priority _or_ deadline is smaller or not.
+	 * Check one's latency criticality and deadline.
 	 */
-	if ((prm_a->lat_cri > prm_b->lat_cri) ||
+	if ((prm_a->lat_cri > prm_b->lat_cri) &&
 	    (prm_a->stopping_tm_est_ns < prm_b->stopping_tm_est_ns))
 		return -1;
-	if ((prm_a->lat_cri < prm_b->lat_cri) ||
+	if ((prm_a->lat_cri < prm_b->lat_cri) &&
 	    (prm_a->stopping_tm_est_ns > prm_b->stopping_tm_est_ns))
 		return 1;
 	return 0;
@@ -73,14 +73,11 @@ static bool can_cpu1_kick_cpu2(struct preemption_info *prm_cpu1,
 static bool is_worth_kick_other_task(struct task_ctx *taskc)
 {
 	/*
-	 * The scx_bpf_kick_cpu() used for preemption is expensive as an IPI is
-	 * involved. Hence, we first judiciously check whether it is worth
-	 * trying to victimize another CPU as the current task is urgent
-	 * enough.
+	 * Preemption is not free. It is expensive involving context switching,
+	 * etc. Hence, we first judiciously check whether it is worth trying to
+	 * victimize another CPU as the current task is urgent enough.
 	 */
-	struct sys_stat *stat_cur = get_sys_stat_cur();
-
-	return (taskc->lat_cri >= stat_cur->thr_lat_cri);
+	return (taskc->lat_cri >= sys_stat.thr_lat_cri);
 }
 
 static bool can_cpu_be_kicked(u64 now, struct cpu_ctx *cpuc)
@@ -118,13 +115,6 @@ static struct cpu_ctx *find_victim_cpu(const struct cpumask *cpumask,
 	}
 	cur_cpu = cpuc->cpu_id;
 
-	/*
-	 * First check if it is worth to try to kick other CPU
-	 * at the expense of IPI.
-	 */
-	if (!is_worth_kick_other_task(taskc))
-		goto null_out;
-
 	/*
 	 * Randomly find _two_ CPUs that run lower-priority tasks than @p. To
 	 * traverse CPUs in a random order, we start from a random CPU ID in a
@@ -195,140 +185,75 @@ null_out:
 	return NULL;
 }
 
-static bool try_kick_cpu(struct task_struct *p, struct cpu_ctx *victim_cpuc)
+static void ask_cpu_yield(struct cpu_ctx *victim_cpuc)
 {
 	/*
-	 * Kicking the victim CPU does _not_ guarantee that task @p will run on
-	 * that CPU. Enqueuing @p to the global queue is one operation, and
-	 * kicking the victim is another asynchronous operation. However, it is
-	 * okay because, anyway, the victim CPU will run a higher-priority task
-	 * than @p.
+	 * Note that we avoid using scx_bpf_kick_cpu() on purpose.
+	 * While scx_bpf_kick_cpu() can trigger a task preemption immediately,
+	 * it incurs an expensive IPI operation. Furthermore, an IPI operation
+	 * is more costly in certain processor architectures or in older
+	 * generations of processors, causing performance variations among
+	 * processors. Thus, let's avoid using the IPI, scx_bpf_kick_cpu(), and
+	 * set the victim task's time slice to zero so the victim task yields
+	 * the CPU in the next scheduling point.
 	 */
-	u64 old;
-	bool ret = false;
+	struct rq *victim_rq;
+	struct task_struct *victim_p;
 
-	/*
-	 * Kick a victim CPU if it is not victimized yet by another
-	 * concurrent kick task.
-	 *
-	 *
-	 */
-	old = p->scx.slice;
-	if (old != 1 && old != 0)
-		ret = __sync_bool_compare_and_swap(&p->scx.slice, old, 1);
-
-	/*
-	 * Kick the remote CPU for preemption.
-	 */
-	if (ret)
-		scx_bpf_kick_cpu(victim_cpuc->cpu_id, SCX_KICK_PREEMPT);
-
-	return ret;
+	victim_rq = scx_bpf_cpu_rq(victim_cpuc->cpu_id);
+	if (victim_rq && (victim_p = victim_rq->curr))
+		WRITE_ONCE(victim_p->scx.slice, 0);
 }
 
-static bool try_find_and_kick_victim_cpu(struct task_struct *p,
-					 struct task_ctx *taskc,
-					 struct cpu_ctx *cpuc_cur,
-					 u64 dsq_id, u64 now)
+static void try_find_and_kick_victim_cpu(struct task_struct *p,
+					 struct task_ctx *taskc, u64 dsq_id)
 {
 	struct bpf_cpumask *cd_cpumask, *cpumask;
 	struct cpdom_ctx *cpdomc;
 	struct cpu_ctx *victim_cpuc;
-	bool ret = false;
+	struct cpu_ctx *cpuc_cur;
+	u64 now;
 
 	/*
 	 * Don't even try to perform expensive preemption for greedy tasks.
 	 */
 	if (!is_eligible(taskc))
-		return false;
+		return;
+
+	/*
+	 * Check if it is worth to try to kick other CPU.
+	 */
+	if (!is_worth_kick_other_task(taskc))
+		return;
 
 	/*
 	 * Prepare a cpumak so we find a victim in @p's compute domain.
 	 */
+	cpuc_cur = get_cpu_ctx();
+	if (!cpuc_cur)
+		return;
+
 	cpumask = cpuc_cur->tmp_t_mask;
 	cpdomc = MEMBER_VPTR(cpdom_ctxs, [dsq_id]);
 	cd_cpumask = MEMBER_VPTR(cpdom_cpumask, [dsq_id]);
 	if (!cpdomc || !cd_cpumask || !cpumask)
-		return false;
+		return;
 
 	bpf_cpumask_and(cpumask, cast_mask(cd_cpumask), p->cpus_ptr);
 
 	/*
 	 * Find a victim CPU among CPUs that run lower-priority tasks.
 	 */
+	now = scx_bpf_now();
 	victim_cpuc = find_victim_cpu(cast_mask(cpumask), taskc, now);
 
 	/*
 	 * If a victim CPU is chosen, preempt the victim by kicking it.
 	 */
-	if (victim_cpuc)
-		ret = try_kick_cpu(p, victim_cpuc);
-
-	return ret;
-}
-
-static bool try_yield_current_cpu(struct task_struct *p_run,
-				  struct cpu_ctx *cpuc_run,
-				  struct task_ctx *taskc_run,
-				  u64 now)
-{
-	struct task_struct *p_wait;
-	struct task_ctx *taskc_wait;
-	struct preemption_info prm_run, prm_wait;
-	bool ret = false;
-
-	/*
-	 * If a task holds a lock, never yield.
-	 */
-	if (is_lock_holder(taskc_run))
-		return false;
-
-	/*
-	 * A slice-extended lock holder finally released the lock,
-	 * give up its extended time slice for fairness.
-	 */
-	if (taskc_run->lock_holder_xted) {
-		p_run->scx.slice = 0;
-		return true;
+	if (victim_cpuc) {
+		ask_cpu_yield(victim_cpuc);
+		cpuc_cur->nr_preempt++;
 	}
-
-	/*
-	 *  If a task already exhausted its time slice, there is nothing to do.
-	 */
-	if (READ_ONCE(p_run->scx.slice) == 0)
-		return false;
-
-	/*
-	 * If there is a higher priority task waiting on the global rq, the
-	 * current running task yield the CPU by shrinking its time slice to
-	 * zero.
-	 */
-	prm_run.stopping_tm_est_ns = taskc_run->last_running_clk +
-				     taskc_run->run_time_ns;
-	prm_run.lat_cri = taskc_run->lat_cri;
-
-	bpf_rcu_read_lock();
-	bpf_for_each(scx_dsq, p_wait, cpuc_run->cpdom_id, 0) {
-		taskc_wait = get_task_ctx(p_wait);
-		if (!taskc_wait)
-			break;
-
-		prm_wait.stopping_tm_est_ns = get_est_stopping_time(taskc_wait, now);
-		prm_wait.lat_cri = taskc_wait->lat_cri;
-
-		ret = can_task1_kick_task2(&prm_wait, &prm_run);
-
-		/*
-		 * Test only the first entry on the DSQ.
-		 */
-		break;
-	}
-	bpf_rcu_read_unlock();
-
-	if (ret)
-		p_run->scx.slice = 0;
-
-	return ret;
 }
 
 static void reset_cpu_preemption_info(struct cpu_ctx *cpuc, bool released)
