@@ -46,12 +46,15 @@ use scx_stats::prelude::*;
 use scx_utils::autopower::{fetch_power_profile, PowerProfile};
 use scx_utils::build_id;
 use scx_utils::compat;
+use scx_utils::read_cpulist;
 use scx_utils::scx_ops_attach;
 use scx_utils::scx_ops_load;
 use scx_utils::scx_ops_open;
 use scx_utils::set_rlimit_infinity;
 use scx_utils::uei_exited;
 use scx_utils::uei_report;
+use scx_utils::Cpumask;
+use scx_utils::EnergyModel;
 use scx_utils::Topology;
 use scx_utils::UserExitInfo;
 use scx_utils::NR_CPU_IDS;
@@ -68,24 +71,42 @@ use stats::SysStats;
 /// See the more detailed overview of the LAVD design at main.bpf.c.
 #[derive(Debug, Parser)]
 struct Opts {
-    /// Automatically decide the scheduler's power mode based on system load.
-    /// This is a default mode if you don't specify the following options:
+    /// Automatically decide the scheduler's power mode (performance vs.
+    /// powersave vs. balanced), CPU preference order, etc, based on system
+    /// load. The options affecting the power mode and the use of core compaction
+    /// (--autopower, --performance, --powersave, --balanced,
+    /// --no-core-compaction) cannot be used with this option. When no option
+    /// is specified, this is a default mode.
     #[clap(long = "autopilot", action = clap::ArgAction::SetTrue)]
     autopilot: bool,
 
-    /// Automatically decide the scheduler's power mode based on the system's active power profile.
+    /// Automatically decide the scheduler's power mode (performance vs.
+    /// powersave vs. balanced) based on the system's active power profile.
+    /// The scheduler's power mode decides the CPU preference order and the use
+    /// of core compaction, so the options affecting these (--autopilot,
+    /// --performance, --powersave, --balanced, --no-core-compaction) cannot
+    /// be used with this option.
     #[clap(long = "autopower", action = clap::ArgAction::SetTrue)]
     autopower: bool,
 
-    /// Run in performance mode to get maximum performance.
+    /// Run the scheduler in performance mode to get maximum performance.
+    /// This option cannot be used with other conflicting options (--autopilot,
+    /// --autopower, --balanced, --powersave, --no-core-compaction)
+    /// affecting the use of core compaction.
     #[clap(long = "performance", action = clap::ArgAction::SetTrue)]
     performance: bool,
 
-    /// Run in powersave mode to minimize power consumption.
+    /// Run the scheduler in powersave mode to minimize powr consumption.
+    /// This option cannot be used with other conflicting options (--autopilot,
+    /// --autopower, --performance, --balanced, --no-core-compaction)
+    /// affecting the use of core compaction.
     #[clap(long = "powersave", action = clap::ArgAction::SetTrue)]
     powersave: bool,
 
-    /// Run in balanced mode aiming for sweetspot between power and performance (default).
+    /// Run the scheduler in balanced mode aiming for sweetspot between power
+    /// and performance. This option cannot be used with other conflicting
+    /// options (--autopilot, --autopower, --performance, --powersave,
+    /// --no-core-compaction) affecting the use of core compaction.
     #[clap(long = "balanced", action = clap::ArgAction::SetTrue)]
     balanced: bool,
 
@@ -97,36 +118,37 @@ struct Opts {
     #[clap(long = "slice-min-us", default_value = "300")]
     slice_min_us: u64,
 
+    /// List of CPUs in preferred order (e.g., "0-3,7,6,5,4"). The scheduler
+    /// uses the CPU preference mode only when the core compaction is enabled
+    /// (i.e., balanced or powersave mode is specified as an option or chosen
+    /// in the autopilot or autopower mode).
+    #[clap(long = "cpu-pref-order", default_value = "")]
+    cpu_pref_order: String,
+
     /// Do not boost futex holders.
     #[clap(long = "no-futex-boost", action = clap::ArgAction::SetTrue)]
     no_futex_boost: bool,
 
-    /// Disable core compaction and schedule tasks across all online CPUs. Core compaction attempts
-    /// to keep idle CPUs idle in favor of scheduling tasks on CPUs that are already
-    /// awake. See main.bpf.c for more info. Normally set by the power mode, but can be set independently if
-    /// desired.
+    /// Disable preemption.
+    #[clap(long = "no-preemption", action = clap::ArgAction::SetTrue)]
+    no_preemption: bool,
+
+    /// Disable an optimization for synchronous wake-up.
+    #[clap(long = "no-wake-sync", action = clap::ArgAction::SetTrue)]
+    no_wake_sync: bool,
+
+    /// Disable core compaction so the scheduler uses all the online CPUs.
+    /// The core compaction attempts to minimize the number of actively used
+    /// CPUs for unaffinitized tasks, respecting the CPU preference order.
+    /// Normally, the core compaction is enabled by the power mode (i.e.,
+    /// balanced or powersave mode is specified as an option or chosen in
+    /// the autopilot or autopower mode). This option cannot be used with the
+    /// other options that control the core compaction (--autopilot,
+    /// --autopower, --performance, --balanced, --powersave).
     #[clap(long = "no-core-compaction", action = clap::ArgAction::SetTrue)]
     no_core_compaction: bool,
 
-    /// Schedule tasks on SMT siblings before using other physcial cores when core compaction is
-    /// enabled. Normally set by the power mode, but can be set independently if desired.
-    #[clap(long = "prefer-smt-core", action = clap::ArgAction::SetTrue)]
-    prefer_smt_core: bool,
-
-    /// Schedule tasks on little (efficiency) cores before big (performance) cores when core compaction is
-    /// enabled. Normally set by the power mode, but can be set independently if desired.
-    #[clap(long = "prefer-little-core", action = clap::ArgAction::SetTrue)]
-    prefer_little_core: bool,
-
-    /// Do not specifically prefer to schedule on turbo cores. Normally set by the power mode, but
-    /// can be set independently if desired.
-    #[clap(long = "no-prefer-turbo-core", action = clap::ArgAction::SetTrue)]
-    no_prefer_turbo_core: bool,
-
-    /// Disable controlling the CPU frequency. In order to improve latency and responsiveness of
-    /// performance-critical tasks, scx_lavd increases the CPU frequency even if CPU usage is low.
-    /// See main.bpf.c for more info. Normally set by the power mode, but can be set independently
-    /// if desired.
+    /// Disable controlling the CPU frequency.
     #[clap(long = "no-freq-scaling", action = clap::ArgAction::SetTrue)]
     no_freq_scaling: bool,
 
@@ -158,46 +180,95 @@ struct Opts {
 }
 
 impl Opts {
-    fn nothing_specified(&self) -> bool {
-        self.autopilot == false
-            && self.autopower == false
+    fn can_autopilot(&self) -> bool {
+        self.autopower == false
             && self.performance == false
             && self.powersave == false
             && self.balanced == false
             && self.no_core_compaction == false
-            && self.prefer_smt_core == false
-            && self.prefer_little_core == false
-            && self.no_prefer_turbo_core == false
-            && self.no_freq_scaling == false
-            && self.monitor == None
-            && self.monitor_sched_samples == None
+    }
+
+    fn can_autopower(&self) -> bool {
+        self.autopilot == false
+            && self.performance == false
+            && self.powersave == false
+            && self.balanced == false
+            && self.no_core_compaction == false
+    }
+
+    fn can_performance(&self) -> bool {
+        self.autopilot == false
+            && self.autopower == false
+            && self.powersave == false
+            && self.balanced == false
+    }
+
+    fn can_balanced(&self) -> bool {
+        self.autopilot == false
+            && self.autopower == false
+            && self.performance == false
+            && self.powersave == false
+            && self.no_core_compaction == false
+    }
+
+    fn can_powersave(&self) -> bool {
+        self.autopilot == false
+            && self.autopower == false
+            && self.performance == false
+            && self.balanced == false
+            && self.no_core_compaction == false
     }
 
     fn proc(&mut self) -> Option<&mut Self> {
-        if self.nothing_specified() {
-            self.autopilot = true;
-            info!("Autopilot mode is enabled by default.");
+        if !self.autopilot {
+            self.autopilot = self.can_autopilot();
+        }
+        if self.autopilot {
+            if !self.can_autopilot() {
+                info!("Autopilot mode cannot be used with conflicting options.");
+                return None;
+            }
+            info!("Autopilot mode is enabled.");
+            return Some(self);
+        }
+
+        if self.autopower {
+            if !self.can_autopower() {
+                info!("Autopower mode cannot be used with conflicting options.");
+                return None;
+            }
+            info!("Autopower mode is enabled.");
             return Some(self);
         }
 
         if self.performance {
+            if !self.can_performance() {
+                info!("Performance mode cannot be used with conflicting options.");
+                return None;
+            }
+            info!("Performance mode is enabled.");
             self.no_core_compaction = true;
-            self.prefer_smt_core = false;
-            self.prefer_little_core = false;
-            self.no_prefer_turbo_core = false;
-            self.no_freq_scaling = true;
-        } else if self.powersave {
+            return Some(self);
+        }
+
+        if self.powersave {
+            if !self.can_powersave() {
+                info!("Powersave mode cannot be used with conflicting options.");
+                return None;
+            }
+            info!("Powersave mode is enabled.");
             self.no_core_compaction = false;
-            self.prefer_smt_core = true;
-            self.prefer_little_core = true;
-            self.no_prefer_turbo_core = true;
-            self.no_freq_scaling = false;
-        } else if self.balanced {
+            return Some(self);
+        }
+
+        if self.balanced {
+            if !self.can_balanced() {
+                info!("Balanced mode cannot be used with conflicting options.");
+                return None;
+            }
+            info!("Balanced mode is enabled.");
             self.no_core_compaction = false;
-            self.prefer_smt_core = false;
-            self.prefer_little_core = false;
-            self.no_prefer_turbo_core = false;
-            self.no_freq_scaling = false;
+            return Some(self);
         }
 
         Some(self)
@@ -222,14 +293,13 @@ impl introspec {
 #[derive(Debug, Clone)]
 struct CpuFlatId {
     node_id: usize,
+    pd_id: usize,
     llc_pos: usize,
     core_pos: usize,
     cpu_pos: usize,
     cpu_id: usize,
-    core_id: usize,
-    l2_id: usize,
-    l3_id: usize,
-    sharing_lvl: usize,
+    smt_level: usize,
+    cache_size: usize,
     cpu_cap: usize,
 }
 
@@ -250,6 +320,7 @@ struct ComputeDomainValue {
 
 #[derive(Debug)]
 struct FlatTopology {
+    all_cpus_mask: Cpumask,
     cpu_fids_performance: Vec<CpuFlatId>,
     cpu_fids_powersave: Vec<CpuFlatId>,
     cpdom_map: BTreeMap<ComputeDomainKey, ComputeDomainValue>,
@@ -275,47 +346,53 @@ impl fmt::Display for FlatTopology {
 impl FlatTopology {
     /// Build a flat-structured topology
     pub fn new() -> Result<FlatTopology> {
-        let topo = Topology::new().expect("Failed to build host topology");
-        let (cpu_fids_performance, avg_cap) = Self::build_cpu_fids(&topo, false, false).unwrap();
-        let (cpu_fids_powersave, _) = Self::build_cpu_fids(&topo, true, true).unwrap();
+        let sys_topo = Topology::new().expect("Failed to build host topology");
+        let sys_em = EnergyModel::new();
+        debug!("{:#?}", sys_topo);
+        debug!("{:#?}", sys_em);
+
+        let (cpu_fids_performance, avg_cap) =
+            Self::build_cpu_fids(&sys_topo, &sys_em, false).unwrap();
+        let (cpu_fids_powersave, _) = Self::build_cpu_fids(&sys_topo, &sys_em, true).unwrap();
 
         // Note that building compute domain is not dependent to CPU orer
         // so it is okay to use any cpu_fids_*.
         let cpdom_map = Self::build_cpdom(&cpu_fids_performance, avg_cap).unwrap();
 
         Ok(FlatTopology {
+            all_cpus_mask: sys_topo.span,
             cpu_fids_performance,
             cpu_fids_powersave,
             cpdom_map,
-            smt_enabled: topo.smt_enabled,
+            smt_enabled: sys_topo.smt_enabled,
         })
     }
 
     /// Build a flat-structured list of CPUs in a preference order
     fn build_cpu_fids(
         topo: &Topology,
-        prefer_smt_core: bool,
-        prefer_little_core: bool,
+        em: &Result<EnergyModel>,
+        prefer_powersave: bool,
     ) -> Option<(Vec<CpuFlatId>, usize)> {
         let mut cpu_fids = Vec::new();
-        debug!("{:#?}", topo);
 
         // Build a vector of cpu flat ids.
         let mut avg_cap = 0;
         for (&node_id, node) in topo.nodes.iter() {
             for (llc_pos, (_llc_id, llc)) in node.llcs.iter().enumerate() {
-                for (core_pos, (core_id, core)) in llc.cores.iter().enumerate() {
+                for (core_pos, (_core_id, core)) in llc.cores.iter().enumerate() {
                     for (cpu_pos, (cpu_id, cpu)) in core.cpus.iter().enumerate() {
+                        let cpu_id = *cpu_id;
+                        let pd_id = Self::get_pd_id(em, cpu_id, node_id);
                         let cpu_fid = CpuFlatId {
                             node_id,
+                            pd_id,
                             llc_pos,
                             core_pos,
                             cpu_pos,
-                            cpu_id: *cpu_id,
-                            core_id: *core_id,
-                            l2_id: cpu.l2_id,
-                            l3_id: cpu.l3_id,
-                            sharing_lvl: 0,
+                            cpu_id,
+                            smt_level: cpu.smt_level,
+                            cache_size: cpu.cache_size,
                             cpu_cap: cpu.cpu_capacity,
                         };
                         cpu_fids.push(RefCell::new(cpu_fid));
@@ -326,28 +403,6 @@ impl FlatTopology {
         }
         avg_cap /= cpu_fids.len() as usize;
 
-        // Initialize cpu's hardware resource sharing level
-        for (a_id, cpu_fid_a) in cpu_fids.iter().enumerate() {
-            for (b_id, cpu_fid_b) in cpu_fids.iter().enumerate() {
-                if a_id == b_id {
-                    continue;
-                }
-
-                let mut a = cpu_fid_a.borrow_mut();
-                let b = cpu_fid_b.borrow();
-
-                if a.core_id == b.core_id {
-                    a.sharing_lvl += 1;
-                }
-                if a.l2_id == b.l2_id {
-                    a.sharing_lvl += 1;
-                }
-                if a.l3_id == b.l3_id {
-                    a.sharing_lvl += 1;
-                }
-            }
-        }
-
         // Convert a vector of RefCell to a vector of plain cpu_fids
         let mut cpu_fids2 = Vec::new();
         for cpu_fid in cpu_fids.iter() {
@@ -356,52 +411,32 @@ impl FlatTopology {
         let mut cpu_fids = cpu_fids2;
 
         // Sort the cpu_fids
-        match (prefer_smt_core, prefer_little_core) {
-            (true, false) => {
-                // Sort the cpu_fids by node, llc, ^cpu_cap, ^sharing_lvl, core, and cpu order
-                cpu_fids.sort_by(|a, b| {
-                    a.node_id
-                        .cmp(&b.node_id)
-                        .then_with(|| a.llc_pos.cmp(&b.llc_pos))
-                        .then_with(|| b.cpu_cap.cmp(&a.cpu_cap))
-                        .then_with(|| b.sharing_lvl.cmp(&a.sharing_lvl))
-                        .then_with(|| a.core_pos.cmp(&b.core_pos))
-                        .then_with(|| a.cpu_pos.cmp(&b.cpu_pos))
-                });
-            }
-            (true, true) => {
-                // Sort the cpu_fids by node, llc, cpu_cap, ^sharing_lvl, core, and cpu order
+        match prefer_powersave {
+            true => {
+                // Sort the cpu_fids by node, llc, cpu_cap, ^smt_level, ^cache_size, perf_dom, core, and cpu order
                 cpu_fids.sort_by(|a, b| {
                     a.node_id
                         .cmp(&b.node_id)
                         .then_with(|| a.llc_pos.cmp(&b.llc_pos))
                         .then_with(|| a.cpu_cap.cmp(&b.cpu_cap))
-                        .then_with(|| b.sharing_lvl.cmp(&a.sharing_lvl))
+                        .then_with(|| b.smt_level.cmp(&a.smt_level))
+                        .then_with(|| b.cache_size.cmp(&a.cache_size))
+                        .then_with(|| a.pd_id.cmp(&b.pd_id))
                         .then_with(|| a.core_pos.cmp(&b.core_pos))
                         .then_with(|| a.cpu_pos.cmp(&b.cpu_pos))
                 });
             }
-            (false, false) => {
-                // Sort the cpu_fids by cpu, node, llc, ^cpu_cap, sharing_lvl, and core order
+            false => {
+                // Sort the cpu_fids by cpu, node, llc, ^cpu_cap, smt_level, ^cache_size, perf_dom, and core order
                 cpu_fids.sort_by(|a, b| {
                     a.cpu_pos
                         .cmp(&b.cpu_pos)
                         .then_with(|| a.node_id.cmp(&b.node_id))
                         .then_with(|| a.llc_pos.cmp(&b.llc_pos))
                         .then_with(|| b.cpu_cap.cmp(&a.cpu_cap))
-                        .then_with(|| a.sharing_lvl.cmp(&b.sharing_lvl))
-                        .then_with(|| a.core_pos.cmp(&b.core_pos))
-                });
-            }
-            (false, true) => {
-                // Sort the cpu_fids by cpu, node, llc, cpu_cap, sharing_lvl, and core order
-                cpu_fids.sort_by(|a, b| {
-                    a.cpu_pos
-                        .cmp(&b.cpu_pos)
-                        .then_with(|| a.node_id.cmp(&b.node_id))
-                        .then_with(|| a.llc_pos.cmp(&b.llc_pos))
-                        .then_with(|| a.cpu_cap.cmp(&b.cpu_cap))
-                        .then_with(|| a.sharing_lvl.cmp(&b.sharing_lvl))
+                        .then_with(|| a.smt_level.cmp(&b.smt_level))
+                        .then_with(|| b.cache_size.cmp(&a.cache_size))
+                        .then_with(|| a.pd_id.cmp(&b.pd_id))
                         .then_with(|| a.core_pos.cmp(&b.core_pos))
                 });
             }
@@ -410,6 +445,15 @@ impl FlatTopology {
         Some((cpu_fids, avg_cap))
     }
 
+    /// Get the performance domain (i.e., CPU frequency domain) ID for a CPU.
+    /// If the energy model is not available, use NUMA node ID instead.
+    fn get_pd_id(em: &Result<EnergyModel>, cpu_id: usize, node_id: usize) -> usize {
+        match em {
+            Ok(em) => em.get_pd(cpu_id).unwrap().id,
+            Err(_) => node_id,
+        }
+    }
+
     /// Build a list of compute domains
     fn build_cpdom(
         cpu_fids: &Vec<CpuFlatId>,
@@ -552,7 +596,7 @@ impl<'a> Scheduler<'a> {
 
         // Initialize CPU topology
         let topo = FlatTopology::new().unwrap();
-        Self::init_cpus(&mut skel, &topo);
+        Self::init_cpus(&mut skel, &opts, &topo);
 
         // Initialize skel according to @opts.
         Self::init_globals(&mut skel, &opts, &topo);
@@ -585,18 +629,44 @@ impl<'a> Scheduler<'a> {
         })
     }
 
-    fn init_cpus(skel: &mut OpenBpfSkel, topo: &FlatTopology) {
-        // Initialize CPU order topologically sorted
-        // by a cpu, node, llc, max_freq, and core order
-        for (pos, cpu) in topo.cpu_fids_performance.iter().enumerate() {
-            skel.maps.rodata_data.cpu_order_performance[pos] = cpu.cpu_id as u16;
-            skel.maps.rodata_data.__cpu_capacity_hint[cpu.cpu_id] = cpu.cpu_cap as u16;
-        }
-        for (pos, cpu) in topo.cpu_fids_powersave.iter().enumerate() {
-            skel.maps.rodata_data.cpu_order_powersave[pos] = cpu.cpu_id as u16;
-        }
+    fn init_cpus(skel: &mut OpenBpfSkel, opts: &Opts, topo: &FlatTopology) {
         debug!("{:#?}", topo);
 
+        // Initialize CPU capacity
+        for (_, cpu) in topo.cpu_fids_performance.iter().enumerate() {
+            skel.maps.rodata_data.cpu_capacity[cpu.cpu_id] = cpu.cpu_cap as u16;
+        }
+
+        // If cpu_pref_order is not specified, initialize CPU order
+        // topologically sorted by a cpu, node, llc, max_freq, and core order.
+        // Otherwise, follow the specified CPU preference order.
+        let mut cpu_pf_order = vec![];
+        let mut cpu_ps_order = vec![];
+        if opts.cpu_pref_order == "" {
+            for cpu in topo.cpu_fids_performance.iter() {
+                cpu_pf_order.push(cpu.cpu_id);
+            }
+            for cpu in topo.cpu_fids_powersave.iter() {
+                cpu_ps_order.push(cpu.cpu_id);
+            }
+        } else {
+            let cpu_list = read_cpulist(&opts.cpu_pref_order).unwrap();
+            let pref_mask = Cpumask::from_cpulist(&opts.cpu_pref_order).unwrap();
+            if pref_mask != topo.all_cpus_mask {
+                panic!("--cpu_pref_order does not cover the whole CPUs.");
+            }
+            cpu_pf_order = cpu_list.clone();
+            cpu_ps_order = cpu_list.clone();
+        }
+        for (pos, cpu) in cpu_pf_order.iter().enumerate() {
+            skel.maps.rodata_data.cpu_order_performance[pos] = *cpu as u16;
+        }
+        for (pos, cpu) in cpu_ps_order.iter().enumerate() {
+            skel.maps.rodata_data.cpu_order_powersave[pos] = *cpu as u16;
+        }
+        info!("CPU pref order in performance mode: {:?}", cpu_pf_order);
+        info!("CPU pref order in powersave mode: {:?}", cpu_ps_order);
+
         // Initialize compute domain contexts
         for (k, v) in topo.cpdom_map.iter() {
             skel.maps.bss_data.cpdom_ctxs[v.cpdom_id].id = v.cpdom_id as u64;
@@ -627,15 +697,12 @@ impl<'a> Scheduler<'a> {
         }
     }
 
-    fn is_powersave_mode(opts: &Opts) -> bool {
-        opts.prefer_smt_core && opts.prefer_little_core
-    }
-
     fn init_globals(skel: &mut OpenBpfSkel, opts: &Opts, topo: &FlatTopology) {
+        skel.maps.bss_data.no_preemption = opts.no_preemption;
+        skel.maps.bss_data.no_wake_sync = opts.no_wake_sync;
         skel.maps.bss_data.no_core_compaction = opts.no_core_compaction;
         skel.maps.bss_data.no_freq_scaling = opts.no_freq_scaling;
-        skel.maps.bss_data.no_prefer_turbo_core = opts.no_prefer_turbo_core;
-        skel.maps.bss_data.is_powersave_mode = Self::is_powersave_mode(&opts);
+        skel.maps.bss_data.is_powersave_mode = opts.powersave;
         skel.maps.rodata_data.nr_cpu_ids = *NR_CPU_IDS as u64;
         skel.maps.rodata_data.is_smt_active = topo.smt_enabled;
         skel.maps.rodata_data.is_autopilot_on = opts.autopilot;
@@ -688,15 +755,15 @@ impl<'a> Scheduler<'a> {
             lat_cri: tc.lat_cri,
             avg_lat_cri: tx.avg_lat_cri,
             static_prio: tx.static_prio,
-            slice_boost_prio: tc.slice_boost_prio,
             run_freq: tc.run_freq,
-            run_time_ns: tc.run_time_ns,
+            avg_runtime: tc.avg_runtime,
             wait_freq: tc.wait_freq,
             wake_freq: tc.wake_freq,
             perf_cri: tc.perf_cri,
             thr_perf_cri: tx.thr_perf_cri,
             cpuperf_cur: tx.cpuperf_cur,
             cpu_util: tx.cpu_util,
+            cpu_sutil: tx.cpu_sutil,
             nr_active: tx.nr_active,
         }) {
             Ok(()) | Err(TrySendError::Full(_)) => 0,
@@ -740,12 +807,13 @@ impl<'a> Scheduler<'a> {
                 self.mseq_id += 1;
 
                 let bss_data = &self.skel.maps.bss_data;
-                let st = bss_data.__sys_stats[0];
+                let st = bss_data.sys_stat;
 
                 let mseq = self.mseq_id;
                 let nr_queued_task = st.nr_queued_task;
                 let nr_active = st.nr_active;
                 let nr_sched = st.nr_sched;
+                let nr_preempt = st.nr_preempt;
                 let pc_pc = Self::get_pc(st.nr_perf_cri, nr_sched);
                 let pc_lc = Self::get_pc(st.nr_lat_cri, nr_sched);
                 let pc_x_migration = Self::get_pc(st.nr_x_migration, nr_sched);
@@ -767,6 +835,7 @@ impl<'a> Scheduler<'a> {
                     nr_queued_task,
                     nr_active,
                     nr_sched,
+                    nr_preempt,
                     pc_pc,
                     pc_lc,
                     pc_x_migration,
@@ -943,7 +1012,7 @@ fn main() -> Result<()> {
     init_log(&opts);
 
     opts.proc().unwrap();
-    debug!("{:#?}", opts);
+    info!("{:#?}", opts);
 
     let shutdown = Arc::new(AtomicBool::new(false));
     let shutdown_clone = shutdown.clone();
