@@ -102,6 +102,7 @@ enum global_stat_id {
 	GSTAT_LO_FB_USAGE,
 	GSTAT_FB_CPU_USAGE,
 	GSTAT_ANTISTALL,
+	GSTAT_SKIP_PREEMPT,
 	NR_GSTATS,
 };
 
@@ -136,6 +137,7 @@ enum layer_stat_id {
 	LSTAT_XLAYER_REWAKE,
 	LSTAT_LLC_DRAIN_TRY,
 	LSTAT_LLC_DRAIN,
+	LSTAT_SKIP_REMOTE_NODE,
 	NR_LSTATS,
 };
 
@@ -248,6 +250,8 @@ enum layer_match_kind {
 	MATCH_IS_KTHREAD,
 	MATCH_USED_GPU_TID,
 	MATCH_USED_GPU_PID,
+	MATCH_AVG_RUNTIME,
+	MATCH_CGROUP_SUFFIX,
 
 	NR_LAYER_MATCH_KINDS,
 };
@@ -255,6 +259,7 @@ enum layer_match_kind {
 struct layer_match {
 	int		kind;
 	char		cgroup_prefix[MAX_PATH];
+	char		cgroup_suffix[MAX_PATH];
 	char		comm_prefix[MAX_COMM];
 	char		pcomm_prefix[MAX_COMM];
 	int		nice;
@@ -269,6 +274,8 @@ struct layer_match {
 	bool		used_gpu_tid;
 	bool		used_gpu_pid;
 	bool		exclude;
+	u64		min_avg_runtime_us;
+	u64		max_avg_runtime_us;
 };
 
 struct layer_match_ands {
@@ -288,10 +295,19 @@ enum layer_growth_algo {
 	GROWTH_ALGO_NODE_SPREAD,
 	GROWTH_ALGO_NODE_SPREAD_REVERSE,
 	GROWTH_ALGO_NODE_SPREAD_RANDOM,
+	GROWTH_ALGO_CPUSET_SPREAD,
+	GROWTH_ALGO_CPUSET_SPREAD_REVERSE,
+	GROWTH_ALGO_CPUSET_SPREAD_RANDOM,
 	GROWTH_ALGO_RANDOM_TOPO,
 	GROWTH_ALGO_STICKY_DYNAMIC,
 };
 
+enum layer_task_place {
+	PLACEMENT_STD,
+	PLACEMENT_STICK,
+	PLACEMENT_FLOAT,
+};
+
 struct layer {
 	struct layer_match_ands	matches[MAX_LAYER_MATCH_ORS];
 	unsigned int		nr_match_ors;
@@ -311,6 +327,7 @@ struct layer {
 	bool			preempt_first;
 	bool			exclusive;
 	bool			allow_node_aligned;
+	bool			skip_remote_node;
 	bool			prev_over_idle_core;
 	int			growth_algo;
 
@@ -329,9 +346,11 @@ struct layer {
 
 	u64			llcs_to_drain;
 	u32			llc_drain_cnt;
+	enum layer_task_place   task_place;
 
 	char			name[MAX_LAYER_NAME];
 	bool			is_protected;
+	bool			periodically_refresh;
 };
 
 struct scx_cmd {
