@@ -52,6 +52,7 @@ const volatile u64 min_open_layer_disallow_preempt_after_ns;
 const volatile u64 lo_fb_wait_ns = 5000000;	/* !0 for veristat */
 const volatile u32 lo_fb_share_ppk = 128;	/* !0 for veristat */
 const volatile bool percpu_kthread_preempt = true;
+volatile u64 layer_refresh_seq_avgruntime;
 
 /* Flag to enable or disable antistall feature */
 const volatile bool enable_antistall = true;
@@ -60,6 +61,13 @@ const volatile bool enable_gpu_support = false;
 const volatile u64 antistall_sec = 3;
 const u32 zero_u32 = 0;
 
+/*
+ * XXX sched classes should be exported kernel
+ * side to avoid having to do this...
+ */
+const volatile u64 ext_sched_class_addr = 0;
+const volatile u64 idle_sched_class_addr = 0;
+
 private(unprotected_cpumask) struct bpf_cpumask __kptr *unprotected_cpumask;
 u64 unprotected_seq = 0;
 
@@ -74,6 +82,18 @@ u32 nr_empty_layer_ids;
 
 UEI_DEFINE(uei);
 
+struct task_hint {
+	u64 hint;
+	u64 __reserved[3];
+};
+
+struct {
+	__uint(type, BPF_MAP_TYPE_TASK_STORAGE);
+	__uint(map_flags, BPF_F_NO_PREALLOC);
+	__type(key, int);
+	__type(value, struct task_hint);
+} scx_layered_task_hint_map SEC(".maps");
+
 static inline s32 prio_to_nice(s32 static_prio)
 {
 	/* See DEFAULT_PRIO and PRIO_TO_NICE in include/linux/sched/prio.h */
@@ -339,16 +359,11 @@ static void layer_llc_drain_disable(struct layer *layer, u32 llc_id)
 static inline bool refresh_layer_cpuc(struct cpu_ctx *cpuc, struct layer *layer)
 {
 	/* a CPU can be shared by multiple open layers */
-	if (layer->kind == LAYER_KIND_OPEN) {
-		cpuc->in_open_layers = true;
-		cpuc->layer_id = MAX_LAYERS;
-	} else {
-		cpuc->in_open_layers = false;
-		cpuc->layer_id = layer->id;
-	}
+	cpuc->in_open_layers = (layer->kind == LAYER_KIND_OPEN);
+	cpuc->layer_id = (layer->kind == LAYER_KIND_OPEN) ? MAX_LAYERS : layer->id;
 
 	if (cpuc->is_protected == layer->is_protected)
-		false;
+		return false;
 
 	cpuc->is_protected = layer->is_protected;
 	if (unlikely(!unprotected_cpumask)) {
@@ -404,8 +419,7 @@ static void refresh_cpumasks(u32 layer_id)
 
 		if ((u8_ptr = MEMBER_VPTR(layers, [layer_id].cpus[cpu / 8]))) {
 			if (*u8_ptr & (1 << (cpu % 8))) {
-				if (refresh_layer_cpuc(cpuc, layer))
-					protected_changed = true;
+				protected_changed = refresh_layer_cpuc(cpuc, layer) || protected_changed;
 
 				bpf_cpumask_set_cpu(cpu, layer_cpumask);
 			} else {
@@ -499,6 +513,7 @@ struct task_ctx {
 	u32			qrt_llc_id;
 
 	char 			join_layer[SCXCMD_COMLEN];
+	u64			layer_refresh_seq;
 };
 
 struct {
@@ -877,6 +892,7 @@ s32 pick_idle_cpu(struct task_struct *p, s32 prev_cpu,
 		  bool from_selcpu)
 {
 	const struct cpumask *idle_smtmask, *layer_cpumask, *layered_cpumask, *cpumask;
+	bool is_float = layer->task_place == PLACEMENT_FLOAT;
 	struct bpf_cpumask *unprot_mask;
 	struct cpu_ctx *prev_cpuc;
 	u32 layer_id = layer->id;
@@ -886,11 +902,14 @@ s32 pick_idle_cpu(struct task_struct *p, s32 prev_cpu,
 	if (layer_id >= MAX_LAYERS || !(layer_cpumask = lookup_layer_cpumask(layer_id)))
 		return -1;
 
+	if (layer->periodically_refresh && taskc->layer_refresh_seq < layer_refresh_seq_avgruntime)
+		taskc->refresh_layer = true;
+
 	/*
 	 * Not much to do if bound to a single CPU. Explicitly handle migration
 	 * disabled tasks for kernels before SCX_OPS_ENQ_MIGRATION_DISABLED.
 	 */
-	if (p->nr_cpus_allowed == 1 || is_migration_disabled(p)) {
+	if (!is_float && (p->nr_cpus_allowed == 1 || is_migration_disabled(p))) {
 		if (scx_bpf_test_and_clear_cpu_idle(prev_cpu)) {
 			if (layer->kind == LAYER_KIND_CONFINED &&
 			    !bpf_cpumask_test_cpu(prev_cpu, layer_cpumask))
@@ -958,6 +977,9 @@ s32 pick_idle_cpu(struct task_struct *p, s32 prev_cpu,
 	if (!(idle_smtmask = scx_bpf_get_idle_smtmask()))
 		return -1;
 
+	if (is_float)
+		goto no_locality;
+
 	/*
 	 * If the system has a big/little architecture and uses any related
 	 * layer growth algos try to find a cpu in that topology first.
@@ -989,6 +1011,7 @@ s32 pick_idle_cpu(struct task_struct *p, s32 prev_cpu,
 		}
 	}
 
+no_locality:
 	/*
 	 * Next try a CPU in the current node
 	 */
@@ -1085,7 +1108,11 @@ s32 BPF_STRUCT_OPS(layered_select_cpu, struct task_struct *p, s32 prev_cpu, u64
 	if (taskc->layer_id == MAX_LAYERS || !(layer = lookup_layer(taskc->layer_id)))
 		return prev_cpu;
 
-	cpu = pick_idle_cpu(p, prev_cpu, cpuc, taskc, layer, true);
+	if (layer->task_place == PLACEMENT_STICK)
+		cpu = prev_cpu;
+	else
+		cpu = pick_idle_cpu(p, prev_cpu, cpuc, taskc, layer, true);
+
 	if (cpu >= 0) {
 		lstat_inc(LSTAT_SEL_LOCAL, layer, cpuc);
 		taskc->dsq_id = SCX_DSQ_LOCAL;
@@ -1106,7 +1133,9 @@ static bool try_preempt_cpu(s32 cand, struct task_struct *p, struct task_ctx *ta
 			    struct layer *layer, bool preempt_first)
 {
 	struct cpu_ctx *cpuc, *cand_cpuc, *sib_cpuc = NULL;
+	struct rq *rq = NULL;
 	s32 sib;
+	struct sched_class *ext_sched_class, *idle_sched_class;
 
 	if (cand >= nr_possible_cpus || !bpf_cpumask_test_cpu(cand, p->cpus_ptr))
 		return false;
@@ -1117,6 +1146,20 @@ static bool try_preempt_cpu(s32 cand, struct task_struct *p, struct task_ctx *ta
 	if (cand_cpuc->current_preempt)
 		return false;
 
+	rq = scx_bpf_cpu_rq(cand);
+	
+	ext_sched_class = (struct sched_class *)(unsigned long long)ext_sched_class_addr;
+	idle_sched_class = (struct sched_class *)(unsigned long long)idle_sched_class_addr;
+
+	if (rq && ext_sched_class_addr && idle_sched_class_addr &&
+		(rq->curr->sched_class != ext_sched_class) &&
+		(rq->curr->sched_class != idle_sched_class)) {
+		if (!(cpuc = lookup_cpu_ctx(-1)))
+			return false;
+		gstat_inc(GSTAT_SKIP_PREEMPT, cpuc);
+		return false;
+	}
+
 	/*
 	 * Don't preempt if protection against is in effect. However, open
 	 * layers share CPUs and using the same mechanism between non-preempt
@@ -1694,12 +1737,16 @@ static __always_inline bool try_consume_layer(u32 layer_id, struct cpu_ctx *cpuc
 {
 	struct llc_prox_map *llc_pmap = &llcc->prox_map;
 	struct layer *layer;
+	u32 nid = llc_node_id(llcc->id);
 	bool xllc_mig_skipped = false;
+	bool skip_remote_node;
 	u32 u;
 
 	if (!(layer = lookup_layer(layer_id)))
 		return false;
 
+	skip_remote_node = layer->skip_remote_node;
+
 	bpf_for(u, 0, llc_pmap->sys_end) {
 		u16 *llc_idp;
 
@@ -1714,6 +1761,11 @@ static __always_inline bool try_consume_layer(u32 layer_id, struct cpu_ctx *cpuc
 			if (!(remote_llcc = lookup_llc_ctx(*llc_idp)))
 				return false;
 
+			if (skip_remote_node && nid != llc_node_id(remote_llcc->id)) {
+				lstat_inc(LSTAT_SKIP_REMOTE_NODE, layer, cpuc);
+				continue;
+			}
+
 			if (remote_llcc->queued_runtime[layer_id] < layer->xllc_mig_min_ns) {
 				xllc_mig_skipped = true;
 				continue;
@@ -1819,7 +1871,7 @@ void BPF_STRUCT_OPS(layered_dispatch, s32 cpu, struct task_struct *prev)
 	 * Low fallback DSQ execution is forced upto lo_fb_share_ppk fraction
 	 * after the DSQ had tasks queued for longer than lo_fb_wait_ns.
 	 */
-	if (!cpuc->is_protected && scx_bpf_dsq_nr_queued(cpuc->lo_fb_dsq_id)) {
+	if (scx_bpf_dsq_nr_queued(cpuc->lo_fb_dsq_id)) {
 		u64 now = scx_bpf_now();
 		u64 dur, usage;
 
@@ -1948,18 +2000,21 @@ static __noinline bool match_one(struct layer_match *match,
 
 	switch (match->kind) {
 	case MATCH_CGROUP_PREFIX: {
-		return match_prefix(match->cgroup_prefix, cgrp_path, MAX_PATH);
+		return match_prefix_suffix(match->cgroup_prefix, cgrp_path, false);
+	}
+	case MATCH_CGROUP_SUFFIX: {
+		return match_prefix_suffix(match->cgroup_suffix, cgrp_path, true);
 	}
 	case MATCH_COMM_PREFIX: {
 		char comm[MAX_COMM];
 		__builtin_memcpy(comm, p->comm, MAX_COMM);
-		return match_prefix(match->comm_prefix, comm, MAX_COMM);
+		return match_prefix_suffix(match->comm_prefix, comm, false);
 	}
 	case MATCH_PCOMM_PREFIX: {
 		char pcomm[MAX_COMM];
 
 		__builtin_memcpy(pcomm, p->group_leader->comm, MAX_COMM);
-		return match_prefix(match->pcomm_prefix, pcomm, MAX_COMM);
+		return match_prefix_suffix(match->pcomm_prefix, pcomm, false);
 	}
 	case MATCH_NICE_ABOVE:
 		return prio_to_nice((s32)p->static_prio) > match->nice;
@@ -2025,8 +2080,8 @@ static __noinline bool match_one(struct layer_match *match,
 		if (!taskc->join_layer[0])
 			return false;
 
-		return match_prefix(match->comm_prefix, taskc->join_layer,
-			SCXCMD_COMLEN);
+		return match_prefix_suffix(match->comm_prefix, taskc->join_layer,
+			false);
 	}
 	case MATCH_IS_GROUP_LEADER: {
 		// There is nuance to this around exec(2)s and group leader swaps.
@@ -2063,6 +2118,24 @@ static __noinline bool match_one(struct layer_match *match,
 
 			return pid_present == match->used_gpu_pid;
 	}
+	case MATCH_AVG_RUNTIME: {
+			struct task_ctx *taskc = lookup_task_ctx_may_fail(p);
+			if (!taskc) {
+				scx_bpf_error("could not find task");
+				return false;
+			}
+
+			u64 avg_runtime_us = taskc->runtime_avg / 1000;
+
+			if (!taskc) {
+				scx_bpf_error("could not find task");
+				return false;
+			}
+
+			/* To match, we must get min <= time < max. */
+			return match->min_avg_runtime_us <= avg_runtime_us &&
+				avg_runtime_us < match->max_avg_runtime_us;
+	}
 
 	default:
 		scx_bpf_error("invalid match kind %d", match->kind);
@@ -2070,26 +2143,20 @@ static __noinline bool match_one(struct layer_match *match,
 	}
 }
 
-int match_layer(u32 layer_id, pid_t pid, const char *cgrp_path)
+int match_layer(u32 layer_id, struct task_struct *p __arg_trusted, const char *cgrp_path)
 {
-
-	struct task_struct *p;
 	struct layer *layer;
 	u32 nr_match_ors;
 	u64 or_id, and_id;
 
-	p = bpf_task_from_pid(pid);
-	if (!p)
-		return -EINVAL;
-
 	if (layer_id >= nr_layers)
-		goto err;
+		return -EINVAL;
 
 	layer = &layers[layer_id];
 	nr_match_ors = layer->nr_match_ors;
 
 	if (nr_match_ors > MAX_LAYER_MATCH_ORS)
-		goto err;
+		return -EINVAL;
 
 	bpf_for(or_id, 0, nr_match_ors) {
 		struct layer_match_ands *ands;
@@ -2097,19 +2164,19 @@ int match_layer(u32 layer_id, pid_t pid, const char *cgrp_path)
 
 		barrier_var(or_id);
 		if (or_id >= MAX_LAYER_MATCH_ORS)
-			goto err;
+			return -EINVAL;
 
 		ands = &layer->matches[or_id];
 
 		if (ands->nr_match_ands > NR_LAYER_MATCH_KINDS)
-			goto err;
+			return -EINVAL;
 
 		bpf_for(and_id, 0, ands->nr_match_ands) {
 			struct layer_match *match;
 
 			barrier_var(and_id);
 			if (and_id >= NR_LAYER_MATCH_KINDS)
-				goto err;
+				return -EINVAL;
 
 			match = &ands->matches[and_id];
 			if (!(match_one(match, p, cgrp_path) == !match->exclude)) {
@@ -2119,29 +2186,23 @@ int match_layer(u32 layer_id, pid_t pid, const char *cgrp_path)
 		}
 
 		if (matched) {
-			bpf_task_release(p);
 			return 0;
 		}
 	}
 
-	bpf_task_release(p);
 	return -ENOENT;
-
-err:
-	bpf_task_release(p);
-	return -EINVAL;
 }
 
-static void maybe_refresh_layer(struct task_struct *p, struct task_ctx *taskc)
+static void maybe_refresh_layer(struct task_struct *p __arg_trusted, struct task_ctx *taskc)
 {
 	const char *cgrp_path;
 	bool matched = false;
 	u64 layer_id;	// XXX - int makes verifier unhappy
-	pid_t pid = p->pid;
 
 	if (!taskc->refresh_layer)
 		return;
 	taskc->refresh_layer = false;
+	taskc->layer_refresh_seq = layer_refresh_seq_avgruntime;
 
 	if (!(cgrp_path = format_cgrp_path(p->cgroups->dfl_cgrp)))
 		return;
@@ -2150,7 +2211,7 @@ static void maybe_refresh_layer(struct task_struct *p, struct task_ctx *taskc)
 		__sync_fetch_and_add(&layers[taskc->layer_id].nr_tasks, -1);
 
 	bpf_for(layer_id, 0, nr_layers) {
-		if (match_layer(layer_id, pid, cgrp_path) == 0) {
+		if (match_layer(layer_id, p, cgrp_path) == 0) {
 			matched = true;
 			break;
 		}
@@ -2454,6 +2515,7 @@ void BPF_STRUCT_OPS(layered_stopping, struct task_struct *p, bool runnable)
 	struct cpu_ctx *cpuc;
 	struct task_ctx *taskc;
 	struct layer *task_layer;
+	struct task_hint *task_hint;
 	u64 now = scx_bpf_now();
 	u64 usage_since_idle;
 	s32 task_lid;
@@ -2524,7 +2586,17 @@ void BPF_STRUCT_OPS(layered_stopping, struct task_struct *p, bool runnable)
 
 	if (cpuc->yielding && runtime < task_layer->slice_ns)
 		runtime = task_layer->slice_ns;
-	p->scx.dsq_vtime += runtime * 100 / p->scx.weight;
+
+	runtime = runtime * 100 / p->scx.weight;
+
+	task_hint = bpf_task_storage_get(&scx_layered_task_hint_map, p, NULL, 0);
+	if (task_hint) {
+		u64 hint = task_hint->hint ?: 1;
+		hint = hint < 1024 ? hint : 1024;
+		runtime = (runtime * hint) / 1024;
+	}
+
+	p->scx.dsq_vtime += runtime;
 	cpuc->maybe_idle = true;
 }
 
@@ -2789,40 +2861,52 @@ void BPF_STRUCT_OPS(layered_disable, struct task_struct *p)
 	task_uncharge_qrt(taskc);
 }
 
-static u64 dsq_first_runnable_for_ms(u64 dsq_id, u64 now)
+static s64 dsq_first_runnable_at_ms(u64 dsq_id, u64 now)
 {
 	struct task_struct *p;
 
 	bpf_for_each(scx_dsq, p, dsq_id, 0) {
 		struct task_ctx *taskc;
 
-		if ((taskc = lookup_task_ctx(p)))
-			return (now - taskc->runnable_at) / 1000000;
+		if ((taskc = lookup_task_ctx(p))) {
+			u64 runnable_at = taskc->runnable_at;
+
+			if (runnable_at >= now)
+				return ((taskc->runnable_at - now) / 1000000);
+			else
+				return -((now - taskc->runnable_at) / 1000000);
+		}
 	}
 
 	return 0;
 }
 
+__hidden void dump_cpumask_word(s32 word, struct cpumask *cpumask)
+{
+	u32 u, v = 0;
+
+	bpf_for(u, 0, 32) {
+		s32 cpu = 32 * word + u;
+		if (cpu < nr_cpu_ids &&
+		    bpf_cpumask_test_cpu(cpu, cpumask))
+			v |= 1 << u;
+	}
+	scx_bpf_dump("%08x", v);
+}
+
 static void dump_layer_cpumask(int id)
 {
 	struct cpumask *layer_cpumask;
-	s32 cpu;
-	char buf[128] = "", *p;
+	u32 word, nr_words = (nr_cpu_ids + 31) / 32;
 
 	if (!(layer_cpumask = lookup_layer_cpumask(id)))
 		return;
 
-	bpf_for(cpu, 0, scx_bpf_nr_cpu_ids()) {
-		if (!(p = MEMBER_VPTR(buf, [cpu])))
-			break;
-		if (bpf_cpumask_test_cpu(cpu, layer_cpumask))
-			*p = '0' + cpu % 10;
-		else
-			*p = '.';
+	bpf_for(word, 0, nr_words) {
+		if (word)
+			scx_bpf_dump(",");
+		dump_cpumask_word(nr_words - word - 1, layer_cpumask);
 	}
-	buf[sizeof(buf) - 1] = '\0';
-
-	scx_bpf_dump("%s", buf);
 }
 
 void BPF_STRUCT_OPS(layered_dump, struct scx_dump_ctx *dctx)
@@ -2846,24 +2930,24 @@ void BPF_STRUCT_OPS(layered_dump, struct scx_dump_ctx *dctx)
 				continue;
 
 			dsq_id = layer_dsq_id(layer->id, j);
-			scx_bpf_dump("LAYER[%d][%s]DSQ[%llx] nr_cpus=%u nr_queued=%d -%llums cpus=",
+			scx_bpf_dump("LAYER[%d](%s)-DSQ[%llx] nr_cpus=%u nr_queued=%d %+lldms\n",
 				     i, layer->name, dsq_id, layer->nr_cpus,
 				     scx_bpf_dsq_nr_queued(dsq_id),
-				     dsq_first_runnable_for_ms(dsq_id, now));
-			scx_bpf_dump("\n");
+				     dsq_first_runnable_at_ms(dsq_id, now));
 		}
+		scx_bpf_dump("LAYER[%d](%s) CPUS=", i, layer->name);
 		dump_layer_cpumask(i);
 		scx_bpf_dump("\n");
 	}
 	bpf_for(i, 0, nr_llcs) {
 		dsq_id = hi_fb_dsq_id(i);
-		scx_bpf_dump("HI_[%llx] nr_queued=%d -%llums\n",
+		scx_bpf_dump("HI_[%llx] nr_queued=%d %+lldms\n",
 			     dsq_id, scx_bpf_dsq_nr_queued(dsq_id),
-			     dsq_first_runnable_for_ms(dsq_id, now));
+			     dsq_first_runnable_at_ms(dsq_id, now));
 		dsq_id = lo_fb_dsq_id(i);
-		scx_bpf_dump("LO_FALLBACK[%llx] nr_queued=%d -%llums\n",
+		scx_bpf_dump("LO_FALLBACK[%llx] nr_queued=%d %+lldms\n",
 			     dsq_id, scx_bpf_dsq_nr_queued(dsq_id),
-			     dsq_first_runnable_for_ms(dsq_id, now));
+			     dsq_first_runnable_at_ms(dsq_id, now));
 	}
 }
 
@@ -3123,6 +3207,14 @@ static s32 init_layer(int layer_id)
 			case MATCH_USED_GPU_PID:
 				dbg("%s GPU_PID %d", header, match->used_gpu_pid);
 				break;
+			case MATCH_AVG_RUNTIME:
+				layer->periodically_refresh = true;
+				dbg("%s AVG_RUNTIME [%lluus, %lluus)", header,
+					match->min_avg_runtime_us,
+					match->max_avg_runtime_us);
+			case MATCH_CGROUP_SUFFIX:
+				dbg("%s CGROUP_SUFFIX \"%s\"", header, match->cgroup_suffix);
+				break;
 			default:
 				scx_bpf_error("%s Invalid kind", header);
 				return -EINVAL;
