@@ -58,9 +58,18 @@ impl LayerSpec {
     }
 }
 
+#[derive(Clone, Debug, Default, Serialize, Deserialize)]
+pub enum LayerPlacement {
+    #[default]
+    Standard,
+    Sticky,
+    Floating,
+}
+
 #[derive(Clone, Debug, Serialize, Deserialize)]
 pub enum LayerMatch {
     CgroupPrefix(String),
+    CgroupSuffix(String),
     CommPrefix(String),
     CommPrefixExclude(String),
     PcommPrefix(String),
@@ -80,6 +89,7 @@ pub enum LayerMatch {
     IsKthread(bool),
     UsedGpuTid(bool),
     UsedGpuPid(bool),
+    AvgRuntime(u64, u64),
 }
 
 #[derive(Clone, Debug, Serialize, Deserialize)]
@@ -101,6 +111,8 @@ pub struct LayerCommon {
     #[serde(default)]
     pub allow_node_aligned: bool,
     #[serde(default)]
+    pub skip_remote_node: bool,
+    #[serde(default)]
     pub prev_over_idle_core: bool,
     #[serde(default)]
     pub weight: u32,
@@ -122,6 +134,8 @@ pub struct LayerCommon {
     pub nodes: Vec<usize>,
     #[serde(default)]
     pub llcs: Vec<usize>,
+    #[serde(default)]
+    pub placement: LayerPlacement,
 }
 
 #[derive(Clone, Debug, Serialize, Deserialize)]
