@@ -38,6 +38,7 @@ const GSTAT_LO_FB_EVENTS: usize = bpf_intf::global_stat_id_GSTAT_LO_FB_EVENTS as
 const GSTAT_LO_FB_USAGE: usize = bpf_intf::global_stat_id_GSTAT_LO_FB_USAGE as usize;
 const GSTAT_FB_CPU_USAGE: usize = bpf_intf::global_stat_id_GSTAT_FB_CPU_USAGE as usize;
 const GSTAT_ANTISTALL: usize = bpf_intf::global_stat_id_GSTAT_ANTISTALL as usize;
+const GSTAT_SKIP_PREEMPT: usize = bpf_intf::global_stat_id_GSTAT_SKIP_PREEMPT as usize;
 
 const LSTAT_SEL_LOCAL: usize = bpf_intf::layer_stat_id_LSTAT_SEL_LOCAL as usize;
 const LSTAT_ENQ_LOCAL: usize = bpf_intf::layer_stat_id_LSTAT_ENQ_LOCAL as usize;
@@ -69,6 +70,7 @@ const LSTAT_XLAYER_WAKE: usize = bpf_intf::layer_stat_id_LSTAT_XLAYER_WAKE as us
 const LSTAT_XLAYER_REWAKE: usize = bpf_intf::layer_stat_id_LSTAT_XLAYER_REWAKE as usize;
 const LSTAT_LLC_DRAIN_TRY: usize = bpf_intf::layer_stat_id_LSTAT_LLC_DRAIN_TRY as usize;
 const LSTAT_LLC_DRAIN: usize = bpf_intf::layer_stat_id_LSTAT_LLC_DRAIN as usize;
+const LSTAT_SKIP_REMOTE_NODE: usize = bpf_intf::layer_stat_id_LSTAT_SKIP_REMOTE_NODE as usize;
 
 const LLC_LSTAT_LAT: usize = bpf_intf::llc_layer_stat_id_LLC_LSTAT_LAT as usize;
 const LLC_LSTAT_CNT: usize = bpf_intf::llc_layer_stat_id_LLC_LSTAT_CNT as usize;
@@ -183,6 +185,8 @@ pub struct LayerStats {
     pub llc_drain_try: f64,
     #[stat(desc = "% LLC draining succeeded")]
     pub llc_drain: f64,
+    #[stat(desc = "% skip LLC dispatch on remote node")]
+    pub skip_remote_node: f64,
     #[stat(desc = "mask of allocated CPUs", _om_skip)]
     pub cpus: Vec<u64>,
     #[stat(desc = "count of CPUs assigned")]
@@ -275,6 +279,7 @@ impl LayerStats {
             xllc_migration_skip: lstat_pct(LSTAT_XLLC_MIGRATION_SKIP),
             llc_drain_try: lstat_pct(LSTAT_LLC_DRAIN_TRY),
             llc_drain: lstat_pct(LSTAT_LLC_DRAIN),
+            skip_remote_node: lstat_pct(LSTAT_SKIP_REMOTE_NODE),
             cpus: layer.cpus.as_raw_slice().to_vec(),
             cur_nr_cpus: layer.cpus.weight() as u32,
             min_nr_cpus: nr_cpus_range.0 as u32,
@@ -366,12 +371,13 @@ impl LayerStats {
 
         writeln!(
             w,
-            "  {:<width$}  xlayer_wake/re={}/{} llc_drain/try={}/{}",
+            "  {:<width$}  xlayer_wake/re={}/{} llc_drain/try={}/{} skip_rnode={}",
             "",
             fmt_pct(self.xlayer_wake),
             fmt_pct(self.xlayer_rewake),
             fmt_pct(self.llc_drain),
             fmt_pct(self.llc_drain_try),
+            fmt_pct(self.skip_remote_node),
             width = header_width,
         )?;
 
@@ -488,6 +494,8 @@ pub struct SysStats {
     pub lo_fb_util: f64,
     #[stat(desc = "Number of tasks dispatched via antistall")]
     pub antistall: u64,
+    #[stat(desc = "Number of times preemptions of non-scx tasks were avoided")]
+    pub skip_preempt: u64,
     #[stat(desc = "fallback CPU")]
     pub fallback_cpu: u32,
     #[stat(desc = "per-layer statistics")]
@@ -543,6 +551,7 @@ impl SysStats {
             lo_fb_util: stats.bpf_stats.gstats[GSTAT_LO_FB_USAGE] as f64 / elapsed_ns as f64
                 * 100.0,
             antistall: stats.bpf_stats.gstats[GSTAT_ANTISTALL],
+            skip_preempt: stats.bpf_stats.gstats[GSTAT_SKIP_PREEMPT],
             fallback_cpu: fallback_cpu as u32,
             fallback_cpu_util: stats.bpf_stats.gstats[GSTAT_FB_CPU_USAGE] as f64
                 / elapsed_ns as f64
@@ -554,7 +563,7 @@ impl SysStats {
     pub fn format<W: Write>(&self, w: &mut W) -> Result<()> {
         writeln!(
             w,
-            "tot={:7} local_sel/enq={}/{} open_idle={} affn_viol={} hi/lo={}/{}",
+            "tot={:7} local_sel/enq={}/{} open_idle={} affn_viol={} hi/lo={}/{} skip_preempt={}",
             self.total,
             fmt_pct(self.local_sel),
             fmt_pct(self.local_enq),
@@ -562,6 +571,7 @@ impl SysStats {
             fmt_pct(self.affn_viol),
             fmt_pct(self.hi_fb),
             fmt_pct(self.lo_fb),
+            self.skip_preempt,
         )?;
 
         writeln!(
