@@ -10,15 +10,17 @@
 #ifdef LSP
 #define __bpf__
 #include "../../../../include/scx/common.bpf.h"
+#include "../../../../include/scx/bpf_arena_common.h"
+#include "../../../../include/lib/sdt_task.h"
 #else
 #include <scx/common.bpf.h>
-#endif
-
-#ifndef P2DQ_CREATE_STRUCT_OPS
-#define P2DQ_CREATE_STRUCT_OPS 1
+#include <scx/bpf_arena_common.h>
+#include <lib/sdt_task.h>
 #endif
 
 #include "intf.h"
+#include "types.h"
+
 
 #include <errno.h>
 #include <stdbool.h>
@@ -26,6 +28,10 @@
 #include <bpf/bpf_helpers.h>
 #include <bpf/bpf_tracing.h>
 
+#ifndef P2DQ_CREATE_STRUCT_OPS
+#define P2DQ_CREATE_STRUCT_OPS 1
+#endif
+
 char _license[] SEC("license") = "GPL";
 
 UEI_DEFINE(uei);
@@ -51,28 +57,36 @@ const volatile u32 min_nr_queued_pick2 = 10;
 const volatile bool autoslice = true;
 const volatile bool dispatch_pick2_disable = false;
 const volatile bool eager_load_balance = true;
-const volatile bool greedy_idle = false;
 const volatile bool interactive_sticky = false;
 const volatile bool keep_running_enabled = true;
 const volatile bool kthreads_local = true;
 const volatile bool max_dsq_pick2 = false;
 const volatile bool select_idle_in_enqueue = true;
+const volatile u64 max_exec_ns = 20 * NSEC_PER_MSEC;
+
+const volatile bool dispatch_lb_interactive = false;
+const volatile u64 dispatch_lb_busy = 75;
+const volatile u64 wakeup_lb_busy = 90;
+const volatile bool wakeup_llc_migrations = false;
+const volatile u64 lb_slack_factor = 5;
 
 const volatile bool smt_enabled = true;
 const volatile bool has_little_cores = false;
 const volatile u32 debug = 2;
 
 const u32 zero_u32 = 0;
+
+const u64 lb_timer_intvl_ns = 250LLU * NSEC_PER_MSEC;
+const u64 lb_backoff_ns = 5LLU * NSEC_PER_MSEC;
+
 u64 cpu_llc_ids[MAX_CPUS];
 u64 cpu_node_ids[MAX_CPUS];
 u64 big_core_ids[MAX_CPUS];
 u64 dsq_time_slices[MAX_DSQS_PER_LLC];
 
-u64 max_exec_ns;
 u64 min_slice_ns = 500;
 u32 sched_mode = MODE_PERFORMANCE;
 
-
 private(A) struct bpf_cpumask __kptr *all_cpumask;
 private(A) struct bpf_cpumask __kptr *big_cpumask;
 
@@ -82,17 +96,22 @@ static u64 max(u64 a, u64 b)
 	return a >= b ? a : b;
 }
 
-static __always_inline u64 dsq_time_slice(int dsq_id)
+static __always_inline u64 dsq_time_slice(int dsq_index)
 {
-	if (dsq_id > nr_dsqs_per_llc || dsq_id < 0) {
-		scx_bpf_error("Invalid DSQ id");
+	if (dsq_index > nr_dsqs_per_llc || dsq_index < 0) {
+		scx_bpf_error("Invalid DSQ index");
 		return 0;
 	}
-	return dsq_time_slices[dsq_id];
+	return dsq_time_slices[dsq_index];
+}
+
+static __always_inline u64 task_slice_ns(struct task_struct *p, int dsq_index)
+{
+	return p->scx.weight * dsq_time_slice(dsq_index) / 100;
 }
 
 struct p2dq_timer p2dq_timers[MAX_TIMERS] = {
-	{250LLU * NSEC_PER_MSEC, CLOCK_BOOTTIME, 0},
+	{lb_timer_intvl_ns, CLOCK_BOOTTIME, 0},
 };
 
 struct timer_wrapper {
@@ -119,10 +138,11 @@ static struct cpu_ctx *lookup_cpu_ctx(int cpu)
 {
 	struct cpu_ctx *cpuc;
 
-	if (cpu < 0)
+	if (cpu < 0) {
 		cpuc = bpf_map_lookup_elem(&cpu_ctxs, &zero_u32);
-	else
+	} else {
 		cpuc = bpf_map_lookup_percpu_elem(&cpu_ctxs, &zero_u32, cpu);
+	}
 
 	if (!cpuc) {
 		scx_bpf_error("no cpu_ctx for cpu %d", cpu);
@@ -186,22 +206,20 @@ static struct node_ctx *lookup_node_ctx(u32 node_id)
 	return nodec;
 }
 
+struct mask_wrapper {
+	struct bpf_cpumask __kptr *mask;
+};
 
 struct {
 	__uint(type, BPF_MAP_TYPE_TASK_STORAGE);
 	__uint(map_flags, BPF_F_NO_PREALLOC);
 	__type(key, int);
-	__type(value, struct task_ctx);
-} task_ctxs SEC(".maps");
+	__type(value, struct mask_wrapper);
+} task_masks SEC(".maps");
 
-static struct task_ctx *lookup_task_ctx_may_fail(struct task_struct *p)
+static task_ctx *lookup_task_ctx(struct task_struct *p)
 {
-	return bpf_task_storage_get(&task_ctxs, p, 0, 0);
-}
-
-static struct task_ctx *lookup_task_ctx(struct task_struct *p)
-{
-	struct task_ctx *taskc = lookup_task_ctx_may_fail(p);
+	task_ctx *taskc = scx_task_data(p);
 
 	if (!taskc)
 		scx_bpf_error("task_ctx lookup failed");
@@ -232,7 +250,7 @@ static inline void stat_inc(enum stat_idx idx)
 /*
  * Returns if the task is interactive based on the tasks DSQ index.
  */
-static bool is_interactive(struct task_ctx *taskc)
+static bool is_interactive(task_ctx *taskc)
 {
 	if (nr_dsqs_per_llc <= 1)
 		return false;
@@ -243,11 +261,12 @@ static bool is_interactive(struct task_ctx *taskc)
 /*
  * Returns if the task is able to load balance using pick2.
  */
-static bool can_pick2(struct task_ctx *taskc)
+static bool can_pick2(task_ctx *taskc)
 {
 	if (is_interactive(taskc) ||
 	    !taskc->all_cpus ||
 	    taskc->is_kworker ||
+	    nr_llcs == 2 ||
 	    (max_dsq_pick2 > 0 && taskc->llc_runs < min_llc_runs_pick2))
 		return false;
 
@@ -286,7 +305,7 @@ static bool keep_running(struct cpu_ctx *cpuc, struct task_struct *p)
 		return false;
 
 
-	u64 slice_ns = dsq_time_slice(cpuc->dsq_index);
+	u64 slice_ns = task_slice_ns(p, cpuc->dsq_index);
 	cpuc->ran_for += slice_ns;
 	p->scx.slice = slice_ns;
 	stat_inc(P2DQ_STAT_KEEP);
@@ -296,44 +315,51 @@ static bool keep_running(struct cpu_ctx *cpuc, struct task_struct *p)
 static struct llc_ctx *pick_two_llc_ctx(struct llc_ctx *cur_llcx, struct llc_ctx *left,
 					struct llc_ctx *right)
 {
-	s32 cur_queued = 0, left_queued = 0, right_queued = 0;
-	u64 cur_load = 0, left_load = 0, right_load = 0;
+	s32 cur_queued = 0;
+	u64 left_load = 0, right_load = 0;
 	int i;
 
 	if (!left || !right)
 		return NULL;
 
+	u64 now = scx_bpf_now();
+	if (now - cur_llcx->last_period_ns < lb_backoff_ns)
+		return NULL;
+
+	u64 max_possible_load = (now - cur_llcx->last_period_ns) * cur_llcx->nr_cpus;
+	u64 cur_load = cur_llcx->load;
+	u64 scaled_load = (100 * cur_load) / max_possible_load;
+
+	// If over the load balancing utilization busy watermark don't load
+	// balance.
+	if (scaled_load > dispatch_lb_busy)
+		return NULL;
+
 	bpf_for(i, 0, nr_llcs) {
 		if (i >= nr_dsqs_per_llc || i < 0)
 			continue;
 
 		u64 cur_dsq_id = *MEMBER_VPTR(cur_llcx->dsqs, [i]);
 		cur_queued += scx_bpf_dsq_nr_queued(cur_dsq_id);
-		cur_load += *MEMBER_VPTR(cur_llcx->dsq_load, [i]);
-		u64 left_dsq_id = *MEMBER_VPTR(left->dsqs, [i]);
-		left_queued += scx_bpf_dsq_nr_queued(left_dsq_id);
-		left_load += *MEMBER_VPTR(left->dsq_load, [i]);
-		u64 right_dsq_id = *MEMBER_VPTR(right->dsqs, [i]);
-		right_queued += scx_bpf_dsq_nr_queued(right_dsq_id);
-		right_load += *MEMBER_VPTR(right->dsq_load, [i]);
 	}
 
-	if (min_nr_queued_pick2 > 0 &&
-	    (left_queued < min_nr_queued_pick2 &&
-	     right_queued < min_nr_queued_pick2))
+	if (min_nr_queued_pick2 > 0 && cur_queued < min_nr_queued_pick2)
 		return NULL;
 
+	left_load = left->load;
+	right_load = right->load;
+
 	// If the current LLCs has more load don't try to pick2.
-	if ((nr_llcs > 2 && cur_queued >= left_queued && cur_queued >= right_queued) ||
-	    (nr_llcs > 2 && cur_load > left_load && cur_load > right_load))
+	cur_load += (lb_slack_factor * cur_load) / 100;
+	if ((nr_llcs > 2 && (cur_load > left_load || cur_load > right_load)))
 	    return NULL;
 
-        if (left_queued < right_queued || left_load < right_load)
+        if (left_load < right_load)
 		return right;
 	return left;
 }
 
-static s32 pick_two_cpu(struct llc_ctx *cur_llcx, struct task_ctx *taskc,
+static s32 pick_two_cpu(struct llc_ctx *cur_llcx, task_ctx *taskc,
 			bool *is_idle)
 {
 	if ((min_llc_runs_pick2 > 0 &&
@@ -343,51 +369,48 @@ static s32 pick_two_cpu(struct llc_ctx *cur_llcx, struct task_ctx *taskc,
 
 	struct llc_ctx *chosen;
 	struct llc_ctx *left, *right;
-	int dsq_index = taskc->dsq_index;
 	s32 cpu;
 
-	if (nr_llcs == 2) {
-		left = lookup_llc_ctx(0);
-		right = lookup_llc_ctx(1);
-	} else {
-		left = rand_llc_ctx();
-		right = rand_llc_ctx();
-	}
+	u64 now = scx_bpf_now();
+	// TODO: Use a moving avg instead for load calculations. The current
+	// method will be too noisy.
+	if (now - cur_llcx->last_period_ns < lb_backoff_ns)
+		return -EINVAL;
 
-	if (!left || !right || dsq_index > nr_dsqs_per_llc) {
+	u64 max_possible_load = (now - cur_llcx->last_period_ns) * cur_llcx->nr_cpus;
+	u64 cur_load = cur_llcx->load;
+	u64 scaled_load = (100 * cur_load) / max_possible_load;
+
+	// If the current LLC is not heavily utilized then don't load balance,
+	// under saturation load balancing is done on the dispatch path.
+	if (scaled_load < wakeup_lb_busy)
+		return -EINVAL;
+
+	left = rand_llc_ctx();
+	right = rand_llc_ctx();
+
+	if (!left || !right) {
 		return -EINVAL;
 	}
 
 	// last ditch effort if same are picked.
-	if (left->id == right->id) {
+	if (unlikely(left->id == right->id)) {
 		right = rand_llc_ctx();
 		if (!right || left->id == right->id)
 			return -EINVAL;
 	}
 
-	u64 cur_dsq_id = *MEMBER_VPTR(cur_llcx->dsqs, [dsq_index]);
-	s32 cur_queued = scx_bpf_dsq_nr_queued(cur_dsq_id);
-	u64 cur_load = *MEMBER_VPTR(cur_llcx->dsq_load, [dsq_index]);
-	u64 left_dsq_id = *MEMBER_VPTR(left->dsqs, [dsq_index]);
-	s32 left_queued = scx_bpf_dsq_nr_queued(left_dsq_id);
-	u64 left_load = *MEMBER_VPTR(left->dsq_load, [dsq_index]);
-	u64 right_dsq_id = *MEMBER_VPTR(right->dsqs, [dsq_index]);
-	s32 right_queued = scx_bpf_dsq_nr_queued(right_dsq_id);
-	u64 right_load = *MEMBER_VPTR(right->dsq_load, [dsq_index]);
+	u64 left_load = left->load;
+	u64 right_load = right->load;
 
 	// If the other LLCs have more load than the current don't bother.
-	if ((nr_llcs > 2 && left_queued >= cur_queued && right_queued >= cur_queued) ||
-	    (nr_llcs > 2 && left_load > cur_load && right_load > cur_load))
+	u64 slack_factor = (1 * cur_load) / 100;
+	if (slack_factor > 0)
+		cur_load += slack_factor;
+	if (left_load > cur_load && right_load > cur_load)
 		return -EINVAL;
 
-	if (min_nr_queued_pick2 > 0 &&
-	    (left_queued < min_nr_queued_pick2 &&
-	     right_queued < min_nr_queued_pick2))
-		return -EINVAL;
-
-	if (((left_queued > 0 || right_queued > 0) ||
-	    left_queued < right_queued) ||
-	    left_load < right_load) {
+	if (left_load < right_load) {
 		chosen = left;
 		goto pick_llc;
 	} else {
@@ -400,7 +423,8 @@ pick_llc:
 		return -EINVAL;
 
 	// First try to find an idle core
-	cpu = scx_bpf_pick_idle_cpu(cast_mask(chosen->cpumask), SCX_PICK_IDLE_CORE);
+	cpu = scx_bpf_pick_idle_cpu(cast_mask(chosen->cpumask),
+				    SCX_PICK_IDLE_CORE);
 	if (cpu >= 0) {
 		*is_idle = true;
 		return cpu;
@@ -422,15 +446,16 @@ pick_llc:
 	return -EINVAL;
 }
 
-static s32 pick_idle_cpu(struct task_struct *p, struct task_ctx *taskc,
+static s32 pick_idle_cpu(struct task_struct *p, task_ctx *taskc,
 			 s32 prev_cpu, u64 wake_flags, bool *is_idle)
 {
 	const struct cpumask *idle_smtmask, *idle_cpumask;
+	struct mask_wrapper *wrapper;
 	struct cpu_ctx *prev_cpuc;
+	struct bpf_cpumask *mask;
 	struct node_ctx *nodec;
 	struct llc_ctx *llcx;
 	bool interactive = is_interactive(taskc);
-	int llc_idle = 0;
 	s32 cpu = prev_cpu;
 
 	idle_cpumask = scx_bpf_get_idle_cpumask();
@@ -439,34 +464,45 @@ static s32 pick_idle_cpu(struct task_struct *p, struct task_ctx *taskc,
 	if (!idle_cpumask || !idle_smtmask)
 		goto found_cpu;
 
+	if (interactive_sticky && interactive) {
+		cpu = prev_cpu;
+		*is_idle = scx_bpf_test_and_clear_cpu_idle(prev_cpu);
+		goto found_cpu;
+	}
+
 	if (!(prev_cpuc = lookup_cpu_ctx(prev_cpu)) ||
 	    !(llcx = lookup_llc_ctx(prev_cpuc->llc_id)) ||
 	    !(nodec = lookup_node_ctx(prev_cpuc->node_id)) ||
-	    !llcx->cpumask || !llcx->tmp_cpumask)
+	    !llcx->cpumask)
 		goto found_cpu;
 
-	bpf_cpumask_and(llcx->tmp_cpumask, idle_cpumask, cast_mask(llcx->cpumask));
-	int nr_idle = bpf_cpumask_weight(idle_cpumask);
-	int nr_queued = scx_bpf_dsq_nr_queued(taskc->dsq_id);
-	if (llcx->tmp_cpumask)
-		llc_idle = bpf_cpumask_weight(cast_mask(llcx->tmp_cpumask));
-
-
 	// Special handling of tasks with custom affinities
 	if (!taskc->all_cpus) {
+		wrapper = bpf_task_storage_get(&task_masks, p, 0, 0);
+		if (!wrapper) {
+			cpu = prev_cpu;
+			goto found_cpu;
+		}
+
+		mask = wrapper->mask;
+		if (!mask) {
+			cpu = prev_cpu;
+			goto found_cpu;
+		}
+
 		// First try last CPU
 		if (bpf_cpumask_test_cpu(prev_cpu, p->cpus_ptr)) {
 			cpu = prev_cpu;
 			goto found_cpu;
 		}
 
-		if (taskc->mask && llcx->cpumask)
-			bpf_cpumask_and(taskc->mask, cast_mask(llcx->cpumask),
+		if (llcx->cpumask)
+			bpf_cpumask_and(mask, cast_mask(llcx->cpumask),
 					p->cpus_ptr);
 
 		// First try to find an idle SMT in the LLC
-		if (smt_enabled && taskc->mask) {
-			cpu = scx_bpf_pick_idle_cpu(cast_mask(taskc->mask),
+		if (smt_enabled) {
+			cpu = scx_bpf_pick_idle_cpu(cast_mask(mask),
 						    SCX_PICK_IDLE_CORE);
 			if (cpu >= 0) {
 				*is_idle = true;
@@ -475,20 +511,27 @@ static s32 pick_idle_cpu(struct task_struct *p, struct task_ctx *taskc,
 		}
 
 		// Next try to find an idle CPU in the LLC
-		if (taskc->mask) {
-			cpu = scx_bpf_pick_idle_cpu(cast_mask(taskc->mask), 0);
-			if (cpu >= 0) {
-				*is_idle = true;
-				goto found_cpu;
-			}
+		cpu = scx_bpf_pick_idle_cpu(cast_mask(mask), 0);
+		if (cpu >= 0) {
+			*is_idle = true;
+			goto found_cpu;
 		}
 
+		// Keep the task sticky to the LLC if possible.
+		if (mask && llcx->cpumask &&
+		    bpf_cpumask_and(mask, cast_mask(llcx->cpumask),
+				    p->cpus_ptr)) {
+			cpu = bpf_cpumask_any_distribute(cast_mask(mask));
+			if (cpu < nr_cpus)
+				goto found_cpu;
+		}
+
+
 		// Next try to find an idle CPU in the node
-		if (nodec->cpumask && taskc->mask) {
-			bpf_cpumask_and(taskc->mask, cast_mask(nodec->cpumask),
+		if (nodec->cpumask && mask) {
+			bpf_cpumask_and(mask, cast_mask(nodec->cpumask),
 					p->cpus_ptr);
-			if (taskc->mask &&
-			    (cpu = scx_bpf_pick_idle_cpu(cast_mask(taskc->mask), 0)) >= 0) {
+			if ((cpu = scx_bpf_pick_idle_cpu(cast_mask(mask), 0)) >= 0) {
 				*is_idle = true;
 				goto found_cpu;
 			}
@@ -507,53 +550,87 @@ static s32 pick_idle_cpu(struct task_struct *p, struct task_ctx *taskc,
 	 */
 	if (wake_flags & SCX_WAKE_SYNC) {
 		struct task_struct *current = (void *)bpf_get_current_task_btf();
-		struct task_ctx *cur_taskc = lookup_task_ctx_may_fail(current);
-		if (cur_taskc) {
-			if (cur_taskc->llc_id == llcx->id) {
-				// First check if the waking task is in the same LLC
-				// and the prev cpu is idle
-				if (scx_bpf_test_and_clear_cpu_idle(cpu)) {
-					stat_inc(P2DQ_STAT_WAKE_PREV);
-					*is_idle = true;
-					goto found_cpu;
-				}
-				// Try an idle core in the LLC.
-				if (llcx->cpumask &&
-				    (cpu = scx_bpf_pick_idle_cpu(cast_mask(llcx->cpumask),
-								 SCX_PICK_IDLE_CORE)) >= 0) {
-					stat_inc(P2DQ_STAT_WAKE_LLC);
-					*is_idle = true;
-					goto found_cpu;
-				}
-			} else {
-				// If not in the local LLC we need to migrate to the waker llc.
-				struct llc_ctx *cur_llcx = lookup_llc_ctx(cur_taskc->llc_id);
-				if (!cur_llcx)
-					goto found_cpu;
-
-				if (cur_llcx->cpumask &&
-				    (cpu = scx_bpf_pick_idle_cpu(cast_mask(cur_llcx->cpumask),
-								 SCX_PICK_IDLE_CORE)) >= 0) {
-					stat_inc(P2DQ_STAT_WAKE_MIG);
-					*is_idle = true;
-					goto found_cpu;
-				}
-
-				// Couldn't find an idle core so just migrate to the CPU
-				if (cur_llcx->cpumask &&
-				    (cpu = scx_bpf_pick_idle_cpu(cast_mask(cur_llcx->cpumask),
-								 0)) >= 0) {
-					stat_inc(P2DQ_STAT_WAKE_MIG);
-					*is_idle = true;
-					goto found_cpu;
-				}
-			}
+		task_ctx *cur_taskc = scx_task_data(current);
+		// Shouldn't happen, but makes code easier to follow
+		if (!cur_taskc) {
+			cpu = prev_cpu;
+			goto found_cpu;
 		}
+		// Interactive tasks aren't worth migrating across LLCs.
+		if (interactive) {
+			cpu = prev_cpu;
+			if (scx_bpf_test_and_clear_cpu_idle(cpu)) {
+				stat_inc(P2DQ_STAT_WAKE_PREV);
+				*is_idle = true;
+				goto found_cpu;
+			}
+			// Try an idle CPU in the LLC.
+			if (llcx->cpumask &&
+			    (cpu = scx_bpf_pick_idle_cpu(cast_mask(llcx->cpumask), 0)) >= 0) {
+				stat_inc(P2DQ_STAT_WAKE_LLC);
+				*is_idle = true;
+				goto found_cpu;
+			}
+			// Nothing idle, stay sticky
+			cpu = prev_cpu;
+			goto found_cpu;
+		}
+		if (cur_taskc->llc_id == llcx->id || !wakeup_llc_migrations) {
+			// First check if the waking task is in the same LLC
+			// and the prev cpu is idle
+			if (scx_bpf_test_and_clear_cpu_idle(prev_cpu)) {
+				cpu = prev_cpu;
+				stat_inc(P2DQ_STAT_WAKE_PREV);
+				*is_idle = true;
+				goto found_cpu;
+			}
+			// Try an idle core in the LLC.
+			if (llcx->cpumask &&
+			    (cpu = scx_bpf_pick_idle_cpu(cast_mask(llcx->cpumask),
+							 SCX_PICK_IDLE_CORE)) >= 0) {
+				stat_inc(P2DQ_STAT_WAKE_LLC);
+				*is_idle = true;
+				goto found_cpu;
+			}
+			// Try an idle core in the LLC.
+			if (llcx->cpumask &&
+			    (cpu = scx_bpf_pick_idle_cpu(cast_mask(llcx->cpumask),
+							 0)) >= 0) {
+				stat_inc(P2DQ_STAT_WAKE_LLC);
+				*is_idle = true;
+				goto found_cpu;
+			}
+			// Nothing idle, stay sticky
+			cpu = prev_cpu;
+			goto found_cpu;
+		}
+		// If wakeup LLC are allowed then migrate to the waker llc.
+		struct llc_ctx *cur_llcx = lookup_llc_ctx(cur_taskc->llc_id);
+		if (!cur_llcx)
+			goto found_cpu;
+
+		if (cur_llcx->cpumask &&
+		    (cpu = scx_bpf_pick_idle_cpu(cast_mask(cur_llcx->cpumask),
+						 SCX_PICK_IDLE_CORE)) >= 0) {
+			stat_inc(P2DQ_STAT_WAKE_MIG);
+			*is_idle = true;
+			goto found_cpu;
+		}
+
+		// Couldn't find an idle core so just migrate to the CPU
+		if (cur_llcx->cpumask &&
+		    (cpu = scx_bpf_pick_idle_cpu(cast_mask(cur_llcx->cpumask),
+						 0)) >= 0) {
+			stat_inc(P2DQ_STAT_WAKE_MIG);
+			*is_idle = true;
+			goto found_cpu;
+		}
+		// Nothing idle, move to waker CPU
+		cpu = cur_taskc->cpu;
+		goto found_cpu;
 	}
 
-	if (nr_llcs > 1 &&
-	    (eager_load_balance ||
-	     (llc_idle == 0 && nr_queued > llcx->nr_cpus && nr_idle > 0))) {
+	if (eager_load_balance && wakeup_lb_busy > 0 && nr_llcs > 1) {
 		cpu = pick_two_cpu(llcx, taskc, is_idle);
 		if (cpu >= 0) {
 			stat_inc(P2DQ_STAT_SELECT_PICK2);
@@ -561,33 +638,34 @@ static s32 pick_idle_cpu(struct task_struct *p, struct task_ctx *taskc,
 		}
 	}
 
-	if (has_little_cores) {
-		if (llcx->big_cpumask &&
-		    (cpu = scx_bpf_pick_idle_cpu(cast_mask(llcx->big_cpumask),
-						 SCX_PICK_IDLE_CORE)) >= 0) {
-			*is_idle = true;
-			goto found_cpu;
-		}
-
-		if (nodec->big_cpumask &&
-		    (cpu = scx_bpf_pick_idle_cpu(cast_mask(nodec->big_cpumask),
-						 SCX_PICK_IDLE_CORE)) >= 0) {
-			*is_idle = true;
-			goto found_cpu;
-		}
-	}
-
 	// First check if last CPU is idle
 	if (llcx->cpumask &&
 	    bpf_cpumask_test_cpu(prev_cpu, cast_mask(llcx->cpumask)) &&
-	    bpf_cpumask_test_cpu(prev_cpu, smt_enabled ? idle_smtmask : idle_cpumask)) {
+	    bpf_cpumask_test_cpu(prev_cpu, (smt_enabled && !interactive) ? idle_smtmask : idle_cpumask)) {
 		cpu = prev_cpu;
 		*is_idle = true;
 		goto found_cpu;
 	}
 
+	if (has_little_cores && llcx->little_cpumask && llcx->big_cpumask) {
+		if (interactive) {
+			if ((cpu = scx_bpf_pick_idle_cpu(cast_mask(llcx->little_cpumask),
+							 0)) >= 0) {
+				*is_idle = true;
+				goto found_cpu;
+			}
+		} else {
+			if ((cpu = scx_bpf_pick_idle_cpu(cast_mask(llcx->big_cpumask),
+							 SCX_PICK_IDLE_CORE)) >= 0) {
+				*is_idle = true;
+				goto found_cpu;
+			}
+		}
+	}
+
 	// Next try in the local LLC
-	if (llcx->cpumask &&
+	if (!interactive &&
+	    llcx->cpumask &&
 	    (cpu = scx_bpf_pick_idle_cpu(cast_mask(llcx->cpumask),
 					 SCX_PICK_IDLE_CORE)) >= 0) {
 		*is_idle = true;
@@ -604,15 +682,18 @@ static s32 pick_idle_cpu(struct task_struct *p, struct task_ctx *taskc,
 	// Non-interactive tasks load balance
 	if (nr_llcs > 1 &&
 	    !interactive &&
-	    nr_idle > 1 &&
+	    wakeup_lb_busy > 0 &&
 	    (cpu = pick_two_cpu(llcx, taskc, is_idle)) >= 0) {
 		stat_inc(P2DQ_STAT_SELECT_PICK2);
 		goto found_cpu;
 	}
 
 	// Couldn't find anything idle just return something in the local LLC
-	if (llcx->cpumask)
+	if (interactive && llcx->cpumask)
 		cpu = bpf_cpumask_any_distribute(cast_mask(llcx->cpumask));
+	else
+		// non interactive tasks stay sticky
+		cpu = prev_cpu;
 
 found_cpu:
 	scx_bpf_put_cpumask(idle_cpumask);
@@ -626,7 +707,7 @@ found_cpu:
 
 static __always_inline s32 p2dq_select_cpu_impl(struct task_struct *p, s32 prev_cpu, u64 wake_flags)
 {
-	struct task_ctx *taskc;
+	task_ctx *taskc;
 	bool is_idle = false;
 	s32 cpu;
 
@@ -636,8 +717,7 @@ static __always_inline s32 p2dq_select_cpu_impl(struct task_struct *p, s32 prev_
 	cpu = pick_idle_cpu(p, taskc, prev_cpu, wake_flags, &is_idle);
 	if (is_idle) {
 		stat_inc(P2DQ_STAT_IDLE);
-		u64 slice_ns = dsq_time_slice(taskc->dsq_index);
-		scx_bpf_dsq_insert(p, SCX_DSQ_LOCAL, slice_ns, 0);
+		scx_bpf_dsq_insert(p, SCX_DSQ_LOCAL, taskc->slice_ns, 0);
 	}
 
 	return cpu;
@@ -660,7 +740,7 @@ static __always_inline void async_p2dq_enqueue(struct enqueue_promise *ret,
 {
 	struct llc_ctx *llcx, *prev_llcx;
 	struct cpu_ctx *cpuc, *task_cpuc;
-	struct task_ctx *taskc;
+	task_ctx *taskc;
 	u64 dsq_id;
 
 	s32 cpu, task_cpu = scx_bpf_task_cpu(p);
@@ -674,7 +754,7 @@ static __always_inline void async_p2dq_enqueue(struct enqueue_promise *ret,
 	}
 
 	u64 vtime_now = llcx->vtime;
-	u64 slice_ns = dsq_time_slice(taskc->dsq_index);
+	taskc->slice_ns = task_slice_ns(p, taskc->dsq_index);
 
 	// If the task in in another LLC need to update vtime.
 	if (taskc->llc_id != cpuc->llc_id) {
@@ -692,11 +772,11 @@ static __always_inline void async_p2dq_enqueue(struct enqueue_promise *ret,
 	u64 vtime = p->scx.dsq_vtime;
 
 	/*
-	 * Limit the amount of budget that an idling task can accumulate
-	 * to one slice for the dsq.
+	 * Limit the amount of budget that an idling task can accumulate to the
+	 * max possible slice.
 	 */
-	if (time_before(vtime, vtime_now - slice_ns))
-		vtime = vtime_now - slice_ns;
+	if (time_before(vtime, vtime_now - dsq_time_slice(nr_dsqs_per_llc - 1)))
+		vtime = vtime_now - taskc->slice_ns;
 
 	p->scx.dsq_vtime = vtime;
 
@@ -706,38 +786,26 @@ static __always_inline void async_p2dq_enqueue(struct enqueue_promise *ret,
 	 * behind other threads which is necessary for forward progress
 	 * guarantee as we depend on the BPF timer which may run from ksoftirqd.
 	 */
-	if ((p->flags & PF_KTHREAD) && (p->nr_cpus_allowed < nr_cpus) &&
+	if ((p->flags & PF_KTHREAD) && !taskc->all_cpus &&
 	    kthreads_local) {
 		stat_inc(P2DQ_STAT_DIRECT);
-		scx_bpf_dsq_insert(p, SCX_DSQ_LOCAL, slice_ns,
+		scx_bpf_dsq_insert(p, SCX_DSQ_LOCAL, taskc->slice_ns,
 				   enq_flags | SCX_ENQ_PREEMPT);
 
 		ret->kind = P2DQ_ENQUEUE_PROMISE_COMPLETE;
 		return;
 	}
 
-	/*
-	 * Affinitized tasks just get dispatched directly, need to handle this
-	 * better.
-	 */
-	if (!taskc->all_cpus && bpf_cpumask_test_cpu(cpuc->id, p->cpus_ptr)) {
-		stat_inc(P2DQ_STAT_DIRECT);
-
-		ret->kind = P2DQ_ENQUEUE_PROMISE_FIFO;
-		ret->fifo.dsq_id = SCX_DSQ_LOCAL;
-		ret->fifo.enq_flags = enq_flags;
-		ret->fifo.slice_ns = slice_ns;
-		return;
-	}
-
 	// If an idle CPU hasn't been found in select_cpu find one now
-	if (select_idle_in_enqueue && !__COMPAT_is_enq_cpu_selected(enq_flags)) {
+	if ((select_idle_in_enqueue && !__COMPAT_is_enq_cpu_selected(enq_flags)) ||
+	    !taskc->all_cpus) {
 		bool is_idle = false;
 		cpu = pick_idle_cpu(p, taskc, taskc->cpu, 0, &is_idle);
 		cpuc = lookup_cpu_ctx(cpu);
 		if (cpuc && taskc->dsq_index >= 0 && taskc->dsq_index < nr_dsqs_per_llc) {
 			dsq_id = cpu_dsq_id(taskc->dsq_index, cpuc);
-			scx_bpf_dsq_insert_vtime(p, dsq_id, slice_ns, vtime, enq_flags);
+			taskc->dsq_id = dsq_id;
+			scx_bpf_dsq_insert_vtime(p, dsq_id, taskc->slice_ns, p->scx.dsq_vtime, enq_flags);
 			if (is_idle) {
 				stat_inc(P2DQ_STAT_IDLE);
 				scx_bpf_kick_cpu(cpu, SCX_KICK_IDLE);
@@ -749,12 +817,13 @@ static __always_inline void async_p2dq_enqueue(struct enqueue_promise *ret,
 	}
 
 	dsq_id = cpu_dsq_id(taskc->dsq_index, cpuc);
+	taskc->dsq_id = dsq_id;
 
 	ret->kind = P2DQ_ENQUEUE_PROMISE_VTIME;
 	ret->vtime.dsq_id = dsq_id;
 	ret->vtime.enq_flags = enq_flags;
-	ret->vtime.slice_ns = slice_ns;
-	ret->vtime.vtime = vtime;
+	ret->vtime.slice_ns = taskc->slice_ns;
+	ret->vtime.vtime = p->scx.dsq_vtime;
 }
 
 static __always_inline void complete_p2dq_enqueue(struct enqueue_promise *pro,
@@ -778,18 +847,18 @@ out:
 
 static __always_inline void p2dq_runnable_impl(struct task_struct *p, u64 enq_flags)
 {
-	struct task_ctx *wakee_ctx;
+	task_ctx *wakee_ctx;
 
 	if (!(wakee_ctx = lookup_task_ctx(p)))
 		return;
 
-	wakee_ctx->is_kworker = p->flags & PF_WQ_WORKER;
+	wakee_ctx->is_kworker = p->flags & (PF_KTHREAD | PF_WQ_WORKER | PF_IO_WORKER);
 }
 
 
 void BPF_STRUCT_OPS(p2dq_running, struct task_struct *p)
 {
-	struct task_ctx *taskc;
+	task_ctx *taskc;
 	struct cpu_ctx *cpuc;
 	struct llc_ctx *llcx;
 	s32 task_cpu = scx_bpf_task_cpu(p);
@@ -812,43 +881,37 @@ void BPF_STRUCT_OPS(p2dq_running, struct task_struct *p)
 		stat_inc(P2DQ_STAT_NODE_MIGRATION);
 	}
 
-	taskc->last_run_at = scx_bpf_now();
 	taskc->llc_id = llcx->id;
 	taskc->node_id = llcx->node_id;
 	cpuc->dsq_index = taskc->dsq_index;
 	cpuc->ran_for = 0;
 	taskc->cpu = task_cpu;
 	// racy, but don't care
-	if (p->scx.dsq_vtime > llcx->vtime)
+	if (p->scx.dsq_vtime > llcx->vtime) {
 		__sync_val_compare_and_swap(&llcx->vtime, llcx->vtime, p->scx.dsq_vtime);
-	else
+	} else {
 		p->scx.dsq_vtime = llcx->vtime;
+	}
 
 	// If the task is running in the least interactive DSQ, bump the
 	// frequency.
 	if (taskc->dsq_index == nr_dsqs_per_llc-1) {
 		scx_bpf_cpuperf_set(task_cpu, SCX_CPUPERF_ONE);
 	}
+	taskc->last_run_at = bpf_ktime_get_ns();
 }
 
-
 void BPF_STRUCT_OPS(p2dq_stopping, struct task_struct *p, bool runnable)
 {
-	struct task_ctx *taskc;
-	struct cpu_ctx *cpuc;
+	task_ctx *taskc;
 	struct llc_ctx *llcx;
 	u64 used, scaled_used, last_dsq_slice_ns;
-	u64 now = scx_bpf_now();
+	u64 now = bpf_ktime_get_ns();
 
-	if (!(taskc = lookup_task_ctx(p)))
+	if (!(taskc = lookup_task_ctx(p)) ||
+	    !(llcx = lookup_llc_ctx(taskc->llc_id)))
 		return;
 
-	if (!(cpuc = lookup_cpu_ctx(-1)) ||
-	    !(llcx = lookup_llc_ctx(cpuc->llc_id)))
-		return;
-
-	last_dsq_slice_ns = dsq_time_slice(taskc->dsq_index);
-
 	// can't happen, appease the verifier
 	int dsq_index = taskc->dsq_index;
 	if (dsq_index < 0 || dsq_index >= nr_dsqs_per_llc) {
@@ -856,14 +919,23 @@ void BPF_STRUCT_OPS(p2dq_stopping, struct task_struct *p, bool runnable)
 		return;
 	}
 
-	used = now - taskc->last_run_at;
-	scaled_used = scale_by_task_weight_inverse(p, used);
-	p->scx.dsq_vtime += scaled_used;
 	taskc->last_dsq_id = taskc->dsq_id;
-	taskc->last_dsq_index = dsq_index;
+	taskc->last_dsq_index = taskc->dsq_index;
+	taskc->used = 0;
+
+	last_dsq_slice_ns = taskc->slice_ns;
+	used = now - taskc->last_run_at;
+	scaled_used = used * 100 / p->scx.weight;
+
+	p->scx.dsq_vtime += scaled_used;
 	__sync_fetch_and_add(&llcx->vtime, scaled_used);
 	__sync_fetch_and_add(&llcx->dsq_max_vtime[dsq_index], scaled_used);
 	__sync_fetch_and_add(&llcx->dsq_load[dsq_index], used);
+	__sync_fetch_and_add(&llcx->load, used);
+
+
+	trace("%s weight %d slice %llu used %llu scaled %llu",
+	      p->comm, p->scx.weight, last_dsq_slice_ns, used, scaled_used);
 
 	// On stopping determine if the task can move to a longer DSQ by
 	// comparing the used time to the scaled DSQ slice.
@@ -871,8 +943,8 @@ void BPF_STRUCT_OPS(p2dq_stopping, struct task_struct *p, bool runnable)
 		if (taskc->dsq_index < nr_dsqs_per_llc - 1) {
 			taskc->dsq_index += 1;
 			stat_inc(P2DQ_STAT_DSQ_CHANGE);
-			trace("%s[%p]: DSQ change %u -> %u, slice %llu", p->comm, p,
-			      taskc->last_dsq_id, taskc->dsq_index, dsq_time_slice(taskc->dsq_index));
+			trace("%s[%p]: DSQ inc %llu -> %u", p->comm, p,
+			      taskc->last_dsq_index, taskc->dsq_index);
 		} else {
 			stat_inc(P2DQ_STAT_DSQ_SAME);
 		}
@@ -881,8 +953,8 @@ void BPF_STRUCT_OPS(p2dq_stopping, struct task_struct *p, bool runnable)
 		if (taskc->dsq_index > 0) {
 			taskc->dsq_index -= 1;
 			stat_inc(P2DQ_STAT_DSQ_CHANGE);
-			trace("%s[%p]: DSQ change %u -> %u slice %llu", p->comm, p,
-			      taskc->last_dsq_id, taskc->dsq_index, dsq_time_slice(taskc->dsq_index));
+			trace("%s[%p]: DSQ dec %llu -> %u", p->comm, p,
+			      taskc->last_dsq_index, taskc->dsq_index);
 		} else {
 			stat_inc(P2DQ_STAT_DSQ_SAME);
 		}
@@ -891,7 +963,7 @@ void BPF_STRUCT_OPS(p2dq_stopping, struct task_struct *p, bool runnable)
 	}
 }
 
-static int __always_inline dispatch_cpu(u64 dsq_id, s32 cpu, struct llc_ctx *llcx, int dsq_index)
+static __always_inline int dispatch_cpu(u64 dsq_id, s32 cpu, struct llc_ctx *llcx, int dsq_index)
 {
 	struct task_struct *p;
 	int dispatched = 0;
@@ -916,7 +988,7 @@ static int __always_inline dispatch_cpu(u64 dsq_id, s32 cpu, struct llc_ctx *llc
 			continue;
 		}
 
-		if (!__COMPAT_scx_bpf_dsq_move(BPF_FOR_EACH_ITER, p, SCX_DSQ_LOCAL_ON|cpu, SCX_ENQ_PREEMPT)) {
+		if (!__COMPAT_scx_bpf_dsq_move(BPF_FOR_EACH_ITER, p, SCX_DSQ_LOCAL_ON | cpu, 0)) {
 			bpf_task_release(p);
 			continue;
 		}
@@ -968,8 +1040,15 @@ static __always_inline int dispatch_pick_two(s32 cpu, struct llc_ctx *cur_llcx,
 		return 0;
 	}
 
-	// Start with least interactive DSQs to avoid migrating interactive
-	// tasks.
+	// First try any interactive tasks.
+	if (dispatch_lb_interactive) {
+		dsq_id = llcx->dsqs[0];
+		if (dispatch_cpu(dsq_id, cpu, llcx, 0) > 0)
+			return 0;
+	}
+
+	// Then migrate least interactive DSQs to find the most throughput
+	// bound tasks.
 	bpf_for(i, 1, nr_dsqs_per_llc) {
 		dsq_id = llcx->dsqs[nr_dsqs_per_llc - i];
 		if (dispatch_cpu(dsq_id, cpu, llcx, nr_dsqs_per_llc - i) > 0)
@@ -1027,7 +1106,7 @@ static __always_inline void p2dq_dispatch_impl(s32 cpu, struct task_struct *prev
 void BPF_STRUCT_OPS(p2dq_set_cpumask, struct task_struct *p,
 		    const struct cpumask *cpumask)
 {
-	struct task_ctx *taskc;
+	task_ctx *taskc;
 
 	if (!(taskc = lookup_task_ctx(p)) || !all_cpumask)
 		return;
@@ -1035,18 +1114,18 @@ void BPF_STRUCT_OPS(p2dq_set_cpumask, struct task_struct *p,
 	taskc->all_cpus = p->cpus_ptr == &p->cpus_mask && p->nr_cpus_allowed == nr_cpus;
 }
 
-s32 BPF_STRUCT_OPS_SLEEPABLE(p2dq_init_task, struct task_struct *p,
-			     struct scx_init_task_args *args)
+static __always_inline s32 p2dq_init_task_impl(struct task_struct *p,
+					       struct scx_init_task_args *args)
 {
+	struct mask_wrapper *wrapper;
 	struct bpf_cpumask *cpumask;
-	struct task_ctx *taskc;
+	task_ctx *taskc;
 	struct cpu_ctx *cpuc;
 	struct llc_ctx *llcx;
 
 	s32 task_cpu = scx_bpf_task_cpu(p);
 
-	taskc = bpf_task_storage_get(&task_ctxs, p, 0,
-				    BPF_LOCAL_STORAGE_GET_F_CREATE);
+	taskc = scx_task_alloc(p);
 	if (!taskc) {
 		scx_bpf_error("task_ctx allocation failure");
 		return -ENOMEM;
@@ -1061,7 +1140,15 @@ s32 BPF_STRUCT_OPS_SLEEPABLE(p2dq_init_task, struct task_struct *p,
 		return -ENOMEM;
 	}
 
-	if ((cpumask = bpf_kptr_xchg(&taskc->mask, cpumask))) {
+	wrapper = bpf_task_storage_get(&task_masks, p, 0,
+				    BPF_LOCAL_STORAGE_GET_F_CREATE);
+	if (!wrapper) {
+		bpf_cpumask_release(cpumask);
+		scx_bpf_error("task mask allocation failure");
+		return -ENOMEM;
+	}
+
+	if ((cpumask = bpf_kptr_xchg(&wrapper->mask, cpumask))) {
 		bpf_cpumask_release(cpumask);
 		scx_bpf_error("task_ctx allocation failure");
 		return -EINVAL;
@@ -1073,6 +1160,7 @@ s32 BPF_STRUCT_OPS_SLEEPABLE(p2dq_init_task, struct task_struct *p,
 	taskc->node_id = cpuc->node_id;
 	taskc->dsq_index = init_dsq_index;
 	taskc->last_dsq_index = init_dsq_index;
+	taskc->slice_ns = dsq_time_slice(init_dsq_index);
 	taskc->runnable = true;
 	taskc->all_cpus = p->cpus_ptr == &p->cpus_mask && p->nr_cpus_allowed == nr_cpus;
 	p->scx.dsq_vtime = llcx->vtime;
@@ -1080,10 +1168,14 @@ s32 BPF_STRUCT_OPS_SLEEPABLE(p2dq_init_task, struct task_struct *p,
 	return 0;
 }
 
+void BPF_STRUCT_OPS(p2dq_exit_task, struct task_struct *p, struct scx_exit_task_args *args)
+{
+	scx_task_free(p);
+}
 
 static int init_llc(u32 llc_id)
 {
-	struct bpf_cpumask *cpumask, *big_cpumask, *tmp_cpumask;
+	struct bpf_cpumask *cpumask, *big_cpumask, *little_cpumask;
 	struct llc_ctx *llcx;
 
 	llcx = bpf_map_lookup_elem(&llc_ctxs, &llc_id);
@@ -1125,16 +1217,16 @@ static int init_llc(u32 llc_id)
 		bpf_cpumask_release(big_cpumask);
 	}
 
-	tmp_cpumask = bpf_cpumask_create();
-	if (!tmp_cpumask) {
+	little_cpumask = bpf_cpumask_create();
+	if (!little_cpumask) {
 		scx_bpf_error("failed to create tmp cpumask");
 		return -ENOMEM;
 	}
 
-	tmp_cpumask = bpf_kptr_xchg(&llcx->tmp_cpumask, tmp_cpumask);
-	if (tmp_cpumask) {
+	little_cpumask = bpf_kptr_xchg(&llcx->little_cpumask, little_cpumask);
+	if (little_cpumask) {
 		scx_bpf_error("kptr already had cpumask");
-		bpf_cpumask_release(tmp_cpumask);
+		bpf_cpumask_release(little_cpumask);
 	}
 
 	return 0;
@@ -1224,6 +1316,10 @@ static s32 init_cpu(int cpu)
 			bpf_cpumask_set_cpu(cpu, llcx->big_cpumask);
 		bpf_rcu_read_unlock();
 	} else {
+		bpf_rcu_read_lock();
+		if (llcx->little_cpumask)
+			bpf_cpumask_set_cpu(cpu, llcx->little_cpumask);
+		bpf_rcu_read_unlock();
 		llcx->all_big = false;
 		nodec->all_big = false;
 	}
@@ -1298,6 +1394,8 @@ reset_load:
 		if (!(llcx = lookup_llc_ctx(llc_id)))
 			return false;
 
+		llcx->load = 0;
+		llcx->last_period_ns = scx_bpf_now();
 		bpf_for(j, 0, nr_dsqs_per_llc) {
 			llcx->dsq_load[j] = 0;
 			if (llc_id == 0 && autoslice) {
@@ -1471,16 +1569,18 @@ static __always_inline s32 p2dq_init_impl()
 		}
 	}
 
-	max_exec_ns = 10 * dsq_time_slice(nr_dsqs_per_llc);
 	min_slice_ns = 1000 * min_slice_us;
 
 	if (start_timers() < 0)
 		return -EINVAL;
 
+	ret = scx_task_init(sizeof(task_ctx));
+	if (ret)
+		return ret;
+
 	return 0;
 }
 
-
 void BPF_STRUCT_OPS(p2dq_exit, struct scx_exit_info *ei)
 {
 	UEI_RECORD(uei, ei);
@@ -1514,6 +1614,12 @@ s32 BPF_STRUCT_OPS(p2dq_select_cpu, struct task_struct *p, s32 prev_cpu, u64 wak
 	return p2dq_select_cpu_impl(p, prev_cpu, wake_flags);
 }
 
+s32 BPF_STRUCT_OPS_SLEEPABLE(p2dq_init_task, struct task_struct *p,
+			     struct scx_init_task_args *args)
+{
+	return p2dq_init_task_impl(p, args);
+}
+
 SCX_OPS_DEFINE(p2dq,
 	       .select_cpu		= (void *)p2dq_select_cpu,
 	       .enqueue			= (void *)p2dq_enqueue,
@@ -1523,6 +1629,7 @@ SCX_OPS_DEFINE(p2dq,
 	       .stopping		= (void *)p2dq_stopping,
 	       .set_cpumask		= (void *)p2dq_set_cpumask,
 	       .init_task		= (void *)p2dq_init_task,
+	       .exit_task		= (void *)p2dq_exit_task,
 	       .init			= (void *)p2dq_init,
 	       .exit			= (void *)p2dq_exit,
 	       .timeout_ms		= 20000,
