@@ -40,7 +40,8 @@ pub struct SchedulerOpts {
     #[clap(short = 'e', long, action = clap::ArgAction::SetTrue)]
     pub eager_load_balance: bool,
 
-    /// Disables greedy idle CPU selection, may cause better load balancing on multi-LLC systems.
+    /// ***DEPRECATED*** Disables greedy idle CPU selection, may cause better load balancing on
+    /// multi-LLC systems.
     #[clap(short = 'g', long, default_value_t = get_default_greedy_disable(), action = clap::ArgAction::Set)]
     pub greedy_idle_disable: bool,
 
@@ -52,10 +53,27 @@ pub struct SchedulerOpts {
     #[clap(short = 'd', long, action = clap::ArgAction::SetTrue)]
     pub dispatch_pick2_disable: bool,
 
+    /// Enables pick2 load balancing on the dispatch path when LLC utilization is under the
+    /// specified utilization.
+    #[clap(long, default_value = "75", value_parser = clap::value_parser!(u64).range(0..100))]
+    pub dispatch_lb_busy: u64,
+
+    /// Enables pick2 load balancing on the dispatch path for interactive tasks.
+    #[clap(long, action = clap::ArgAction::SetTrue)]
+    pub dispatch_lb_interactive: bool,
+
     /// Enable tasks to run beyond their timeslice if the CPU is idle.
     #[clap(long, action = clap::ArgAction::SetTrue)]
     pub keep_running: bool,
 
+    /// Minimum load for load balancing on the wakeup path, 0 to disable.
+    #[clap(long, default_value = "0", value_parser = clap::value_parser!(u64).range(0..99))]
+    pub wakeup_lb_busy: u64,
+
+    /// Allow LLC migrations on the wakeup path.
+    #[clap(long, action = clap::ArgAction::SetTrue)]
+    pub wakeup_llc_migrations: bool,
+
     /// Set idle QoS resume latency based in microseconds.
     #[clap(long)]
     pub idle_resume_us: Option<u32>,
@@ -68,6 +86,11 @@ pub struct SchedulerOpts {
     #[clap(short = 's', long, default_value = "100")]
     pub min_slice_us: u64,
 
+    /// Slack factor for load balancing, load balancing is not performed if load is within slack
+    /// factor percent.
+    #[clap(long, default_value = "5", value_parser = clap::value_parser!(u64).range(0..99))]
+    pub lb_slack_factor: u64,
+
     /// Number of runs on the LLC before a task becomes eligbile for pick2 migration on the wakeup
     /// path.
     #[clap(short = 'l', long, default_value_t = get_default_llc_runs())]
@@ -164,18 +187,24 @@ macro_rules! init_open_skel {
             $skel.maps.rodata_data.init_dsq_index = opts.init_dsq_index as i32;
             $skel.maps.rodata_data.nr_llcs = $crate::TOPO.all_llcs.clone().keys().len() as u32;
             $skel.maps.rodata_data.nr_nodes = $crate::TOPO.nodes.clone().keys().len() as u32;
+            $skel.maps.rodata_data.lb_slack_factor = opts.lb_slack_factor;
 
             $skel.maps.rodata_data.autoslice = opts.autoslice;
             $skel.maps.rodata_data.debug = verbose as u32;
             $skel.maps.rodata_data.dispatch_pick2_disable = opts.dispatch_pick2_disable;
+            $skel.maps.rodata_data.dispatch_lb_busy = opts.dispatch_lb_busy;
+            $skel.maps.rodata_data.dispatch_lb_interactive = opts.dispatch_lb_interactive;
             $skel.maps.rodata_data.eager_load_balance = !opts.eager_load_balance;
-            $skel.maps.rodata_data.greedy_idle = !opts.greedy_idle_disable;
             $skel.maps.rodata_data.has_little_cores = $crate::TOPO.has_little_cores();
             $skel.maps.rodata_data.interactive_sticky = opts.interactive_sticky;
             $skel.maps.rodata_data.keep_running_enabled = opts.keep_running;
             $skel.maps.rodata_data.max_dsq_pick2 = opts.max_dsq_pick2;
             $skel.maps.rodata_data.smt_enabled = $crate::TOPO.smt_enabled;
             $skel.maps.rodata_data.select_idle_in_enqueue = true;
+            $skel.maps.rodata_data.wakeup_lb_busy = opts.wakeup_lb_busy;
+            $skel.maps.rodata_data.wakeup_llc_migrations = opts.wakeup_llc_migrations;
+            $skel.maps.rodata_data.max_exec_ns =
+                2 * $skel.maps.bss_data.dsq_time_slices[opts.dumb_queues - 1];
 
             Ok(())
         }
@@ -197,3 +226,27 @@ macro_rules! init_skel {
         }
     };
 }
+
+pub mod bpf_srcs {
+
+    pub fn intf_h() -> &'static [u8] {
+        const INTF_H: &'static [u8] =
+            include_bytes!(concat!(env!("CARGO_MANIFEST_DIR"), "/src/bpf/intf.h"));
+
+        INTF_H
+    }
+
+    pub fn main_bpf_c() -> &'static [u8] {
+        const MAIN_BPF_C: &'static [u8] =
+            include_bytes!(concat!(env!("CARGO_MANIFEST_DIR"), "/src/bpf/main.bpf.c"));
+
+        MAIN_BPF_C
+    }
+
+    pub fn types_h() -> &'static [u8] {
+        const TYPES_H: &'static [u8] =
+            include_bytes!(concat!(env!("CARGO_MANIFEST_DIR"), "/src/bpf/types.h"));
+
+        TYPES_H
+    }
+}
