@@ -33,9 +33,6 @@ const SCHEDULER_NAME: &'static str = "RustLand";
 
 const VERSION: &'static str = env!("CARGO_PKG_VERSION");
 
-/// Task is being enqueued as a result of a wakeup event.
-const SCX_ENQ_WAKEUP: u64 = 1;
-
 /// scx_rustland: user-space scheduler written in Rust
 ///
 /// scx_rustland is designed to prioritize interactive workloads over background CPU-intensive
@@ -45,24 +42,26 @@ const SCX_ENQ_WAKEUP: u64 = 1;
 /// scx_rustland is also designed to be an "easy to read" template that can be used by any
 /// developer to quickly experiment more complex scheduling policies fully implemented in Rust.
 ///
-/// The scheduler is made of a BPF component (scx_rustland_core) that implements the low level
-/// sched-ext functionalities and a user-space counterpart (scheduler), written in Rust, that
-/// implements the actual scheduling policy.
+/// The scheduler is based on scx_rustland_core, which implements the low level sched-ext
+/// functionalities.
 ///
-/// The default scheduling policy implemented in the user-space scheduler is a based on virtual
-/// runtime (vruntime):
+/// The scheduling policy implemented in user-space is a based on a deadline, evaluated as
+/// following:
 ///
-/// - each task receives the same time slice of execution (slice_ns)
+///       deadline = vruntime + exec_runtime
 ///
-/// - the actual execution time, adjusted based on the task's static priority (weight), determines
-///   the vruntime
+/// Where, vruntime reflects the task's total runtime scaled by weight (ensuring fairness), while
+/// exec_runtime accounts the CPU time used since the last sleep (capturing responsiveness). Tasks
+/// are then dispatched from the lowest to the highest deadline.
 ///
-/// - tasks are then dispatched from the lowest to the highest vruntime
+/// This approach favors latency-sensitive tasks: those that frequently sleep will accumulate less
+/// exec_runtime, resulting in earlier deadlines. In contrast, CPU-intensive tasks that don’t sleep
+/// accumulate a larger exec_runtime and thus get scheduled later.
 ///
-/// All the tasks are stored in a BTreeSet (TaskTree), using vruntime as the ordering key.
-/// Once the order of execution is determined all tasks are sent back to the BPF counterpart to be
-/// dispatched. To keep track of the accumulated cputime and vruntime the scheduler maintain a
-/// HashMap (TaskInfoMap) indexed by pid.
+/// All the tasks are stored in a BTreeSet (TaskTree), using the deadline as the ordering key.
+/// Once the order of execution is determined all tasks are sent back to the BPF counterpart
+/// (scx_rustland_core) to be dispatched. To keep track of the accumulated execution time and
+/// vruntime, the scheduler maintains a HashMap (TaskInfoMap), indexed by pid.
 ///
 /// The BPF dispatcher is completely agnostic of the particular scheduling policy implemented in
 /// user-space. For this reason developers that are willing to use this scheduler to experiment
@@ -71,17 +70,16 @@ const SCX_ENQ_WAKEUP: u64 = 1;
 ///
 /// === Troubleshooting ===
 ///
-/// - Reduce the time slice (option `-s`) if you experience audio issues (i.e., cracking audio or
-///   audio packet loss).
+/// - Reduce the time slice (option `-s`) if you experience lag or cracking audio.
 ///
 #[derive(Debug, Parser)]
 struct Opts {
     /// Scheduling slice duration in microseconds.
-    #[clap(short = 's', long, default_value = "5000")]
+    #[clap(short = 's', long, default_value = "20000")]
     slice_us: u64,
 
     /// Scheduling minimum slice duration in microseconds.
-    #[clap(short = 'S', long, default_value = "500")]
+    #[clap(short = 'S', long, default_value = "1000")]
     slice_us_min: u64,
 
     /// If specified, only tasks which have their scheduling policy set to SCHED_EXT using
@@ -118,19 +116,11 @@ struct Opts {
 
 // Time constants.
 const NSEC_PER_USEC: u64 = 1_000;
-const NSEC_PER_SEC: u64 = 1_000_000_000;
-
-// Maximum multiplier for the dynamic task priority.
-const MAX_LATENCY_WEIGHT: u64 = 1_000;
 
 // Basic item stored in the task information map.
 #[derive(Debug)]
 struct TaskInfo {
-    nvcsw: u64,
-    nvcsw_ts: u64,
-    avg_nvcsw: u64,
-    sum_exec_runtime: u64, // total cpu time used by the task
-    vruntime: u64,         // total vruntime of the task
+    vruntime: u64, // total vruntime of the task
 }
 
 // Task information map: store total execution time and vruntime of each task in the system.
@@ -229,10 +219,17 @@ struct Scheduler<'a> {
 
 impl<'a> Scheduler<'a> {
     fn init(opts: &Opts, open_object: &'a mut MaybeUninit<OpenObject>) -> Result<Self> {
-        // Low-level BPF connector.
-        let bpf = BpfScheduler::init(open_object, opts.exit_dump_len, opts.partial, opts.verbose)?;
         let stats_server = StatsServer::new(stats::server_data()).launch()?;
 
+        // Low-level BPF connector.
+        let bpf = BpfScheduler::init(
+            open_object,
+            opts.exit_dump_len,
+            opts.partial,
+            opts.verbose,
+            true, // Enable built-in idle CPU selection policy
+        )?;
+
         info!("{} scheduler attached", SCHEDULER_NAME);
 
         // Return scheduler object.
@@ -281,80 +278,46 @@ impl<'a> Scheduler<'a> {
         ts.as_nanos() as u64
     }
 
-    fn calc_avg(old_val: u64, new_val: u64) -> u64 {
-        (old_val - (old_val >> 2)) + (new_val >> 2)
-    }
-
     // Update task's vruntime based on the information collected from the kernel and return to the
     // caller the evaluated task's deadline.
     //
     // This method implements the main task ordering logic of the scheduler.
     fn update_enqueued(&mut self, task: &QueuedTask) -> u64 {
-        let now = Self::now();
-
         // Get task information if the task is already stored in the task map,
         // otherwise create a new entry for it.
-        let task_info = self
-            .task_map
-            .tasks
-            .entry(task.pid)
-            .or_insert_with_key(|&_pid| TaskInfo {
-                nvcsw: task.nvcsw,
-                nvcsw_ts: now,
-                avg_nvcsw: 0,
-                sum_exec_runtime: task.sum_exec_runtime,
-                vruntime: self.min_vruntime,
-            });
+        let task_info = self.task_map.tasks.entry(task.pid).or_insert(TaskInfo {
+            vruntime: self.min_vruntime,
+        });
 
-        // Refresh voluntary context switches metrics.
-        let delta_t = now - task_info.nvcsw_ts;
-        if delta_t >= NSEC_PER_SEC {
-            let delta_nvcsw = task.nvcsw - task_info.nvcsw;
-            let avg_nvcsw = (delta_nvcsw * NSEC_PER_SEC / delta_t).min(1000);
-
-            task_info.nvcsw = task.nvcsw;
-            task_info.nvcsw_ts = now;
-            task_info.avg_nvcsw = Self::calc_avg(task_info.avg_nvcsw, avg_nvcsw);
+        // Update global minimum vruntime based on the previous task's vruntime.
+        if self.min_vruntime < task.vtime {
+            self.min_vruntime = task.vtime;
         }
 
-        // Evaluate used task time slice.
-        let slice = task
-            .sum_exec_runtime
-            .saturating_sub(task_info.sum_exec_runtime)
-            .min(self.slice_ns);
-
-        // Update total task cputime.
-        task_info.sum_exec_runtime = task.sum_exec_runtime;
-
-        // Update task's vruntime re-aligning it to min_vruntime.
+        // Estimate the used time slice based on total runtime since the last sleep.
         //
-        // The amount of vruntime budget an idle task can accumulate is adjusted in function of its
-        // latency weight, which is derived from the average number of voluntary context switches.
-        // This ensures that latency-sensitive tasks receive a priority boost.
-        let latency_weight = {
-            let base_weight = task_info.avg_nvcsw.min(MAX_LATENCY_WEIGHT);
-            // Boost tasks being enqueued as a result of a wakeup event by an additional 2x factor.
-            let weight_multiplier = if task.flags & SCX_ENQ_WAKEUP != 0 {
-                2
-            } else {
-                1
-            };
-            (base_weight * weight_multiplier) + 1
-        };
-        let min_vruntime = self
-            .min_vruntime
-            .saturating_sub(self.slice_ns * latency_weight);
+        // Cap the value to slice_ns, since exec_runtime accumulates across multiple enqueue
+        // events, but what matters here is the time used in the most recent slice, so:
+        //  - if the task didn't sleep, it's the full slice_ns,
+        //  - if it did sleep, it's exec_runtime.
+        //
+        // Note that there may be some inaccuracies here, as a task can exceed its assigned time
+        // slice due to factors like holding locks or becoming non-deschedulable. These
+        // inaccuracies are tolerated to ensure smoother vruntime progression and prevent excessive
+        // gaps between tasks' vruntimes.
+        let slice = task.exec_runtime.min(self.slice_ns);
+
+        // Update task's vruntime re-aligning it to min_vruntime (never allow a task to accumulate
+        // a budget of more than a time slice to prevent starvation).
+        let min_vruntime = self.min_vruntime.saturating_sub(self.slice_ns);
         if task_info.vruntime < min_vruntime {
             task_info.vruntime = min_vruntime;
         }
         let vslice = slice * 100 / task.weight;
         task_info.vruntime += vslice;
 
-        // Update global minimum vruntime.
-        self.min_vruntime += vslice;
-
         // Return the task's deadline.
-        task_info.vruntime
+        task_info.vruntime + task.exec_runtime.min(self.slice_ns * 100)
     }
 
     // Drain all the tasks from the queued list, update their vruntime (Self::update_enqueued()),
@@ -440,7 +403,7 @@ impl<'a> Scheduler<'a> {
         self.drain_queued_tasks();
         self.dispatch_tasks();
 
-        // Notify the dispatcher if there are still peding tasks to be processed,
+        // Notify the dispatcher if there are still pending tasks to be processed,
         self.bpf.notify_complete(self.task_pool.tasks.len() as u64);
     }
 
