@@ -37,24 +37,41 @@ struct {
 volatile scx_bitmap_t node_data[MAX_NUMA_NODES];
 
 volatile dom_ptr dom_ctxs[MAX_DOMS];
-struct sdt_allocator lb_domain_allocator;
+struct scx_stk lb_domain_allocator;
+
+/*
+ * Declared in its own section because loading the object otherwise fails
+ * with the following error message (where bpf_bpf the fully linked object):
+ *
+ * [WARN] libbpf: map 'bpf_bpf.bss': failed to re-mmap() contents: -524
+ * [WARN] libbpf: map 'bpf_bpf.bss': failed to create: unknown error (-524)(-524)
+ * [WARN] libbpf: failed to load object 'bpf_bpf'
+ */
+private(LBDOMAIN_BUDDY) struct scx_buddy buddy;
+
+#define LBALLOC_PAGES_PER_ALLOC (16)
 
 __weak
 int lb_domain_init(void)
 {
-	return sdt_alloc_init(&lb_domain_allocator, sizeof(struct dom_ctx));
+	if (scx_buddy_init(&buddy, PAGE_SIZE)) {
+		scx_bpf_error("failed to initialize buddy allocator");
+		bpf_printk("failed to initialize buddy allocator");
+	}
+
+	return scx_stk_init(&lb_domain_allocator,
+		sizeof(struct dom_ctx),
+		LBALLOC_PAGES_PER_ALLOC);
 }
 
 __hidden
 dom_ptr lb_domain_alloc(u32 dom_id)
 {
-	struct sdt_data __arena *data = NULL;
 	dom_ptr domc;
 
-	data = sdt_alloc(&lb_domain_allocator);
-
-	domc = (dom_ptr)data->payload;
-	domc->tid = data->tid;
+	domc = (dom_ptr)scx_stk_alloc(&lb_domain_allocator);
+	if (!domc)
+		return NULL;
 
 	domc->cpumask = scx_bitmap_alloc();
 	if (!domc->cpumask) {
@@ -83,13 +100,13 @@ dom_ptr lb_domain_alloc(u32 dom_id)
 __hidden
 void lb_domain_free(dom_ptr domc)
 {
-	sdt_subprog_init_arena();
+	scx_arena_subprog_init();
 
 	scx_bitmap_free(domc->node_cpumask);
 	scx_bitmap_free(domc->direct_greedy_cpumask);
 	scx_bitmap_free(domc->cpumask);
 
-	sdt_free_idx(&lb_domain_allocator, domc->tid.idx);
+	scx_stk_free(&lb_domain_allocator, (void __arena *)domc);
 }
 
 __hidden
@@ -266,7 +283,7 @@ int dom_xfer_task(struct task_struct *p __arg_trusted, u32 new_dom_id, u64 now)
 	dom_ptr from_domc, to_domc;
 	task_ptr taskc;
 
-	taskc = (task_ptr)sdt_task_data(p);
+	taskc = (task_ptr)scx_task_data(p);
 	if (!taskc)
 		return 0;
 
