@@ -1,8 +1,8 @@
+#pragma once
 
 #include <lib/sdt_task.h>
 
 extern volatile dom_ptr dom_ctxs[MAX_DOMS];
-extern struct sdt_allocator lb_domain_allocator;
 
 int lb_domain_init(void);
 dom_ptr lb_domain_alloc(u32 dom_id);
@@ -21,7 +21,7 @@ extern volatile u64 slice_ns;
 extern const volatile u32 nr_doms;
 extern const volatile u32 nr_nodes;
 
-#define lookup_task_ctx(p) ((task_ptr) sdt_task_data(p))
+#define lookup_task_ctx(p) ((task_ptr) scx_task_data(p))
 u32 dom_node_id(u32 dom_id);
 void dom_dcycle_adj(dom_ptr domc, u32 weight, u64 now, bool runnable);
 
