@@ -39,6 +39,7 @@ enum event_type {
 	GPU_MEM,
 	HW_PRESSURE,
 	IPI,
+	PSTATE_SAMPLE,
 	SCHED_REG,
 	SCHED_SWITCH,
 	SCHED_UNREG,
@@ -53,7 +54,7 @@ enum event_type {
 struct sched_switch_event {
 	u32		cpu;
 	bool		preempt;
-	char		next_comm[MAX_COMM];
+	u8		next_comm[MAX_COMM];
 	u64		next_dsq_id;
 	u64		next_dsq_lat_us;
 	u32		next_dsq_nr;
@@ -62,7 +63,7 @@ struct sched_switch_event {
 	u32		next_pid;
 	u32		next_tgid;
 	int		next_prio;
-	char		prev_comm[MAX_COMM];
+	u8		prev_comm[MAX_COMM];
 	u64		prev_dsq_id;
 	u64		prev_used_slice_ns;
 	u64		prev_slice_ns;
@@ -74,8 +75,9 @@ struct sched_switch_event {
 
 struct wakeup_event {
 	u32		pid;
+	u32		tgid;
 	int		prio;
-	char		comm[MAX_COMM];
+	u8		comm[MAX_COMM];
 };
 
 struct set_perf_event {
@@ -93,14 +95,14 @@ struct exit_event {
 	u32		pid;
 	u32		prio;
 	u32		tgid;
-	char		comm[MAX_COMM];
+	u8		comm[MAX_COMM];
 };
 
 struct fork_event {
 	u32		parent_pid;
 	u32		child_pid;
-	char		parent_comm[MAX_COMM];
-	char		child_comm[MAX_COMM];
+	u8		parent_comm[MAX_COMM];
+	u8		child_comm[MAX_COMM];
 };
 
 struct exec_event {
@@ -144,6 +146,10 @@ struct trace_started_event {
 	bool		stop_scheduled;
 };
 
+struct pstate_sample_event {
+	u32             busy;
+};
+
 struct bpf_event {
 	int		type;
 	u64		ts;
@@ -157,6 +163,7 @@ struct bpf_event {
 		struct  cpuhp_enter_event chp;
 		struct  cpuhp_exit_event cxp;
 		struct	ipi_event ipi;
+		struct  pstate_sample_event pstate;
 		struct	sched_switch_event sched_switch;
 		struct	set_perf_event perf;
 		struct	softirq_event softirq;
