@@ -14,6 +14,7 @@ pub mod edm;
 mod event_data;
 mod keymap;
 mod llc_data;
+pub mod mangoapp;
 mod node_data;
 mod perf_event;
 mod perfetto_trace;
@@ -76,6 +77,8 @@ pub enum AppState {
     Scheduler,
     /// Application is in the tracing  state.
     Tracing,
+    /// Application is in the mangoapp state.
+    MangoApp,
 }
 
 #[derive(Clone, Debug, Eq, Hash, Ord, PartialEq, PartialOrd)]
@@ -166,6 +169,7 @@ pub struct SchedWakeActionCtx {
     pub ts: u64,
     pub cpu: u32,
     pub pid: u32,
+    pub tgid: u32,
     pub prio: i32,
     pub comm: SsoString,
 }
@@ -237,6 +241,25 @@ pub struct HwPressureAction {
     pub cpu: u32,
 }
 
+#[derive(Clone, Debug, Eq, Hash, Ord, PartialEq, PartialOrd)]
+pub struct PstateSampleAction {
+    pub cpu: u32,
+    pub busy: u32,
+}
+
+#[derive(Clone, Debug, Eq, Hash, Ord, PartialEq, PartialOrd)]
+pub struct MangoAppAction {
+    pub pid: u32,
+    pub vis_frametime: u64,
+    pub app_frametime: u64,
+    pub fsr_upscale: u8,
+    pub fsr_sharpness: u8,
+    pub latency_ns: u64,
+    pub output_width: u32,
+    pub output_height: u32,
+    pub display_refresh: u16,
+}
+
 #[derive(Clone, Debug, Eq, Hash, Ord, PartialEq, PartialOrd)]
 pub enum Action {
     ChangeTheme,
@@ -257,11 +280,13 @@ pub enum Action {
     IncBpfSampleRate,
     IncTickRate,
     IPI(IPIAction),
+    MangoApp(MangoAppAction),
     NextEvent,
     NextViewState,
     PageDown,
     PageUp,
     PrevEvent,
+    PstateSample(PstateSampleAction),
     Quit,
     RequestTrace,
     TraceStarted(TraceStartedAction),
@@ -364,17 +389,14 @@ impl TryFrom<&bpf_event> for Action {
             #[allow(non_upper_case_globals)]
             bpf_intf::event_type_SCHED_WAKEUP => {
                 let wakeup = unsafe { event.event.wakeup };
-                let comm = unsafe {
-                    std::str::from_utf8(std::slice::from_raw_parts(
-                        event.event.wakeup.comm.as_ptr() as *const u8,
-                        16,
-                    ))
-                    .unwrap()
-                };
+
+                let comm = String::from_utf8_lossy(&wakeup.comm);
+
                 Ok(Action::SchedWakeup(SchedWakeupAction {
                     ts: event.ts,
                     cpu: event.cpu,
                     pid: wakeup.pid,
+                    tgid: wakeup.tgid,
                     prio: wakeup.prio,
                     comm: comm.into(),
                 }))
@@ -382,14 +404,14 @@ impl TryFrom<&bpf_event> for Action {
             #[allow(non_upper_case_globals)]
             bpf_intf::event_type_SCHED_WAKING => {
                 let waking = unsafe { &event.event.waking };
-                let comm = std::str::from_utf8(unsafe {
-                    std::slice::from_raw_parts(waking.comm.as_ptr() as *const u8, 16)
-                })
-                .unwrap();
+
+                let comm = String::from_utf8_lossy(&waking.comm);
+
                 Ok(Action::SchedWaking(SchedWakingAction {
                     ts: event.ts,
                     cpu: event.cpu,
                     pid: waking.pid,
+                    tgid: waking.tgid,
                     prio: waking.prio,
                     comm: comm.into(),
                 }))
@@ -402,13 +424,8 @@ impl TryFrom<&bpf_event> for Action {
             })),
             bpf_intf::event_type_EXIT => {
                 let exit = unsafe { &event.event.exit };
-                let comm = unsafe {
-                    std::str::from_utf8(std::slice::from_raw_parts(
-                        exit.comm.as_ptr() as *const u8,
-                        exit.comm.len(),
-                    ))
-                    .unwrap()
-                };
+                let comm = String::from_utf8_lossy(&exit.comm);
+
                 Ok(Action::Exit(ExitAction {
                     ts: event.ts,
                     cpu: event.cpu,
@@ -421,20 +438,9 @@ impl TryFrom<&bpf_event> for Action {
             #[allow(non_upper_case_globals)]
             bpf_intf::event_type_FORK => {
                 let fork = unsafe { &event.event.fork };
-                let parent_comm = unsafe {
-                    std::str::from_utf8(std::slice::from_raw_parts(
-                        fork.parent_comm.as_ptr() as *const u8,
-                        fork.parent_comm.len(),
-                    ))
-                    .unwrap()
-                };
-                let child_comm = unsafe {
-                    std::str::from_utf8(std::slice::from_raw_parts(
-                        fork.child_comm.as_ptr() as *const u8,
-                        fork.child_comm.len(),
-                    ))
-                    .unwrap()
-                };
+                let parent_comm = String::from_utf8_lossy(&fork.parent_comm);
+                let child_comm = String::from_utf8_lossy(&fork.child_comm);
+
                 Ok(Action::Fork(ForkAction {
                     ts: event.ts,
                     cpu: event.cpu,
@@ -445,22 +451,15 @@ impl TryFrom<&bpf_event> for Action {
                 }))
             }
             #[allow(non_upper_case_globals)]
+            bpf_intf::event_type_PSTATE_SAMPLE => Ok(Action::PstateSample(PstateSampleAction {
+                cpu: event.cpu,
+                busy: unsafe { event.event.pstate.busy },
+            })),
+            #[allow(non_upper_case_globals)]
             bpf_intf::event_type_SCHED_SWITCH => {
                 let sched_switch = unsafe { &event.event.sched_switch };
-                let prev_comm = unsafe {
-                    std::str::from_utf8(std::slice::from_raw_parts(
-                        sched_switch.prev_comm.as_ptr() as *const u8,
-                        sched_switch.prev_comm.len(),
-                    ))
-                    .unwrap()
-                };
-                let next_comm = unsafe {
-                    std::str::from_utf8(std::slice::from_raw_parts(
-                        sched_switch.next_comm.as_ptr() as *const u8,
-                        sched_switch.next_comm.len(),
-                    ))
-                    .unwrap()
-                };
+                let prev_comm = String::from_utf8_lossy(&sched_switch.prev_comm);
+                let next_comm = String::from_utf8_lossy(&sched_switch.next_comm);
 
                 Ok(Action::SchedSwitch(SchedSwitchAction {
                     ts: event.ts,
